/* FILE: out2asc.c             Copyright (c), Symmetric Research 2001-2007
 *
 * This program reads in a binary .hdr and a binary .out file generated
 * by one of the SR acquisition programs and writes out the information
 * in an ASCII text version.  This file can be compiled for use with 
 * either the SR PAR24B or PARxCH A/D board.
 *
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>       // for toupper


#include "srhelper.h"    // for SrFileExist


//****** A/D Specific Data ******


#if defined( SRA2D_PARANY )
#include "parxch.h"

#define ANALOGCHANNELS   PARXCH_MAX_ANALOG_CHANNELS
#define DIGITALCHANNELS  PARXCH_MAX_DIGITAL_CHANNELS
#define GPSCHANNELS      PARXCH_MAX_GPS_CHANNELS
#define COUNTERCHANNELS  PARXCH_MAX_COUNTER_CHANNELS

        

#elif defined( ATOD_IS_PAR24B )

#define ANALOGCHANNELS   8
#define DIGITALCHANNELS  0
#define GPSCHANNELS      0
#define COUNTERCHANNELS  0


#else

#pragma message( "Valid A/D options are SRA2D_PARANY or ATOD_IS_PAR24B." )
#error Invalid A/D option defined.

#define ANALOGCHANNELS   8
#define DIGITALCHANNELS  0
#define GPSCHANNELS      0
#define COUNTERCHANNELS  0

#endif



#define TOTALCHANNELS    (ANALOGCHANNELS + DIGITALCHANNELS +  \
                          GPSCHANNELS + COUNTERCHANNELS)



// SR out header structure

#define MAXCHANNELS             TOTALCHANNELS
#define MAXTITLE                12

#define SROUT_DATACODING_OFFSET  0
#define SROUT_DATACODING_SIGNED  1

typedef struct SrOutHdrLayout {

          char Id[32];

          long Channels;
          long RecordPts;
          long WordSize;
          long DataCoding;

        double SamplingRate;
        double Gain;
        double FilterCoeff;
        double FilterScale;

        struct channel_data { long channel; char title[ MAXTITLE ]; }

                        ChannelList[ MAXCHANNELS ];
        } SROUT_HDR;

SROUT_HDR HDR;

#define HDRSIZE         sizeof( SROUT_HDR )


// This must match the value of SRDAT_HDR_ID in srdat.h
// Dos can not handle multi-character constants so use a long

#if defined( SROS_MSDOS )
#define CHECK_HDR_ID     (0x52535253L)   // = "SRSR"
#else
#define CHECK_HDR_ID    ((long)('RSRS')) // = "SRSR"
#endif





// Prototypes

int  ProcessCommandLine( int argc, char** argv );
void ShowHdr( FILE *fout, SROUT_HDR *Hdr );
void SetSpecialIndex( int *DigitalIndex, int *GpsMarkIndex, int *GpsCtrIndex );
void Usage( char *msg );
void ErrorWarning( char *msg, int status );
void CleanExit( int status );



// Global defines and data

char* UsageMsg[] = {
    "USAGE: out2asc file.HDR file.OUT [/seq | /seqhex] [/hex] [/bare] [/?]",
    "                                                                     ",
    "Out2asc converts the specified .HDR header and .OUT data files to    ",
    "ASCII text format.  The ASCII output filename is the same as the     ",
    "input data filename except the suffix is changed from .OUT to .ASC.  ",
    "If the input file prefix contains a number, it can be treated as the ",
    "first of a sequence where out2asc repeately increments the file      ",
    "prefix and converts the next file.  The optional /seq and /seqhex    ",
    "arguments mean out2asc should increment the prefix in decimal        ",
    "(base 10) and in hex (base 16) respectively.  The optional /hex      ",
    "argument causes the analog data to be written in hex instead of      ",
    "decimal format.  The /bare option shows only the analog data values  ",
    "and none of the header info.  Use /? to show this usage message.     ",
    "                                                                     ",
    "EXAMPLES: out2asc SCOPE.HDR SCOPE.OUT                                ",
    "          out2asc SIMP.HDR 00000010.OUT /seq /hex                    ",
    NULL
    };


#define MAXFILENAME      256
#define MAXSTR           256
#define MAXFMT           20

#define SEQTYPE_NONE  0
#define SEQTYPE_DEC   1
#define SEQTYPE_HEX   2



/* Sign extension constants
 *
 * PARxCH data can be acquired in 24-bit signed or offset format.
 * When stored as a 32-bit number, the 24-bit signed format data must
 * be sign extended if the 24th bit is set.
 *
 */

#define NEG24BIT        0x00800000
#define SIGNEXTEND      0xFF000000
#define NOSIGNEXTEND    0x00000000



int  SeqType, Sequential, HexFormat, ShowAllInfo;
long SeqNum;
FILE *HeaderFile, *BinaryFile, *AsciiFile;
char HeaderName[MAXFILENAME], BinaryName[MAXFILENAME], AsciiName[MAXFILENAME];
char Msg[MAXSTR];




//------------------------------------------------------------------------------
// ROUTINE: MAIN (out2asc)
// PURPOSE: To convert binary .hdr header and .out data files into ASCII.  See
//          UsageMsg above for a description of the command line arguments.
//------------------------------------------------------------------------------
int main( int argc, char *argv[] ) {

        char         MarkStar, DataFmt[MAXFMT], DataTitle[MAXFMT];
	int          i, DigitalIndex, GpsMarkIndex, GpsCtrIndex;
        long         Nsample, SetTopBits, DataBuffer[TOTALCHANNELS];
        unsigned int Nbytes;


        // Announce program version.

        printf( "OUT2ASC [/?]: (c) SR 2007 : Rev {01/15/07}\n" );




        // Process command line arguments

        ProcessCommandLine( argc, argv );
        printf( "\n" );



        // Init file handles

        HeaderFile = NULL;
        BinaryFile = NULL;
        AsciiFile  = NULL;

	
	// Process header file - check existance, open, read, close.

        if ( !SrFileExist( HeaderName ) ) {
                sprintf( Msg, "Input header file %s does not exist", HeaderName );
                ErrorWarning( Msg, 1 );
                }

	HeaderFile = fopen( HeaderName, "rb" );
	if ( !HeaderFile ) {
                sprintf( Msg, "Can't open input header file %s", HeaderName );
                ErrorWarning( Msg, 1 );
                }


        printf( "Reading file %s ...\n", HeaderName );

        fread( &HDR, 1, HDRSIZE, HeaderFile );

        if ( ( strlen( HDR.Id ) > 32 )        ||
             ( HDR.Id[0] != 'S' )             ||
             ( HDR.Id[1] != 'R' )             ||
             ( HDR.Channels < 0 )             ||
             ( HDR.Channels > TOTALCHANNELS ) ) {
                sprintf( Msg, "Header file %s is invalid", HeaderName );
                ErrorWarning( Msg, 1 );
                }

        fclose( HeaderFile );  HeaderFile = NULL;



        // Set parameters based on header info ...

        // Sign extension constant for signed vs offset data

        if (HDR.DataCoding == 1)
                SetTopBits = SIGNEXTEND;
        else
                SetTopBits = NOSIGNEXTEND;


        // Format and title strings for hex vs decimal output

        if ( HexFormat == 1 ) {
                strncpy( DataTitle, "\nHex Data:\n", MAXFMT );
                strncpy( DataFmt, "%08lX ", MAXFMT);
                }
        else {
                strncpy( DataTitle, "\nData:\n", MAXFMT );
                strncpy( DataFmt, "%8ld ", MAXFMT);
                }


        // Special channel indicies for digital, GPS mark and counter channels

        SetSpecialIndex( &DigitalIndex, &GpsMarkIndex, &GpsCtrIndex );
        




        
        // Loop until no more out files requested or found

        printf( "Converting to ASCII:\n" );

        while ( 1 ) {

                // First check for existance of input data file

                if ( !SrFileExist( BinaryName ) ) {
                        sprintf( Msg, "No more files to process.  Stopping at %s.", BinaryName);
                        ErrorWarning( Msg, 0 );
                        }
                
        
                // Open input and output files

                BinaryFile = fopen( BinaryName, "rb" );
                if ( !BinaryFile ) {
                        sprintf( Msg, "Can not open input data file %s", BinaryName );
                        ErrorWarning( Msg, 1 );
                        }
                
                AsciiFile = fopen( AsciiName, "w" );
                if ( !AsciiFile ) {
                        sprintf( Msg, "Can not open output ASCII file %s", AsciiName );
                        ErrorWarning( Msg, 1 );
                        }


                // There is no definitive test to ensure the binary file IS
                // an OUT file, but at least make sure it is NOT a DAT file.
                // Get first point, check it, and rewind.
                
                fread( DataBuffer, 1, 4, BinaryFile );
                if ( DataBuffer[0] == CHECK_HDR_ID ) {
                        sprintf( Msg, "Input file %s must be .OUT not .DAT format", BinaryName );
                        ErrorWarning( Msg, 1 );
                        }
                fseek( BinaryFile, 0L, SEEK_SET );



                // Write header info to ASCII file

                if ( ShowAllInfo )
                        ShowHdr( AsciiFile, &HDR );



                // Write data column titles to ASCII file

                if ( ShowAllInfo ) {
                        fprintf( AsciiFile, DataTitle );

                        for ( i = 0 ; i < HDR.Channels ; i++ )
                                fprintf( AsciiFile, "%8s ", HDR.ChannelList[i].title );

                        fprintf( AsciiFile, "   Index\n" );
                        }




                // Convert the binary data file to ASCII

                printf( "Reading file %s ...\n", BinaryName );
                printf( "Writing file %s ...\n", AsciiName );


                Nbytes  = (unsigned int)(HDR.Channels * sizeof( long ));
                Nsample = 0;

                while ( fread( DataBuffer, 1, Nbytes, BinaryFile ) == Nbytes ) {

                        // Convert the samples for each channel at this time point.

                        for ( i = 0 ; i < HDR.Channels ; i++ ) {

                                // If not a special channel, 
                                // sign extend 24 bit data to 32 bits.

                                if ( (i != DigitalIndex) &&
                                     (i != GpsMarkIndex) &&
                                     (i != GpsCtrIndex)  &&
                                     (DataBuffer[i] & NEG24BIT) )
                                        DataBuffer[i] |= SetTopBits;

                                fprintf( AsciiFile, DataFmt, DataBuffer[i] );
                                }



                        // Add final columns with the time index
                        // and GPS mark indicator.

                        if ( GpsMarkIndex == -1 || DataBuffer[GpsMarkIndex] == 0 )
                                MarkStar = ' ';
                        else
                                MarkStar = '<';

                        if ( ShowAllInfo )
                                fprintf( AsciiFile, "%8ld  %c\n", Nsample, MarkStar );
                        else
                                fprintf( AsciiFile, "\n" );

                        Nsample++;

                        } // end while fread



                // Close current files

                fclose( BinaryFile );  BinaryFile = NULL;
                fclose( AsciiFile );   AsciiFile  = NULL;


                
                // If only a single file was requested, we're done
                
                if (!Sequential)
                        break;

                // Otherwise, compute the next set of sequential file names
                
                else {
                        SeqNum++;
                        if ( SeqType == SEQTYPE_HEX ) {
                                sprintf( BinaryName, "%08lX.OUT", SeqNum );
                                sprintf( AsciiName,  "%08lX.ASC", SeqNum );
                                }
                        else { // ( SeqType == SEQTYPE_DEC ) 
                                sprintf( BinaryName, "%08ld.OUT", SeqNum );
                                sprintf( AsciiName,  "%08ld.ASC", SeqNum );
                                }
                        }


                } // end while 1 (more files to process)


        return( 0 );

}

//------------------------------------------------------------------------------
// ROUTINE: ProcessCommandLine
// PURPOSE: Extract information provided by the user on the program command line.
//          See the UsageMsg structure defined above for allowed options.
//------------------------------------------------------------------------------
int ProcessCommandLine( int argc, char** argv ) {

        int  i, iprefix, len, useSeq, hdrFileArg, outFileArg;
        long ltmp;
        char *p, FilePrefix[MAXFILENAME], PrefixCheck[MAXFILENAME];


        // Set defaults (global and local)

        SeqType     = SEQTYPE_NONE;
        HexFormat   = 0;
        ShowAllInfo = 1;

        useSeq      = 0;
        hdrFileArg  = 0;
        outFileArg  = 0;


        // Process command line arguments

        for ( i = 1 ; i < argc ; i++ ) {

                if (strcmp("/?",argv[i]) == 0)
                        Usage( "?" );

                else if (strcmp("/seq",argv[i]) == 0) {
                        if (useSeq == 1)
                                Usage( "Only one of /seq and /seqhex is allowed" );
                        useSeq = 1;
                        SeqType = SEQTYPE_DEC;
                        }

                else if (strcmp("/seqhex",argv[i]) == 0) {
                        if (useSeq == 1)
                                Usage( "Only one of /seq and /seqhex is allowed" );
                        useSeq = 1;
                        SeqType = SEQTYPE_HEX;
                        }

                else if (strcmp("/hex",argv[i]) == 0)
                        HexFormat = 1;

                else if (strcmp("/bare",argv[i]) == 0)
                        ShowAllInfo = 0;

                else if (hdrFileArg == 0)
                        hdrFileArg = i;

                else if (outFileArg == 0)
                        outFileArg = i;

                else
                        Usage( "Bad command line argument" );
                }

        if (hdrFileArg == 0) // Did not find any hdr file name
                Usage( "Command line must include a header file name." );
        
        if (outFileArg == 0) // Did not find any out file name
                Usage( "Command line must include a .out data file name." );

        if ( (argc < 3) || (argc > 6) )
                Usage( "Wrong number of command line arguments.  Must be 2 to 5." );

        
        // Fill some file name parameters
        
        strncpy( HeaderName, argv[hdrFileArg], MAXFILENAME-1 );
        strncpy( BinaryName, argv[outFileArg], MAXFILENAME-1 );
        HeaderName[MAXFILENAME-1] = '\0';
        BinaryName[MAXFILENAME-1] = '\0';
        

        // Find the end of the input file prefix

        len = strlen( BinaryName );
        iprefix = -1;
        for ( i = len ; i > 0 ; i-- ) {
                if (BinaryName[i-1] == '.') {
                        iprefix = i-1;
                        break;
                        }
                }
        
        if (iprefix < 0 || iprefix > len || iprefix > MAXFILENAME-4) {
                sprintf( Msg, "File name %s is invalid", BinaryName );
                Usage( Msg );
                }


        // Extract the prefix, null terminate, and uppercase

        strncpy( FilePrefix, BinaryName, iprefix );
        FilePrefix[iprefix] = '\0';

        p   = FilePrefix;
        len = strlen(FilePrefix);
        for ( i = 0 ; i < len ; i++ ) {
                *p = (char)toupper( *p );
                *p++;
                }



        // If optional sequential argument was found, determine starting 
        // sequential number from prefix and verify.

        Sequential = 0;
        SeqNum     = -1;
        
        if ( useSeq ) {
                        
                Sequential = 1;

                if (SeqType == SEQTYPE_HEX) {
                        if ( sscanf(FilePrefix, "%lX", &ltmp) == 1 )
                                SeqNum = ltmp;

                        sprintf( PrefixCheck, "%08lX", SeqNum );
                        if ( strcmp( FilePrefix, PrefixCheck ) != 0 )
                                Usage( "Can't set sequential hex value from filename" );
                        }
                else { // (SeqType == SEQTYPE_DEC)
                        if ( sscanf(FilePrefix, "%ld", &ltmp) == 1 )
                                SeqNum = ltmp;

                        sprintf( PrefixCheck, "%08ld", SeqNum );
                        if ( strcmp( FilePrefix, PrefixCheck ) != 0 )
                                Usage( "Can't set sequential decimal value from filename" );
                        }

                } // end useSeq



        // Make ASCII output name

        sprintf( AsciiName, "%s.ASC", FilePrefix );


        return( 1 );
}


//------------------------------------------------------------------------------
// ROUTINE: ShowHdr
// PURPOSE: Write header information to the ASCII output file.
//------------------------------------------------------------------------------
void ShowHdr( FILE *fout, SROUT_HDR *Hdr ) {

        int i;
        char *DataCodeName;

        if ( Hdr->DataCoding == SROUT_DATACODING_SIGNED )
                DataCodeName = "signed";
        else // Hdr->DataCoding == SROUT_DATACODING_OFFSET
                DataCodeName = "offset";
        
        
        fprintf( fout, "Header:\n" );
        fprintf( fout, "           Id = %s\n",       Hdr->Id           );
        fprintf( fout, "     Channels = %ld\n",      Hdr->Channels     );
        fprintf( fout, "    RecordPts = %ld\n",      Hdr->RecordPts    );
        fprintf( fout, "     WordSize = %ld\n",      Hdr->WordSize     );
        fprintf( fout, "   DataCoding = %ld (%s)\n", Hdr->DataCoding, DataCodeName );
        fprintf( fout, " SamplingRate = %f\n",       Hdr->SamplingRate );
        fprintf( fout, "         Gain = %f\n",       Hdr->Gain         );
        fprintf( fout, "  FilterCoeff = %f\n",       Hdr->FilterCoeff  );
        fprintf( fout, "  FilterScale = %f\n",       Hdr->FilterScale  );

        for ( i = 0 ; i < Hdr->Channels ; i++ )

                fprintf( fout, "   Channel[%1ld] = %s\n",
                                 Hdr->ChannelList[i].channel,
                                 Hdr->ChannelList[i].title
                               );
}


//------------------------------------------------------------------------------
// ROUTINE: SetSpecialIndex
// PURPOSE: Determine the index values of the analog channels that have special
//          meanings.  These include the GPS mark channel and, on the PAR8CH,
//          the packed digital channels and the GPS OnBoardCounter.
//------------------------------------------------------------------------------
void SetSpecialIndex( int *DigitalIndex, int *GpsMarkIndex, int *GpsCtrIndex ) {

        int i;

        // Set defaults to indicate none

        *DigitalIndex = -1;
        *GpsMarkIndex = -1;
        *GpsCtrIndex  = -1;


        // Look for digital and GPS mark indicies by channel name

        for ( i = 0 ; i < HDR.Channels ; i++ ) {
                
                if (strcmp(HDR.ChannelList[i].title,"DGTL") == 0)
                        *DigitalIndex = i;

                else if (strcmp(HDR.ChannelList[i].title,"MARK") == 0)
                        *GpsMarkIndex = i;
                }

        
        // Compute counter index from GPS mark index

        if (*GpsMarkIndex != -1 && *GpsMarkIndex < HDR.Channels)
                *GpsCtrIndex = *GpsMarkIndex + 1;
}

//------------------------------------------------------------------------------
// ROUTINE: Usage
// PURPOSE: Print program usage message to the screen.
//------------------------------------------------------------------------------
void Usage( char *msg ) {

        int i;

        // Select error line depending on 
        // 1) no message, 2) help requested, or 3) message given

        if ( !msg )
                printf( "\nERROR: Command line usage.\n\n" );

        else if ( strcmp( msg, "?" ) == 0 )
                printf( "\n" );

        else
                printf( "\nERROR: Command line usage (%s).\n\n", msg );


        for ( i = 0 ; UsageMsg[i] ; i++ )
                printf( "%s\n", UsageMsg[i] );

        exit( 1 );
}


//------------------------------------------------------------------------------
// ROUTINE: ErrorWarning
// PURPOSE: Print the supplied error message to the screen and exit cleanly
//          with the supplied status.
//------------------------------------------------------------------------------
void ErrorWarning( char *msg, int status ) {
        printf("OUT2ASC ERROR: %s\n", msg );
        CleanExit( status );
}


//------------------------------------------------------------------------------
// ROUTINE: CleanExit
// PURPOSE: Exit cleanly by ensuring that all open files are closed before exit.
//------------------------------------------------------------------------------
void CleanExit( int status ) {

        // Close any open files

        if ( HeaderFile != NULL )
                fclose( HeaderFile );

        if ( BinaryFile != NULL )
                fclose( BinaryFile );

        if ( AsciiFile != NULL )
                fclose( AsciiFile );

        // Exit

        exit( status );
}
