/* FILE: indriver.c              Copyright (c), Symmetric Research, 2003-2007
 *
 * Program to install the Win2K PARxCH kernel mode driver.
 *
 * This is done by filling in registry entries for the driver service
 * key, preparing a device information list and element for the selected
 * driver, and calling the UpdateDriversForPlugAndPlayDevices function
 * to do the install.  The symres.inf file is required for correct operation.
 *
 */



// Needed in setupapi for Win9x and WinNT compatibility (Win2K could use V2)
#define  USE_SP_DRVINFO_DATA_V1 1

#include <windows.h>
#include <setupapi.h>   // for SetupDi functions
#include <newdev.h>     // for UpdateDriverForPlugAndPlayDevices function
#include <stdio.h>
#include <stdlib.h>

#include "parxchkd.h"   // for REG_PARM constants


// Defines and typedefs

#define MAXSTR    256

#define NOT_FATAL 0
#define FATAL     1

#define INF_PATH  "\\symres.inf"

#define DRIVER_NAME_BASE "SrParXch"
#define DRIVER_NAME_FMT  "%s%d"
#define DRIVER_NUM_MAX   3

typedef struct RegValueStruct { // used for making registry entries
        CHAR *Name;
        DWORD Type;
        DWORD ValueDw;
        CHAR *ValueSz;
        } REGVALUE;


//GUID ClassGuidInf = {
//        0x00000000,                                      // Data1
//        0x0000,                                          // Data2
//        0x0000,                                          // Data3
//        {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00} // Data4[8]
//        };




// Usage and Error message arrays:

char *UsageMsg[] = {
  "USAGE: indriver XchModel PortAddr PortIrq                                ",
  "                                                                         ",
  "This utility installs the PARxCH device driver which supports all        ",
  "XchModels of the PARxCH family of A/D boards for x = 1, 4, or 8.         ",
  "                                                                         ",
  "   XchModel = PAR1CH, PAR4CH, or PAR8CH                                  ",
  "   PortAddr = parallel port address  (like 0x378, 0x278 or 0xFFF8)       ",
  "   PortIrq  = parallel port IRQ      (typically 7, 5, or 17)             ",
  "                                                                         ",
  "Use the Device Manager to find your PC's available addresses and irqs    ",
  "by selecting the Ports group, choosing a parallel port, viewing its      ",
  "properties, and checking the Resources tab.                              ",
  "                                                                         ",
  "The default XchModel can be changed with the setmodel utility or         ",
  "overridden within applications like scope.                               ",
  "                                                                         ",
  "Use PortIrq of 0 to install without interrupts.  PARGPS timestamping     ",
  "requires interrupts to work.                                             ",
  "                                                                         ",
  "INDRIVER EXAMPLES: > indriver PAR4CH 0x378 7                             ",
  "                   > indriver PAR8CH 0x278 5                             ",
  "                   > indriver PAR1CH 0xFFF8 17                           ",
  NULL
  };

char *StartupMsg[] = {
  "                                                                         ",
  "Note: Parallel port interrupts must be enabled before installing the     ",
  "PARxCH device driver when using PARGPS.  See readme.txt for more details.",
  "                                                                         ",
  NULL
  };

char *UpdatingMsg[] = {
  "                                                                       ",
  "   Preparing to update driver.  This may take a minute or two while    ",
  "   the inf file and registry are searched for matching devices ...     ",
  NULL
  };


char *SuccessMsg[] = {
  "                                                                       ",
  "SUCCESS: The device has been installed.                                ",
  "                                                                       ",
  "Now run the PARxCH diagnostic program diag to verify the PARxCH is     ",
  "working correctly.  If using the PARGPS, please install its driver     ",
  "using the indriver utility located in the \\sr\\pargps\\driver directory.",
  "                                                                       ",
  "You will need the exact driver name to access the PARxCH.  If you      ",
  "forget it, showdriver will display it again.  The name of the installed",
  "PARxCH device driver is:                                               ",
  "                                                                       ",
  NULL
  };


char *ErrorNoMoreItemsMsg[] = {
  "                                                                      ",
  "This message indicates the underlying parallel port driver could not  ",
  "provide the needed resources.  You can check the system Event Viewer  ",
  "for more details, but the most likely causes are:                     ",
  "                                                                      ",
  "   1) The port is in use by another device like a printer or PAR4CH.  ",
  "   2) The parallel port is using a different address.  eg you asked   ",
  "      for 0x378, but it is using 0x278.                               ",
  "   3) Parallel port interrupts are not enabled.                       ",
  "                                                                      ",
  "Run rmdriver to remove the failed PARxCH installation and see the     ",
  "troubleshooting section of readme.txt for details on using the Device ",
  "Manager to correct the parallel port settings or to remove conflicting",
  "drivers by other vendors.  The cleanest way to remove a conflicting SR",
  "driver is by using the rmdriver program for that product.  Then try   ",
  "the PARxCH installation again.                                        ",
  NULL
  };



// Helper function protos:

int RemoveFailedDriver( void );
int ScanDrivers( void );
int CheckForExisting( GUID *ClassGuidInf, char *HardwareId );
int CheckForDriverStart( CHAR *DriverName );
BOOL CreateServiceKey( CHAR *DriverName, int DriverNum, int DriverAddr, int DriverIrq, CHAR *XchModelName );
BOOL FillRegKeyWithValues( HKEY Key, REGVALUE Values[] );
void Usage( char *msg );
void PrintMsg( char *MsgArray[] );
void Error( char *Msg, int Fatal );


// Global variables

char DriverName[MAXSTR];
int  DriverNum, PortAddr, PortIrq;

struct Driver_Info {
        char DriverName[MAXSTR];
        int  DriverNumber;
        int  PortAddr;
        int  PortIrq;
        int  InUse;
        } DriverInfo[DRIVER_NUM_MAX];





// Main:

void main( int argc, char **argv ) {

        int  i, nscan1, nscan2;
        BOOL Ok, RebootRequired;
        DWORD BufferSize, HardwareIdSize;
        CHAR Buffer[MAXSTR], HardwareId[MAXSTR];
        CHAR XchModelName[MAXSTR], ClassName[MAXSTR], InfFileName[MAX_PATH];
        GUID ClassGuidInf;
        HDEVINFO DeviceInfoSet;
        SP_DEVINFO_DATA DeviceInfoData;
        SP_DEVINSTALL_PARAMS DeviceInstallParams;


        // Banner message

        printf( "PARxCH Kernel Mode Driver Installation Program : Rev {08/10/07}\n");




        // Process command line arguments for addr, irq, xchmodel

        if ( argc != 4 ) 
                Usage( "Wrong number of arguments" );

        strncpy( XchModelName, argv[1], MAXSTR );
        nscan1 = sscanf( argv[2], "0x%X", &PortAddr );
        nscan2 = sscanf( argv[3], "%d",   &PortIrq );

        
        // Do preliminary error check on arguments

        if ( nscan1 != 1 )
                Usage( "Could not read port address" );

        if ( nscan2 != 1 )
                Usage( "Could not read port IRQ " );

        if ( PortIrq < 0 || PortIrq > 30 )  // should be 5,7,17,19, etc
                Usage( "Port IRQ out of range" );

        if ( (stricmp(XchModelName,"PAR1CH") != 0) &&
             (stricmp(XchModelName,"PAR4CH") != 0) &&
             (stricmp(XchModelName,"PAR8CH") != 0) )
                Usage( "Invalid XchModel, must be PAR1CH, PAR4CH, PAR8CH" );



        // Scan installed drivers

        if ( !ScanDrivers() ) {
                printf( "\n\nERROR: All available driver names are currently in use.\n");
                printf( "       Please use rmdriver to uninstall some before running indriver again.\n");
                exit( 1 );
                }
        


        // Error check for duplicate address

        for ( i = 0 ; i < DRIVER_NUM_MAX ; i++ ) {
                if ( PortAddr == DriverInfo[i].PortAddr ) {
                        printf( "\n\nERROR: Port Address 0x%X already in use for driver %s\n",
                                  PortAddr, DriverInfo[i].DriverName );
                        printf( "       Please select another address.\n" );
                        exit( 1 );
                        }
                }
        

        // Select first available driver name 

        for ( i = 0 ; i < DRIVER_NUM_MAX ; i++ ) {
                if ( DriverInfo[i].InUse == 0 ) {
                        strcpy( DriverName, DriverInfo[i].DriverName );
                        DriverNum = DriverInfo[i].DriverNumber;
                        break;
                        }
                }

        
        


        // Start up message with note about ParPort interrupts

        PrintMsg( StartupMsg );
        printf( "ATTEMPTING TO INSTALL NEW DRIVER %s ...\n\n", DriverName);
        printf( "   ... with Address = 0x%X, IRQ = %d, XchModel = %s\n",
                  PortAddr, PortIrq, XchModelName );




        // Get Class GUID from inf file assuming that this program
        // is being run from the sr\parxch\driver directory and that
        // the inf file is in the same directory.

        strncpy( Buffer, ".", MAXSTR-1 );  Buffer[MAXSTR-1] = '\0';
        InfFileName[0] = '\0';
        ClassName[0]   = '\0';
        
        GetCurrentDirectory( sizeof(Buffer), Buffer );

        if ( (strlen(Buffer) + strlen(INF_PATH)) > MAXSTR-1)
                Error( ".inf file relative path name is too long", FATAL );
        else
                sprintf( InfFileName, "%s%s", Buffer, INF_PATH );

        Ok = SetupDiGetINFClass(
                                InfFileName,
                                &ClassGuidInf,
                                ClassName,
                                sizeof(ClassName),
                                0
                               );
        if (!Ok)
                Error( "Failed to read GUID from inf file", FATAL );
        else
                printf( "   ... read GUID from inf file %s\n", InfFileName );




        // Create the HardwareID.
        // It is a multi-sz string, so it must be double null terminated.

        sprintf( HardwareId, "Root\\%s&%s", ClassName, DriverName );
        HardwareIdSize = (lstrlen(HardwareId)+1+1)*sizeof(CHAR);
        HardwareId[HardwareIdSize-1] = '\0';
        HardwareId[HardwareIdSize]   = '\0';
        HardwareId[HardwareIdSize+1] = '\0';




        // Check that the device is not already installed.

        if ( CheckForExisting( &ClassGuidInf, HardwareId ) ) {
                printf( "\n\nERROR: Device %s already exists.\n\n", DriverName );
                printf( "Use rmdriver to remove, then reinstall.\n" );
                exit( 1 );
                }




        // Add Service key entries to the registry
        // WCT: It would be better if the service install section in
        // the INF file would do this work instead.

        Ok = CreateServiceKey( DriverName, DriverNum, PortAddr, PortIrq, XchModelName );
        if ( !Ok )
                Error( "Failed to add service and parameters keys", NOT_FATAL );
        else
                printf( "   ... service key %s added to registry\n", DriverName );





        // Create container for the to-be-created Device Information Element.
        
        DeviceInfoSet = SetupDiCreateDeviceInfoList( &ClassGuidInf, 0 );
        
        if (DeviceInfoSet == INVALID_HANDLE_VALUE)
                Error( "Failed to create a device list", FATAL );


        
        // Now create the new device element. 

        DeviceInfoData.cbSize = sizeof(SP_DEVINFO_DATA);
        
        Ok = SetupDiCreateDeviceInfo(
                                     DeviceInfoSet,
                                     ClassName,
                                     &ClassGuidInf,
                                     NULL,          // device description
                                     0,
                                     DICD_GENERATE_ID,
                                     &DeviceInfoData
                                    );
        if (!Ok ) {
                SetupDiDestroyDeviceInfoList( DeviceInfoSet );
                Error( "Failed a device element", FATAL );
                }




        // Add the HardwareID to the new device element.

        Ok = SetupDiSetDeviceRegistryProperty(
                                              DeviceInfoSet,
                                              &DeviceInfoData,
                                              SPDRP_HARDWAREID,
                                              (LPBYTE)HardwareId,
                                              HardwareIdSize
                                             );
        
        if (!Ok ) {
                SetupDiDestroyDeviceInfoList( DeviceInfoSet );
                Error( "Failed to set hardware id property", FATAL );
                }



        // Set location information for the new device element.
        
        strncpy( Buffer, "External Device", MAXSTR-1 );  Buffer[MAXSTR-1] = '\0';
        BufferSize = strlen( Buffer ) + 1;
        
        Ok = SetupDiSetDeviceRegistryProperty(
                                              DeviceInfoSet,
                                              &DeviceInfoData,
                                              SPDRP_LOCATION_INFORMATION,
                                              (LPBYTE)Buffer,
                                              BufferSize
                                             );
        if ( !Ok ) {
                SetupDiDestroyDeviceInfoList( DeviceInfoSet );
                Error( "Failed to set location property", NOT_FATAL );
                }




        // Transform the registry element into an actual devnode 
        // in the PnP HW tree.

        Ok = SetupDiCallClassInstaller(
                                       DIF_REGISTERDEVICE,
                                       DeviceInfoSet,
                                       &DeviceInfoData
                                      );
        if ( !Ok ) {
                SetupDiDestroyDeviceInfoList( DeviceInfoSet );
                Error( "Failed to register the device", FATAL );
                }
        else
                printf( "   ... device %s registered\n", HardwareId );




        // Update device installation parameters
        
        DeviceInstallParams.cbSize = sizeof(SP_DEVINSTALL_PARAMS);

        Ok = SetupDiGetDeviceInstallParams(
                                           DeviceInfoSet,
                                           &DeviceInfoData,
                                           &DeviceInstallParams
                                          );
        if ( !Ok ) {
                SetupDiDestroyDeviceInfoList( DeviceInfoSet );
                Error( "Failed to get old install parms", FATAL );
                }


        // Possibly add DI_QUIETINSTALL to flags below
        DeviceInstallParams.Flags |= DI_ENUMSINGLEINF;
        strncpy( DeviceInstallParams.DriverPath, InfFileName, MAX_PATH-1 );
        DeviceInstallParams.DriverPath[MAX_PATH-1] = '\0';

        Ok = SetupDiSetDeviceInstallParams(
                                           DeviceInfoSet,
                                           &DeviceInfoData,
                                           &DeviceInstallParams
                                          );
        if ( !Ok ) {
                SetupDiDestroyDeviceInfoList( DeviceInfoSet );
                Error( "Failed to set new install parms", FATAL );
                }





        // Install the driver.

        PrintMsg( UpdatingMsg );
        
        Ok = UpdateDriverForPlugAndPlayDevices(
                        0,
                        HardwareId,
                        InfFileName,
                        0,              // or INSTALLFLAG_FORCE
                        &RebootRequired
                        );

        if (RebootRequired) {
                printf( "\n\nERROR: Driver can not get needed resources.\n" );
                RemoveFailedDriver();
                PrintMsg( ErrorNoMoreItemsMsg );
                exit( 1 );
                }

        if ( !Ok ) {
                SetupDiDestroyDeviceInfoList( DeviceInfoSet );
                Error( "Failed to update driver.  Please run rmdriver", FATAL );
                }



        // Destroy driver and device lists

        SetupDiDestroyDeviceInfoList( DeviceInfoSet );



        // Verify that the driver really started by checking for a
        // valid PortFound registry entry in the driver service key.
        // It is important to do this since XP does not appear to
        // recognize the driver failing to start as an error condition.

        if ( !CheckForDriverStart( DriverName ) ) {
                printf( "\n\nERROR: Driver can not start.\n" );
                RemoveFailedDriver();
                PrintMsg( ErrorNoMoreItemsMsg );
                exit( 1 );
                }

        
        // Announce success

        PrintMsg( SuccessMsg );
        printf( "\t\t%s   (at 0x%03lX, irq %d)\n", DriverName, PortAddr, PortIrq );
        
}

int RemoveFailedDriver( void ) {

        int  ret;
        char RemoveCommandString[MAXSTR];

        sprintf( RemoveCommandString, "rmdriver %s /s", DriverName );
        
        ret = system( RemoveCommandString );
        
        return( ret );
}

int ScanDrivers( void ) {

        int   i, AnyFree, CurrentAddress, CurrentIrq;
        char  ServiceParmKeyName[MAXSTR];
        HKEY  ServiceParmKey;
        DWORD Code, BufferSize;


        // Check for drivers that are currently installed

        AnyFree = 0;

        printf( "\nCurrent Driver Status:\n\n" );
        
        for ( i = 0 ; i < DRIVER_NUM_MAX ; i++ ) {

                // Set up trial driver name and default parameters
                
                sprintf( DriverInfo[i].DriverName, DRIVER_NAME_FMT, DRIVER_NAME_BASE, i );
                DriverInfo[i].DriverNumber = i;
                DriverInfo[i].PortAddr     = 0;
                DriverInfo[i].PortIrq      = 0;
                DriverInfo[i].InUse        = 0;
                

                // Open service parameters registry key

                sprintf( ServiceParmKeyName,
                         "System\\CurrentControlSet\\Services\\%s\\Parameters",
                         DriverInfo[i].DriverName );

                Code = RegOpenKeyEx(
                               HKEY_LOCAL_MACHINE,      // already open key
                               ServiceParmKeyName,      // subkey to open
                               0,                       // reserved
                               KEY_ALL_ACCESS,          // full access permission
                               &ServiceParmKey          // handle to key
                              );



                // If key was opened, driver is installed so get parameters
                
                if (Code == ERROR_SUCCESS) {

                        DriverInfo[i].InUse = 1;

                        BufferSize = sizeof( CurrentAddress );
                        Code = RegQueryValueEx(
                                               ServiceParmKey,
                                               REGENVSTR_PORTADDR,
                                               NULL,
                                               NULL,
                                               (LPBYTE)&CurrentAddress,
                                               &BufferSize
                                              );

                        if (Code == ERROR_SUCCESS)
                                DriverInfo[i].PortAddr = CurrentAddress;
                        else
                                printf( "  Error: Could not determine address (err = %d).\n", Code );


                        BufferSize = sizeof( CurrentIrq );
                        Code = RegQueryValueEx(
                                               ServiceParmKey,
                                               REGENVSTR_PORTIRQ,
                                               NULL,
                                               NULL,
                                               (LPBYTE)&CurrentIrq,
                                               &BufferSize
                                              );

                        if (Code == ERROR_SUCCESS)
                                DriverInfo[i].PortIrq = CurrentIrq;
                        else
                                printf( "  Error: Could not determine irq (err = %d).\n", Code );

                        printf( "  Driver %s is  installed with Addr 0x%X, Irq %d.\n",
                                  DriverInfo[i].DriverName,
                                  DriverInfo[i].PortAddr,
                                  DriverInfo[i].PortIrq
                              );

                        Code = RegCloseKey( ServiceParmKey );
                        if (Code != ERROR_SUCCESS)
                                printf( "  Error: Could not close registry key (err = %d).\n", Code );
                        }
                
                // Else key does not open and driver is not installed
                
                else {
                        printf( "  Driver %s not installed.\n", DriverInfo[i].DriverName );
                        AnyFree = 1;
                        }

                } // end for i

        return( AnyFree );
}

int CheckForExisting( GUID *ClassGuidInf, char *HardwareId ) {

        int             i;
        DWORD           Result, BufferSize, RegType;
        CHAR            Buffer[MAXSTR], *Ptr;
        HDEVINFO        DeviceInfoSet;
        SP_DEVINFO_DATA DeviceInfoData;

        // Check if a device with the specified guid and hardware id
        // already exists.  If so, return 1.  Otherwise, return 0.


        // Create a Device Information Set with all present devices
        // having specified class guid.
        
        DeviceInfoSet = SetupDiGetClassDevs( ClassGuidInf, 0, 0, DIGCF_PRESENT );
        
        if ( DeviceInfoSet == INVALID_HANDLE_VALUE )
                Error( "Failed to create a device list", FATAL );
    


        // Enumerate through all listed devices getting their hardware 
        // ids and comparing it to the one we're looking for.

        DeviceInfoData.cbSize = sizeof(SP_DEVINFO_DATA);

        for ( i = 0 ;
              SetupDiEnumDeviceInfo( DeviceInfoSet, i, &DeviceInfoData ) ;
              i++ ) {
                
                Buffer[0] = '\0';
                Result = SetupDiGetDeviceRegistryProperty(
                        DeviceInfoSet,
                        &DeviceInfoData,
                        SPDRP_HARDWAREID,
                        &RegType,
                        (PBYTE)Buffer,
                        MAXSTR,
                        &BufferSize
                        );

                
                //  Hardware id is a multi-sz, so check each entry.
                
                Ptr = Buffer;
                while ( *Ptr && (Ptr < &Buffer[BufferSize]) ) {
                      
                        if ( strcmp( Ptr, HardwareId ) == 0 ) {
                                SetupDiDestroyDeviceInfoList( DeviceInfoSet );
                                return( 1 );
                                }

                        Ptr += ( lstrlen( Ptr ) + sizeof(char) );
                              
                        } // end while
                
                } // end for i, EnumDeviceInfo

        SetupDiDestroyDeviceInfoList( DeviceInfoSet );

        return( 0 );
}

int CheckForDriverStart( CHAR *DriverName ) {

        HKEY  ParmKey;
        DWORD Code, PortFound, RegType, BufferSize;
        CHAR  ParmKeyName[MAXSTR];

        PortFound = 0;

        // Open driver parameters registry key

        sprintf( ParmKeyName,
                 "System\\CurrentControlSet\\Services\\%s\\Parameters",
                 DriverName );

        Code = RegOpenKeyEx(
                            HKEY_LOCAL_MACHINE,  // already open  key
                            ParmKeyName,         // subkey to open
                            0,                   // reserved
                            KEY_READ,            // security access mask
                            &ParmKey             // handle to key
                           );


        if (Code == ERROR_SUCCESS) {

                // Query the PortFound value, 1 means the driver started ok

                BufferSize = sizeof(PortFound);
                Code = RegQueryValueEx(
                                       ParmKey,            // handle to key
                                       "PortFound",        // value name
                                       NULL,               // reserved
                                       &RegType,           // type buffer
                                       (LPBYTE)&PortFound, // data buffer
                                       &BufferSize         // size of buffer
                                      );
                
                if (Code != ERROR_SUCCESS) {
                        PortFound = 0;
                        Error( "Could not read PortFound value", NOT_FATAL );
                        }

                RegCloseKey( ParmKey );
                }

        else
                Error( "Failed to open driver parameters registry key", NOT_FATAL );
        
        return( PortFound );
}



// Note: Do not change the order of these values unless you also change
//       the array indexes used below for filling in the variable values

REGVALUE ServiceKeyValues[] = {
        "DisplayName",     REG_SZ,       0, "SrParXch378 AtoD Service Name",
        "ImagePath",       REG_SZ,       0, "System32\\DRIVERS\\srxch378.sys",
        "Group",           REG_SZ,       0, "Data acquisition",
        "DependOnService", REG_MULTI_SZ, 0, "Parport\0\0",
        "DependOnGroup",   REG_MULTI_SZ, 0, "Parallel arbitrator\0\0",
        "ErrorControl",    REG_DWORD,    SERVICE_ERROR_NORMAL,  "",
        "Start",           REG_DWORD,    SERVICE_DEMAND_START,  "",
        "Type",            REG_DWORD,    SERVICE_KERNEL_DRIVER, "",
        "",                REG_NONE,     0,                     "",
        };

REGVALUE ParametersKeyValues[] = {
        REG_PARM_DRIVERNAME,  REG_SZ,        0,     "SrParXch378",
        REG_PARM_PORTADDR,    REG_DWORD,     0x378, "",
        REG_PARM_PORTIRQ,     REG_DWORD,     7,     "",
        REG_PARM_PORTFOUND,   REG_DWORD,     0,     "",
        REGENVSTR_MODELNAME,  REG_SZ,        0,     "PAR4CH",
        "",                   REG_NONE,      0,     "",
        };


BOOL CreateServiceKey( CHAR *DriverName, int DriverNum,
                       int DriverAddr, int DriverIrq, CHAR *XchModelName ) {

        BOOL  Result1, Result2;
        DWORD Code, Disposition;
        HKEY  ServiceKey, ParametersKey;
        CHAR  ServiceKeyName[MAXSTR];
        CHAR  DisplayNameString[MAXSTR], SysFileNameString[MAXSTR];



        // Create service and service parameters registry subkeys

        sprintf( ServiceKeyName,
                 "System\\CurrentControlSet\\Services\\%s",
                 DriverName );

        Code = RegCreateKeyEx(
                       HKEY_LOCAL_MACHINE,      // already open key
                       ServiceKeyName,          // subkey to open
                       0,                       // reserved
                       NULL,                    // object class string
                       REG_OPTION_NON_VOLATILE, // special options
                       KEY_ALL_ACCESS,          // full access permission
                       NULL,                    // no inheritance       
                       &ServiceKey,             // handle to key
                       &Disposition             // new or existing
                      );

        if (Code != ERROR_SUCCESS) {
                Error( "Failed to create service key", NOT_FATAL );
                return( FALSE );
                }

        Code = RegCreateKeyEx(
                       ServiceKey,              // already open key
                       "Parameters",            // subkey to open
                       0,                       // reserved
                       NULL,                    // object class string
                       REG_OPTION_NON_VOLATILE, // special options
                       KEY_ALL_ACCESS,          // full access permission
                       NULL,                    // no inheritance       
                       &ParametersKey,          // handle to key
                       &Disposition             // new or existing
                      );

        if (Code != ERROR_SUCCESS) {
                Error( "Failed to create service parameters key", NOT_FATAL );
                RegCloseKey( ServiceKey );
                return( FALSE );
                }




        
        // Update static service and parameter key entries with their
        // variable components.  Note that the array indexes must match
        // the correct position in the static structure defined above.

        sprintf( DisplayNameString, "%s AtoD Service Name", DriverName );
        sprintf( SysFileNameString, "System32\\DRIVERS\\srxch%d.sys", DriverNum );

        
        ServiceKeyValues[0].ValueSz = DisplayNameString;
        ServiceKeyValues[1].ValueSz = SysFileNameString;
        
        ParametersKeyValues[0].ValueSz = DriverName;
        ParametersKeyValues[1].ValueDw = DriverAddr;
        ParametersKeyValues[2].ValueDw = DriverIrq;
        ParametersKeyValues[3].ValueDw = 0; // No PortFound 
        ParametersKeyValues[4].ValueSz = XchModelName;



        // Fill values for service and parameters keys

        Result1 = FillRegKeyWithValues(    ServiceKey,    ServiceKeyValues );
        Result2 = FillRegKeyWithValues( ParametersKey, ParametersKeyValues );


        RegCloseKey( ParametersKey );
        RegCloseKey( ServiceKey );
        
        return( Result1 && Result2 );
}

BOOL FillRegKeyWithValues( HKEY Key, REGVALUE Values[] ) {
        
        REGVALUE   *ValueEntry;
        BOOL        Result;
        DWORD       Code, DataSize;
        CONST BYTE *DataValue;


        // Initialize variables
        
        Result = TRUE;
        ValueEntry = Values;

        
        // Loop for all entries
        
        while ( ValueEntry->Type != REG_NONE ) {

                
                // Prepare data size and value 

                if (ValueEntry->Type == REG_DWORD) {
                        DataValue = (CONST BYTE *)&ValueEntry->ValueDw;
                        DataSize  = sizeof( DWORD );
                        }
                else if (ValueEntry->Type == REG_SZ ||
                         ValueEntry->Type == REG_EXPAND_SZ) {
                        DataValue = (CONST BYTE *)ValueEntry->ValueSz;
                        DataSize  = strlen(ValueEntry->ValueSz)+1;
                        }
                else if (ValueEntry->Type == REG_MULTI_SZ ) {
                        DataValue = (CONST BYTE *)ValueEntry->ValueSz;
                        DataSize  = strlen(ValueEntry->ValueSz)+3;
                        }
                else {
                        DataValue = (CONST BYTE *)&ValueEntry->ValueDw;
                        DataSize  = sizeof( DWORD );
                        printf( "Unsupported REG type %d\n", ValueEntry->Type );
                        Result = FALSE;
                        }


                // Fill registry entry
                        
                Code = RegSetValueEx(
                                     Key,              // handle to key
                                     ValueEntry->Name, // value name
                                     0,                // reserved
                                     ValueEntry->Type, // value type
                                     DataValue,        // value data
                                     DataSize          // size of value data
                            );
                        
                if (Code != ERROR_SUCCESS) {
                        printf( "\n\nERROR: Failed to write registry value %s\n",
                                  ValueEntry->Name);
                        printf( "   with data %u or %s\n",
                                  ValueEntry->ValueDw, ValueEntry->ValueSz );
                        Result = FALSE;
                        }



                // Move to next value
                
                ValueEntry++;
                
                } // end while loop
        

        return( Result );
}

void Usage( char *msg ) {

        if ( !msg )
                printf( "\nERROR: Command line usage.\n\n" );
        else
                printf( "\nERROR: Command line usage (%s).\n\n", msg );
        
        PrintMsg( UsageMsg );

        exit( 1 );
}

void PrintMsg( char *MsgArray[] ) {

        int i;

        for ( i = 0 ; MsgArray[i] ; i++ )
                printf( "%s\n", MsgArray[i] );

}

void Error( char *Msg, int Fatal ) {

        DWORD  Err;
        LPVOID lpMessageBuffer;


        // Display local message, error number, and system message

        if (Fatal)
                printf( "\n\nERROR: " );

        Err = GetLastError();

        if ( Err == 0 ) {
                printf( "%s.\n", Msg );
                }
        
        else {
                FormatMessage(
                              FORMAT_MESSAGE_ALLOCATE_BUFFER |
                              FORMAT_MESSAGE_FROM_SYSTEM,
                              NULL,
                              Err,
                              MAKELANGID(LANG_NEUTRAL,SUBLANG_DEFAULT),
                              (LPTSTR) &lpMessageBuffer,
                              0,
                              NULL
                             );

                printf( "%s.  Error #%d\n%s\n", Msg, Err, lpMessageBuffer );

                LocalFree( lpMessageBuffer );
                }

        
        // Give additional explanation for common errors
        
        if (Err == ERROR_NO_MORE_ITEMS)  // #259
                PrintMsg( ErrorNoMoreItemsMsg );


        // Exit if fatal.
        
        if ( Fatal ) {
                printf( "Please run rmdriver to remove failed install.\n" );
                exit( 1 );
                }

}
