/* FILE: rmdriver.c              Copyright (c), Symmetric Research, 2003-2007
 *
 * Program to remove the Win2K PARxCH kernel mode driver.
 *
 * This is done by getting the GUID of device from the original
 * symres.inf file, preparing a driver information list with all
 * matching devices, finding the name of the copied oem#.inf file
 * from the service key, removing the device, removing the class 
 * key and oem#.inf file if they are not in use by another device, 
 * removing the copied sys file, and deleting the service key from
 * the registry.
 *
 */



// Needed in setupapi for Win9x and WinNT compatibility (Win2K could use V2)
#define  USE_SP_DRVINFO_DATA_V1 1

#include <windows.h>
#include <setupapi.h>   // for SetupDi functions
#include <stdio.h>
#include <stdlib.h>


// Defines and typedefs

#define MAXSTR    256

#define NOT_FATAL 0
#define FATAL     1

#define INF_PATH  "\\symres.inf"

#ifndef SUOI_FORCEDELETE          // This is defined in Win XP and later
#define SUOI_FORCEDELETE 0x0001
#endif



// Usage message arrays:

char *UsageMsg[] = {

  "                                                                      ",
  "ERROR: Command line usage                                             ",
  "                                                                      ",
  "USAGE: rmdriver ParXchDriverName                                      ",
  "                                                                      ",
  "This utility removes a PARxCH driver.  ParXchDriverName must be a     ",
  "name of the form SrParXch# where # is a small integer like 0, 1, etc. ",
  "To see what drivers are currently installed, run showdriver, run the  ",
  "Windows Device Manager or under Linux run /sbin/lsmod.                ",
  "                                                                      ",
  "RMDRIVER EXAMPLE: > rmdriver SrParXch0                                ",
  "                                                                      ",
  NULL
  };


char DriverName[MAXSTR], MsgBuffer[MAXSTR];



// Protos for helper functions

void FormatGUID( GUID *Guid, CHAR *GuidString );
BOOL UninstallSysFile( CHAR *ServiceName, CHAR *SystemRoot, CHAR *SysFileName );
BOOL DeleteServiceKey( CHAR *ServiceName );
BOOL DeleteEventLogKey( CHAR *ServiceName );
BOOL OsUninstallOemInf( CHAR *InfFileName, DWORD Flags, PVOID Reserved );
void Usage( void );
void PrintMsg( char *MsgArray[] );
void Error( char *Msg, int Fatal );



// Main:

void main( int argc, char **argv ) {

        int    i;
        LONG   Code;
        BOOL   Ok, FoundDevice, Normal;
        DWORD  Result, BufferSize, RegType, NumSubKeys, Flags, HardwareIdSize;
        CHAR   Buffer[MAXSTR], ClassName[MAXSTR], HardwareId[MAXSTR];
        CHAR   SysRoot[MAX_PATH], SysFileName[MAX_PATH], InfFileName[MAX_PATH];
        CHAR   ClassGuidInfStr[40], *Ptr;
        GUID   ClassGuidInf;
        REGSAM DesiredAccess;
        HKEY   DevRegKey, ClassRegKey, ParentRegKey;
        HDEVINFO DeviceInfoSet;
        SP_DEVINFO_DATA DeviceInfoData;


        // Check secret command line argument /s.  This is used to give
        // silent operation only when rmdriver is called from indriver.
        
        Normal = TRUE;
        if ( argc == 3  &&  strcmp( argv[2], "/s" ) == 0 )
                Normal = FALSE;



        // Banner message

        if (Normal)
                printf( "PARxCH Kernel Mode Driver Removal Program : Rev {01/08/07}\n");

        
        // Process command line arguments for driver name

        if ( Normal  &&  argc != 2 )
                Usage();


        strcpy( DriverName, argv[1] );


        if (Normal)
                printf( "\nATTEMPTING TO REMOVE DRIVER %s ...\n", DriverName );




        // Get SystemRoot environment string value

        Result = ExpandEnvironmentStrings( "%SYSTEMROOT%", SysRoot, MAX_PATH );
        if (Result == 0  ||  Result > MAX_PATH)
                Error( "Failed to properly expand system root path", NOT_FATAL );





        // Get Class GUID from inf file assuming that this program
        // is being run from the sr\parxch\driver directory and that
        // the inf file is in the same directory.
        
        strncpy( Buffer, ".", MAXSTR-1 );  Buffer[MAXSTR-1] = '\0';
        InfFileName[0] = '\0';
        ClassName[0]   = '\0';
        
        GetCurrentDirectory( sizeof(Buffer), Buffer );

        if ( (strlen(Buffer) + strlen(INF_PATH)) > MAXSTR-1)
                Error( ".inf file relative path name is too long", FATAL );
        else
                sprintf( InfFileName, "%s%s", Buffer, INF_PATH );
        
        Ok = SetupDiGetINFClass(
                                InfFileName,
                                &ClassGuidInf,
                                ClassName,
                                sizeof(ClassName),
                                0
                               );
        if (!Ok)
                Error( "Failed to read GUID from inf file", FATAL );
        else
                if (Normal)
                        printf( "   ... read GUID from inf file %s\n", InfFileName );

        FormatGUID( &ClassGuidInf, ClassGuidInfStr );




        // Create the HardwareID.
        // It is a multi-sz string, so it must be double null terminated.

        sprintf( HardwareId, "Root\\%s&%s", ClassName, DriverName );
        HardwareIdSize = (lstrlen(HardwareId)+1+1)*sizeof(CHAR);
        HardwareId[HardwareIdSize-1] = '\0';
        HardwareId[HardwareIdSize]   = '\0';
        HardwareId[HardwareIdSize+1] = '\0';



        

        // Create a Device Information Set with all present devices
        // having specified class guid.

        DeviceInfoSet = SetupDiGetClassDevs( &ClassGuidInf, 0, 0, DIGCF_PRESENT);
        
        if (DeviceInfoSet == INVALID_HANDLE_VALUE)
                Error( "Failed to create a device list", FATAL );
    

        // Enumerate through all listed devices getting their hardware keys
        // and comparing it to the one we're looking for.
        
        FoundDevice = FALSE;
        DeviceInfoData.cbSize = sizeof(SP_DEVINFO_DATA);

        for ( i = 0 ;
              SetupDiEnumDeviceInfo( DeviceInfoSet, i, &DeviceInfoData ) ;
              i++ ) {
                
                Buffer[0] = '\0';
                Ok = SetupDiGetDeviceRegistryProperty(
                        DeviceInfoSet,
                        &DeviceInfoData,
                        SPDRP_HARDWAREID,
                        &RegType,
                        (PBYTE)Buffer,
                        MAXSTR,
                        &BufferSize
                        );

                if (!Ok) {
                        printf( "Could not read hardware id for device %d\n", i);
                        Buffer[0] = '\0';
                        }

                //  Hardware id is a multi-sz, so check each entry.
                
                Ptr = Buffer;
                while ( *Ptr && (Ptr < &Buffer[BufferSize]) ) {
                      
                        if ( strcmp( Ptr, HardwareId ) == 0 ) {
                                FoundDevice = TRUE;
                                break; // This only gets us out of the while
                                }

                        Ptr += ( lstrlen( Ptr ) + sizeof(char) );
                              
                        } // end while
                
                
                if ( FoundDevice )
                        break; // This gets us the rest of the way out

                } // end for i, EnumDeviceInfo
              

        InfFileName[0] = '\0';
        DesiredAccess = STANDARD_RIGHTS_READ | GENERIC_READ;

        
        // If device not installed, exit

        if ( !FoundDevice ) {
                SetupDiDestroyDeviceInfoList( DeviceInfoSet );
                printf( "\n\nERROR: Device %s not installed.  ", DriverName );
                printf( "Use indriver to install it.\n");
                exit( 1 );
                }


        // else device exists, get inf name and remove service
        
        else {

                // Get the oem#.inf name
              
                DevRegKey = SetupDiOpenDevRegKey(
                                                 DeviceInfoSet,
                                                 &DeviceInfoData,
                                                 DICS_FLAG_GLOBAL,
                                                 0,
                                                 DIREG_DRV,
                                                 DesiredAccess
                                                );

                if (DevRegKey == INVALID_HANDLE_VALUE) {
                        Error( "Failed to open device reg key", NOT_FATAL );
                        }
                else {
                        BufferSize = sizeof(Buffer);
                        Code = RegQueryValueEx(
                                               DevRegKey,
                                               "InfPath",
                                               NULL,
                                               &RegType,
                                               Buffer,
                                               &BufferSize
                                              );

                        if (Code != ERROR_SUCCESS)
                                Error( "Could not read inf path", NOT_FATAL );
                        else {
                                sprintf(
                                        InfFileName,
                                        "%s\\inf\\%s",
                                        SysRoot,
                                        Buffer
                                     );
                                }

                        RegCloseKey( DevRegKey );
                        if (Code != ERROR_SUCCESS)
                                Error( "Could not close device key", NOT_FATAL );

                        } // end if DevRegKey


                // Remove the device

                Ok = SetupDiCallClassInstaller(
                                               DIF_REMOVE,
                                               DeviceInfoSet,
                                               &DeviceInfoData
                                              );
                if ( !Ok ) {
                        SetupDiDestroyDeviceInfoList( DeviceInfoSet );
                        Error( "Failed to remove device", FATAL );
                        }
                else
                        if (Normal)
                                printf( "   ... device %s removed\n", DriverName );


                } // End found device
      
        SetupDiDestroyDeviceInfoList( DeviceInfoSet );




        // See if the Class key has any other devices

        NumSubKeys = 0;
        ClassRegKey = SetupDiOpenClassRegKeyEx(
                                               &ClassGuidInf,
                                               DesiredAccess,
                                               DIOCR_INSTALLER,
                                               NULL,
                                               NULL
                                              );

        if (ClassRegKey == INVALID_HANDLE_VALUE)
                Error( "Failed to open class reg key", NOT_FATAL );

        else {
                Code = RegQueryInfoKey(
                                       ClassRegKey,
                                       NULL,            // ClassBuff,
                                       NULL,            // ClassBuffSize,
                                       NULL,            // Reserved
                                       &NumSubKeys,
                                       NULL,            // MaxSubKeyLen,
                                       NULL,            // MaxClassStrLen,
                                       NULL,            // NumValues,
                                       NULL,            // MaxValueNameLen,
                                       NULL,            // MaxValueLen,
                                       NULL,            // SecurityDescLen,
                                       NULL             // LastWriteTime
                                      );

                if (Code != ERROR_SUCCESS)
                        Error( "Failed to get class key info", NOT_FATAL );
                
                RegCloseKey( ClassRegKey );
                
                }




        // No other device users, so its ok to delete class key and oem#.inf

        if (NumSubKeys != 0) {
                if (Normal) {
                        printf( "   ... copied inf file %s is in use\n", InfFileName ); 
                        Error( "       by other devices and not deleted", NOT_FATAL);
                        }
                }

        else {

                ParentRegKey = SetupDiOpenClassRegKeyEx(
                                                        NULL,
                                                        DesiredAccess,
                                                        DIOCR_INSTALLER,
                                                        NULL,
                                                        NULL
                                                        );

                if (ParentRegKey == INVALID_HANDLE_VALUE)
                        Error( "Failed to open class parent reg key", NOT_FATAL );

                else {
                        Code = RegDeleteKey( ParentRegKey, ClassGuidInfStr );
                        if (Code != ERROR_SUCCESS) {
                                sprintf(
                                         MsgBuffer,
                                         "Failed to delete Class key %s",
                                         ClassGuidInfStr
                                       );
                                Error( MsgBuffer, NOT_FATAL );
                                }

                        Code = RegCloseKey( ParentRegKey );
                        if (Code != ERROR_SUCCESS)
                                Error( "Failed to close parent key", NOT_FATAL );
                        
                        } // else opened class parent key



                // Delete the oem#.inf file name we found earlier

                Flags = SUOI_FORCEDELETE; // don't check for other oem.inf users
                Ok = OsUninstallOemInf( InfFileName, Flags, NULL );
                if (!Ok) {
                        sprintf( MsgBuffer,
                                 "Failed to delete file %s or its pnf",
                                 InfFileName );
                        Error( MsgBuffer, NOT_FATAL );
                        }
                else
                        if (Normal)
                                printf( "   ... copied inf file %s deleted\n",
                                          InfFileName ); 


                } // else num subkeys = 0




        // Remove old sys file

        Ok = UninstallSysFile( DriverName, SysRoot, SysFileName );
        if (!Ok) {
                sprintf( MsgBuffer, "Failed to delete sys file %s", SysFileName );
                Error( MsgBuffer, NOT_FATAL );
                }
        else
                if (Normal)
                        printf( "   ... copied sys file %s deleted\n", SysFileName ); 



        // Remove service key

        Ok = DeleteServiceKey( DriverName );
        if (!Ok) 
                Error( "Failed to delete service key", NOT_FATAL );
        else
                if (Normal)
                        printf( "   ... service key %s deleted from registry\n",
                                  DriverName );


        // Remove event log key

        Ok = DeleteEventLogKey( DriverName );
        if (!Ok) 
                Error( "Failed to delete event log key", NOT_FATAL );
        else
                if (Normal)
                        printf( "   ... event log key System\\%s deleted from registry\n",
                                  DriverName );


        if (Normal)
                printf( "\nSUCCESS: %s driver removed.\n", DriverName );
}

void FormatGUID( GUID *Guid, CHAR *GuidString ) {

        if (GuidString == NULL)
                return;
        
        sprintf( GuidString,
                 "{%08lX-%04X-%04X-%02X%02X-%02X%02X%02X%02X%02X%02X}",
                 Guid->Data1, Guid->Data2, Guid->Data3,
                 Guid->Data4[0], Guid->Data4[1],
                 Guid->Data4[2], Guid->Data4[3],
                 Guid->Data4[4], Guid->Data4[5],
                 Guid->Data4[6], Guid->Data4[7] );
}

BOOL UninstallSysFile( CHAR *DriverName, CHAR *SystemRoot, CHAR *SysFileName ) {

        BOOL  Ok;
        DWORD ValueType, ValueSize, Code;
        CHAR  ServiceKeyName[MAXSTR], ImagePath[MAX_PATH];
        HKEY  ServiceKey;


        SysFileName[0] = '\0';


        // Open service key 

        sprintf( ServiceKeyName,
                 "System\\CurrentControlSet\\Services\\%s",
                 DriverName );

        Code = RegOpenKeyEx(
                            HKEY_LOCAL_MACHINE,  // already open parent key
                            ServiceKeyName,      // subkey to open
                            0,                   // reserved
                            KEY_ALL_ACCESS,      // full access permission
                            &ServiceKey          // handle to key
                           );

        if (Code != ERROR_SUCCESS) {
                Error( "Failed to open service key", NOT_FATAL );
                return( FALSE );
                }


        // Get sys file name

        ValueSize = MAX_PATH;
        Code = RegQueryValueEx(
                               ServiceKey,       // use this key
                               "ImagePath",      // value name
                               0,                // reserved
                               &ValueType,       // value data type
                               ImagePath,        // ptr to value contents
                               &ValueSize        // size of buffer
                              );

        RegCloseKey( ServiceKey );
        
        if (Code != ERROR_SUCCESS) {
                Error( "Failed to read service key ImagePath", NOT_FATAL );
                return( FALSE );
                }



        // Build complete name and delete the sys file 

        sprintf( SysFileName, "%s\\%s", SystemRoot, ImagePath );

        Ok = DeleteFile( SysFileName );

        return( Ok );
}


BOOL DeleteServiceKey( CHAR *DriverName ) {

        int   i;
        BOOL  Ok;
        DWORD Code, SubKeyLen, NumSubKeys, MaxSubKeyLen;
        CHAR  SubKeyName[MAXSTR], ParentKeyName[MAXSTR] ;
        HKEY  ServiceKey, ServiceParentKey;


        Ok = TRUE;
        

        // Get parent service key

        strncpy( ParentKeyName, "System\\CurrentControlSet\\Services", MAXSTR-1 );
        ParentKeyName[MAXSTR-1] = '\0';
        
        Code = RegOpenKeyEx(
                            HKEY_LOCAL_MACHINE,    // already open parent key
                            ParentKeyName,         // subkey to open
                            0,                     // reserved
                            KEY_ALL_ACCESS,        // full access permission
                            &ServiceParentKey      // handle to key
                           );

        if (Code != ERROR_SUCCESS) {
                Error( "Failed to open parent service key", NOT_FATAL );
                return( FALSE );
                }


        
        
        // Open service key and get number of subkeys

        Code = RegOpenKeyEx(
                            ServiceParentKey,      // already open parent key
                            DriverName,            // subkey to open
                            0,                     // reserved
                            KEY_ALL_ACCESS,        // full access permission
                            &ServiceKey            // handle to key
                           );

        if (Code != ERROR_SUCCESS) {
                RegCloseKey( ServiceParentKey );
                Error( "Failed to open service key", NOT_FATAL );
                return( FALSE );
                }


        Code = RegQueryInfoKey(
                               ServiceKey,    // handle to key
                               NULL,          // class buffer
                               NULL,          // size of class buffer
                               NULL,          // reserved
                               &NumSubKeys,   // number of subkeys
                               &MaxSubKeyLen, // longest subkey name
                               NULL,          // longest class string
                               NULL,          // number of value entries
                               NULL,          // longest value name
                               NULL,          // longest value data
                               NULL,          // descriptor length
                               NULL           // last write time
                              );

        if (Code != ERROR_SUCCESS) {
                NumSubKeys = 0;
                Error( "Failed to get service key info", NOT_FATAL );
                Ok = FALSE;
                }

        if (MAXSTR < MaxSubKeyLen) {
                sprintf(
                        MsgBuffer,
                        "SubKey too long.  Please increase buffer to %d",
                        MaxSubKeyLen
                       );
                Error( MsgBuffer, NOT_FATAL );
                Ok = FALSE;
                }
        


        // Enumerate and delete all service subkeys

        for ( i = NumSubKeys-1 ; i >= 0 ; i-- ) {

                SubKeyLen = MAXSTR;
                Code = RegEnumKeyEx(
                                      ServiceKey,     // handle to key
                                      i,              // subkey index
                                      SubKeyName,     // subkey name
                                      &SubKeyLen,     // size of subkey buffer
                                      NULL,           // reserved
                                      NULL,           // class string buffer
                                      NULL,           // size of class string buffer
                                      NULL            // last write time
                                     );
                
                if (Code == ERROR_NO_MORE_ITEMS) {
                        Error( "Finished enumerating service key too soon", NOT_FATAL );
                        break;
                        }
        
                Code = RegDeleteKey( ServiceKey, SubKeyName );

                if (Code != ERROR_SUCCESS) {
                        sprintf(
                                 MsgBuffer,
                                 "Failed to delete subkey key %s",
                                 SubKeyName
                               );
                        Error( MsgBuffer, NOT_FATAL );
                        Ok = FALSE;
                        }
                
                } // end for i



        // Close service key, delete it, then close parent key
        
        Code = RegCloseKey( ServiceKey );
        if (Code != ERROR_SUCCESS) {
                Error( "Failed to close service key", NOT_FATAL );
                Ok = FALSE;
                }

        Code = RegDeleteKey( ServiceParentKey, DriverName );
        if (Code != ERROR_SUCCESS) {
                Error( "Failed to delete service key", NOT_FATAL );
                Ok = FALSE;
                }

        Code = RegCloseKey( ServiceParentKey );
        if (Code != ERROR_SUCCESS) {
                Error( "Failed to close parent service key", NOT_FATAL );
                Ok = FALSE;
                }


        return( Ok );
}


BOOL DeleteEventLogKey( CHAR *DriverName ) {

        BOOL  Ok;
        DWORD Code;
        CHAR  ParentKeyName[MAXSTR];
        HKEY  EventLogParentKey;


        Ok = TRUE;
        

        // Get parent event log key

        strncpy( ParentKeyName,
                "System\\CurrentControlSet\\Services\\EventLog\\System",
                 MAXSTR-1 ); ParentKeyName[MAXSTR-1] = '\0';

        Code = RegOpenKeyEx(
                            HKEY_LOCAL_MACHINE,    // already open parent key
                            ParentKeyName,         // subkey to open
                            0,                     // reserved
                            KEY_ALL_ACCESS,        // full access permission
                            &EventLogParentKey     // handle to key
                           );

        if (Code != ERROR_SUCCESS) {
                Error( "Failed to open parent event log key", NOT_FATAL );
                return( FALSE );
                }

        

        // Delete event log key
        
        Code = RegDeleteKey( EventLogParentKey, DriverName );
        if (Code != ERROR_SUCCESS) {
                Error( "Failed to delete driver event log key", NOT_FATAL );
                Ok = FALSE;
                }



        // Close parent key

        Code = RegCloseKey( EventLogParentKey );
        if (Code != ERROR_SUCCESS) {
                Error( "Failed to close parent event log key", NOT_FATAL );
                Ok = FALSE;
                }


        return( Ok );
}


BOOL OsUninstallOemInf( CHAR *InfFileName, DWORD Flags, PVOID Reserved ) {

#if defined( SROS_WINXP )
        return( SetupUninstallOEMInf( (PCWSTR)InfFileName, 0, NULL ) );
#else
        
        
        BOOL Ok;
        int  NameLen;
        char PnfFileName[MAX_PATH];
        
        // Error check name string
        
        if ( !InfFileName )
                return( FALSE );

        
        // Error check inf name length

        NameLen = strlen( InfFileName );

        if ( NameLen == 0  ||  NameLen > MAX_PATH-1 )
                return( FALSE );


        // WCT: Should check for other devices unless force flag is set

        if ( !(Flags & SUOI_FORCEDELETE) ) {
                printf( "No check for other devices using %s\n", InfFileName );
                }
        

        
        // Delete the Ini file
        
        Ok = DeleteFile( InfFileName );

        
        // Compute name of related PNF file and delete it.

        strcpy( PnfFileName, InfFileName ); // length checked above
        PnfFileName[NameLen-3] = 'p';
        
        Ok &= DeleteFile( PnfFileName );



        // Note: The XP version also deletes related catalog files ...

        
        return( Ok );

#endif        
}

void Usage( void ) {

        PrintMsg( UsageMsg );
                
        exit( 1 );
}

void PrintMsg( char *MsgArray[] ) {

        int i;

        for ( i = 0 ; MsgArray[i] ; i++ )
                printf( "%s\n", MsgArray[i] );
}

void Error( char *Msg, int Fatal ) {

        DWORD Err;
        LPVOID lpMessageBuffer;
        
        // Display local message, error number, and system message
        
        if (Fatal)
                printf( "\n\nERROR: " );

        Err = GetLastError();

        if (Err == 0) {
                printf( "%s.\n", Msg );
                }
        else {
                FormatMessage(
                              FORMAT_MESSAGE_ALLOCATE_BUFFER |
                              FORMAT_MESSAGE_FROM_SYSTEM,
                              NULL,
                              Err,
                              MAKELANGID(LANG_NEUTRAL,SUBLANG_DEFAULT),
                              (LPTSTR) &lpMessageBuffer,
                              0,
                              NULL
                             );

                printf( "%s.  Error #%d\n%s\n", Msg, Err, lpMessageBuffer );

                LocalFree( lpMessageBuffer );
                }



        // Give additional explanation for common errors
        
//        if (Err == ERROR_INVALID_SERVICE_CONTROL)  // #1052
//                PrintMsg( ErrorInvalidServiceControlMsg );
//        else
//                printf( "Sorry, no extra help available for this error.\n" );



        // Clean up and exit if fatal.

        if ( Fatal )
                exit( 1 );
}
