// FILE: scope.cpp              Copyright (c), Symmetric Research, 2003-2008
//
// Source file for generic portion of the Scope program includes
//
//      CScopeApp       should be virtual (ie must declare derived class)
//      CScopeWindow    should be virtual (ie must declare derived class)
//      CScopeDialog    should be virtual (ie must declare derived class)
//      CTriggerDlg
//      COutputDlg
//      CDisplayDlg
//      CGpsDlg
//      CContentsDlg
//      CTopicDlg
//      CSrParser
//
//

#include <direct.h>     // needed for _getcwd (GET Current Working Directory)

#include "scope.h"





// Global pointer to application object

CScopeApp *pAPP;



//***********************************************************************
//
// CLASS CScopeApp
//
// This class, derived from CWinApp, contains the generic application
// portion of the Scope data acquisition program.  You should not
// declare objects of this type.  Instead, declare application objects
// of one of the derived classes, CA64App, C24bApp or CParXchApp.  These
// are for use with the DSPA64/DSPHLF, PAR24B and PARxCH acquisition
// systems respectively.  The application portion contains code dealing
// with acquisition, files, and ini.  CScopeWindow takes care of the
// graphics and main window functions.
//
//

BEGIN_MESSAGE_MAP(CScopeApp, CWinApp)
        //{{AFX_MSG_MAP(CScopeApp)
        //}}AFX_MSG_MAP
END_MESSAGE_MAP()



BOOL CScopeApp::InitInstance() {

        TRACE0("CScopeApp::InitInstance\n");

        RECT screenrect;


        // Call just in case this isn't a no-op

        CWinApp::InitInstance();


        // Get full screen size (could also have used GetSystemMetrics)
        //
        //     int screenx = GetSystemMetrics(SM_CXSCREEN);
        //     int screeny = GetSystemMetrics(SM_CYSCREEN);
        
        SystemParametersInfo(SPI_GETWORKAREA,0,&screenrect,0);

        DefaultScreenWidth  = screenrect.right  - screenrect.left;
        DefaultScreenHeight = screenrect.bottom - screenrect.top;




        // Set initial data

        InitAppData();



        // Allocate scan and offset arrays

        Ini.DoAutoRun = FALSE;     // Pre-set this in case error occurs
        if (!AllocateScanArrays())
                return(FALSE);



        // Initialize ini file data values

        IniSetDefaults();





        // Check command line for ini filename or empty

        Ini.IniFile = m_lpCmdLine;
        Ini.IniFile.TrimLeft();
        Ini.IniFile.TrimRight();

        if ( Ini.IniFile.Compare( "/?" ) == 0 ||
             Ini.IniFile.Compare( "?" )  == 0 ) {
                ErrString  = "Scope usage message.";
                TmpString  = "\nScope can be started with zero or one arguments.\n";
                TmpString += "If one argument is provided, it must be the name\n";
                TmpString += "of a valid initialization file.  If no arguments\n";
                TmpString += "are given, the program will use default values\n";
                TmpString += "for the option settings.\n";
                LogMessage( MSGLEVEL_ERR, MSGDLG_SHOW, ERROR_AUTO_FATAL, &TmpString );
                ErrCode = 1;
                return( FALSE );
                }




        // If ini file not specified, verify defaults and save them

        IniFileSetShortName();

        if ( Ini.IniFile.IsEmpty() ) {
                IniVerifyData();
                IniSaveCurrent();
                }

        // Else read and verify from file, printing message if error

        else if ( !IniInputFromFile() ) {
                ErrString  = "Input from .ini file failed.";
                TmpString  = "\nProgram will start with defaults.\n";
                TmpString += "Use Options>Read from ini file... ";
                TmpString += "to specify another ini file.";
                LogMessage( MSGLEVEL_WARN, MSGDLG_SHOW, ERROR_AUTO_FATAL, &TmpString );
                }



        // Compute dependent values

        ComputeActiveChannels();
        if (!ComputeBufferSizes())
                return( FALSE );
        UpdatePlotTraceInfo();



        // Check for multiple instances

        if ( !MultipleInstanceCondition() )
                return( FALSE );








        // Create main window and check validity.
        //
        // If invalid, delete it and NULL out the pointer so
        // DestroyWindow is not called when InitInstance fails.
        // The window class is already gone because of failing
        // during its construction phase.


        m_pMainWnd = (CWnd*)NewScopeWindow();

        if ( ! ::IsWindow(m_pMainWnd->m_hWnd) ) {

                delete m_pMainWnd;
                m_pMainWnd = NULL;

                return( FALSE );
                }




        // Show main window

        m_pMainWnd->ShowWindow(m_nCmdShow);
        m_pMainWnd->UpdateWindow();



        // Post run command message if AutoRun mode requested

        if (Ini.DoAutoRun)
                m_pMainWnd->PostMessage( WM_COMMAND, IDM_PROGRAM_RUN, NULL );

        return( TRUE );

}


int CScopeApp::ExitInstance() {

        TRACE0("CScopeApp::ExitInstance\n");


        // Clean up any existing common arrays

        if (pDemuxArray)
                 { free(pDemuxArray); pDemuxArray = NULL; }

        if (pGpsPpsArray)
                 { free(pGpsPpsArray); pGpsPpsArray = NULL; }

        if (pGpsSerialArray)
                 { free(pGpsSerialArray); pGpsSerialArray = NULL; }


        FreeScanArrays();


        return( CWinApp::ExitInstance() );
}

CScopeWindow* CScopeApp::NewScopeWindow() {

        TRACE0("CScopeApp::NewScopeWindow\n");

        return( NULL );
}

int CScopeApp::InitAppData() {

        TRACE0("CScopeApp::InitAppData\n");

        SYSTEMTIME systime;


        // Set general global application pointer
        // Although declared as base class, actually points to
        // the global object which is of the derived class

        pAPP = this;



        // Initialize common global constants (not stored in ini file)

        ErrCode          = 0;
        FileError        = 0;
        CurrBlock        = 0;
        CurrTriggerFile  = 0;
        TriggerState     = TRIGSTATE_WAITING;
        Iplotbuff        = 0;
        LastNbufferbytes = 0;
        LastNgpspts      = 0;

        IsRunning        = FALSE;
        IsAllocated      = FALSE;
        ResetPaint       = TRUE;
        ResetDC          = TRUE;
        IsTriggered      = FALSE;

        pTMPoffset       = NULL;
        pDCoffset        = NULL;
        pDemuxArray      = NULL;

        hWndPrev         = NULL;

        ErrString        = "Error:";
        szTemp[0]        = '\0';

        DigitalIndex     = SRDAT_INDEXCHANNEL_NONE;
        GpsMarkIndex     = SRDAT_INDEXCHANNEL_NONE;
        CounterIndex     = SRDAT_INDEXCHANNEL_NONE;

        pGpsPpsArray     = NULL;
        pGpsSerialArray  = NULL;
        GpsHandle        = BAD_DEVHANDLE;
        GpsNumSat        = -1;
        

        TmpFileName.Empty();
        FinalFileName.Empty();
        ShortFileName.Empty();

        GetSystemTime( &systime );    // in UTC
        StartTime         = systime;  // "local" now has UTC correction
        StartTimeUsec     = systime.wMilliseconds * 1000;
        AcquireTime       = StartTime;
        AcquireTimeUsec   = StartTimeUsec;
        ReadyTime         = StartTime;
        SingleBufferTime  = 0.0L;
        ElapsedBufferTime = 0.0L;
        NpointsFromStart  = 0;
        TimeSource        = SRDAT_SOURCE_NONE;




        // Get current working directory
        // This is curr dir from Dos prompt, and exe dir from Win Run...

        if (_getcwd( szTemp, MAXSTR ) != NULL)
                CurrDir = szTemp;
        else
                CurrDir = ".";


        // NOTE: Many of the other variables are assumed to be
        //       set from the derived class InitAppData call
        //       which should complete after having called this
        //       parent class version


        return( 1 );
}

int CScopeApp::MultipleInstanceCondition() {

        TRACE0("CScopeApp::MultipleInstanceCondition\n");

        HANDLE  hSem;
        CString SemName;


        // If multiple instances are forbidden, check for them

        if (Ini.DoForbidMultiple) {

                // Create or open a named semaphore.

                SemName.Format( "SemaphoreFor%s", (LPCTSTR)AppTitle );

                hSem = CreateSemaphore( NULL, 0, 1, SemName );


                // If existing semaphore was opened, close handle.

                if ((hSem != NULL) && (GetLastError() == ERROR_ALREADY_EXISTS)) {

                        CloseHandle(hSem);

                        ErrString = "Another instance of this program is already running.\n";
                        LogMessage( MSGLEVEL_MSG, MSGDLG_SHOW, ERROR_NOT_FATAL, NULL );

                        return( 0 );

                        }

                }


        // Either multiples are allowed, or there is only one

        return( 1 );
}


int CScopeApp::ComputeBufferSizes( void ) {

        TRACE0("CScopeApp::ComputeBufferSizes\n");

        Nbufferpts     = Ini.Nsamples * Ini.Nchannels;
        Nbufferbytes   = Nbufferpts   * DataPtSize;

        NsamplesPlot   = Nplotbuffers * Ini.Nsamples;
        NbufferptsPlot = NsamplesPlot * NchannelsPlot;


        // Compute some GPS parameters
        //
        // GPS data comes once per second.  So, to determine the number
        // of points needed in the GPS data arrays, we computed the
        // length of time for one buffer and add 2 for good measure.

        Ngpspts        = (int)(Ini.Nsamples / Ini.AcqSPS) + 2;

#ifdef GPS_AVAILABLE

        // Don't ask for more data than MAX_PARGPS_STORAGE since that is
        // the total amount that can be stored in the PARGPS device driver
        // software FIFO without losing data.

        if (Ini.GpsEnable && Ngpspts > MAX_PARGPS_STORAGE) {

                ErrString = "Low sample rate overflows MAX_PARGPS_STORAGE for the PARGPS driver software FIFO.";
                LogMessage( MSGLEVEL_ERR, MSGDLG_SHOW, ERROR_AUTO_FATAL, NULL );
                return( 0 );
                }
#endif



        return( 1 );
}


int CScopeApp::AllocateBuffers( void ) {

        TRACE0("CScopeApp::AllocateBuffers\n");

        int   i, j, iret, nbytes;
        long *pData, joff;


        if (!ComputeBufferSizes()) {
                IsAllocated = FALSE;
                return( IsAllocated );
                }


        // Only re-dim if array needs to get bigger

        if (LastNbufferbytes < Nbufferbytes) {

                LastNbufferbytes = Nbufferbytes;


                // Clean up any previous demux array and allocate new one

                if (pDemuxArray)
                        { free(pDemuxArray); pDemuxArray = NULL; }

                pDemuxArray = (long*) malloc(NbufferptsPlot * sizeof(long));



                // Repeat for hardware specific data array

                iret = AllocateDataArray();

                }
        else
                iret = 1;






        // Check for successful allocation

        if (!iret || !pDemuxArray) {

                IsAllocated = FALSE;

                ErrString = "Can not allocate data buffers.";
                LogMessage( MSGLEVEL_ERR, MSGDLG_SHOW, ERROR_NOT_FATAL, NULL );
                }

        else
                IsAllocated = TRUE;




        // When allocated, clear array (even if remains allocated from before)

        if (IsAllocated == TRUE) {

                pData = pDemuxArray;

                for ( j = 0 ; j < NchannelsPlot ; j++ ) {

                        joff = pDCoffset[j];

                        for ( i = 0 ; i < NsamplesPlot ; i++ )

                                *pData++ = joff;

                        }

                }





        // Allocate new GPS PPS and serial arrays if requested
        // and current ones are not large enough

        if ( (Ini.GpsEnable) && (LastNgpspts < Ngpspts) ) {

                LastNgpspts = Ngpspts;


                // Clean up previous GPS arrays

                if (pGpsPpsArray)
                        { free(pGpsPpsArray); pGpsPpsArray = NULL; }

                if (pGpsSerialArray)
                        { free(pGpsSerialArray); pGpsSerialArray = NULL; }



                // Allocate new GPS PPS and serial arrays

                nbytes          = Ngpspts * sizeof( PPSTDATA );
                pGpsPpsArray    = (PPSTDATA *) malloc( nbytes );

                nbytes          = Ngpspts * sizeof( SERIALDATA );
                pGpsSerialArray = (SERIALDATA *) malloc( nbytes );

                }



        return( IsAllocated );
}


int CScopeApp::ComputeActiveChannels( ) {

        TRACE0("CScopeApp::ComputeActiveChannels\n");

        int i;


        // Count number of active channels and set their order
        // This forces channels to be scanned in numerically ascending order

        Ini.Nchannels = 0;

        for ( i = 0 ; i < MaxTotalChannel ; i++ ) {

                if (Ini.pScanActive[i] == CHECKED) {
                        Ini.pScanOrder[Ini.Nchannels] = i;
                        Ini.Nchannels++;
                        }
                }


        // Ensure at least one active channel

        if (Ini.Nchannels == 0) {
                Ini.pScanActive[0] = CHECKED;
                Ini.pScanOrder[0]  = 0;
                Ini.Nchannels      = 1;
                if (Ini.DigitalEnable) { // turn on digital channel too
                        i = MaxAnalogChannel;
                        Ini.pScanActive[i] = CHECKED;
                        Ini.pScanOrder[1]  = Ini.Nchannels;
                        Ini.Nchannels++;
                        } // end if DigitalEnable
                } // end if Ini.Nchannels == 0


        NchannelsPlot = Ini.Nchannels;
        if (Ini.GpsEnable)
                NchannelsPlot += (MaxGpsPlotChannel-MaxGpsChannel);
        if (Ini.DigitalEnable)
                NchannelsPlot += (MaxDigPlotChannel-MaxDigitalChannel);
        if (Ini.CounterEnable)
                NchannelsPlot += (MaxCntPlotChannel-MaxCounterChannel);


        // Ensure valid display number

        if (Ini.PltNdisplay > NchannelsPlot) {
                Ini.PltNdisplay   = NchannelsPlot;
                Ini.PltFirstTrace = 0;
                }


        return( 1 );
}


int CScopeApp::UpdateActiveChannels( ) {

        TRACE0("CScopeApp::UpdateActiveChannels\n");

        int  i, j, State;
        char SigString[MAXSTR];


        // This function turns on all analog channels.  It should be called
        // whenever the number of analog channels changes (eg from hardware
        // dialog).


        // First turn on all analog channels

        for ( i = 0 ; i < MaxAnalogChannel ; i++ ) {

                sprintf( SigString, "CH%02d", i );
                SrStringCopy( &Ini.pScanTitle[i*MAXTITLE], SigString, MAXTITLE );
                Ini.pScanActive[i]  = CHECKED;
                Ini.pScanTrigger[i] = CHECKED;
                Ini.pScanChType[i]  = CH_TYPE_ANALOG;
                }

        Ini.Nchannels = MaxAnalogChannel;



        // Then set digital channels

        if (Ini.DigitalEnable) {
                State          = CHECKED;
                DigitalIndex   = MaxAnalogChannel;
                Ini.Nchannels += MaxDigitalChannel;
                }
        else {
                State          = UNCHECKED;
                DigitalIndex   = SRDAT_INDEXCHANNEL_NONE;
                }

        for ( i = 0 ; i < MaxDigitalChannel ; i++ ) {
                j = MaxAnalogChannel + i;
                SrStringCopy( &Ini.pScanTitle[j*MAXTITLE], "DGTL", MAXTITLE );
                Ini.pScanActive[j]  = State;
                Ini.pScanTrigger[j] = UNCHECKED;
                Ini.pScanChType[j]  = CH_TYPE_DIGITAL;
                }

        for ( j = 0 ; j < MaxDigPlotChannel ; j++ ) {
                sprintf( SigString, "DIG%1d", j );
                SrStringCopy( &Ini.pDigTitle[j*MAXTITLE], SigString, MAXTITLE );
                }





        // Finally set gps and counter channels
        
        if (Ini.GpsEnable) {
                State        = CHECKED;
                GpsMarkIndex = MaxAnalogChannel;
                if (Ini.DigitalEnable)
                        GpsMarkIndex += MaxDigitalChannel;
                Ini.Nchannels += MaxGpsChannel;

                }
        else {
                State        = UNCHECKED;
                GpsMarkIndex = SRDAT_INDEXCHANNEL_NONE;
                }

        for ( i = 0 ; i < MaxGpsChannel ; i++ ) {
                j = MaxAnalogChannel + MaxDigitalChannel + i;
                SrStringCopy( &Ini.pScanTitle[j*MAXTITLE], "MARK", MAXTITLE );
                Ini.pScanActive[j]  = State;
                Ini.pScanTrigger[j] = UNCHECKED;
                Ini.pScanChType[j]  = CH_TYPE_GPSMARK;
                }


        if (Ini.CounterEnable) {
                State          = CHECKED;
                Ini.Nchannels += MaxCounterChannel;
                CounterIndex   = Ini.Nchannels-1;
                }
        else {
                State          = UNCHECKED;
                CounterIndex   = SRDAT_INDEXCHANNEL_NONE;
                }

        for ( i = 0 ; i < MaxCounterChannel ; i++ ) {
                j = MaxAnalogChannel + MaxDigitalChannel + MaxGpsChannel + i;
                SrStringCopy( &Ini.pScanTitle[j*MAXTITLE], "CNTR", MAXTITLE );
                Ini.pScanActive[j]  = State;
                Ini.pScanTrigger[j] = UNCHECKED;
                Ini.pScanChType[j]  = CH_TYPE_GPSCOUNTER;
                }
        



        NchannelsPlot = Ini.Nchannels;
        if (Ini.GpsEnable)
                NchannelsPlot += (MaxGpsPlotChannel-MaxGpsChannel);
        if (Ini.DigitalEnable)
                NchannelsPlot += (MaxDigPlotChannel-MaxDigitalChannel);
        if (Ini.CounterEnable)
                NchannelsPlot += (MaxCntPlotChannel-MaxCounterChannel);


        // Fill scan order array with the active channels

        j = 0;
        for ( i = 0 ; i < MaxTotalChannel ; i++ ) {
                if (Ini.pScanActive[i] == CHECKED)
                        Ini.pScanOrder[j++] = i;
                }


        // Set display to show all channels

        Ini.PltNdisplay   = NchannelsPlot;
        Ini.PltFirstTrace = 0;


        return( 1 );
}


int CScopeApp::AllocateScanArrays() {

        TRACE0("CScopeApp::AllocateScanArrays\n");

        int j;


        // Clean up old scan buffers and offset

        FreeScanArrays();


        // Allocate new scan buffers and offset

        pTMPoffset  = (double*) malloc( MaxAllocTotalPlotChannel * sizeof(double) );
        pDCoffset   = (long*)   malloc( MaxAllocTotalPlotChannel * sizeof(long) );
        pPlotChType = (int*)    malloc( MaxAllocTotalPlotChannel * sizeof(int) );
        pPlotTitle  = (char*)   malloc( MaxAllocTotalPlotChannel * sizeof(char) * MAXTITLE );

        Ini.pScanActive  = (long*) malloc( MaxAllocTotalChannel * sizeof(long) );
        Ini.pScanTrigger = (long*) malloc( MaxAllocTotalChannel * sizeof(long) );
        Ini.pScanChType  = (int*)  malloc( MaxAllocTotalChannel * sizeof(int) );
        Ini.pScanOrder   = (int*)  malloc( MaxAllocTotalChannel * sizeof(int) );
        Ini.pScanTitle   = (char*) malloc( MaxAllocTotalChannel * sizeof(char) * MAXTITLE );

        Ini.pDigTitle   = (char*) malloc( MaxAllocDigPlotChannel * sizeof(char) * MAXTITLE );

        IniSave.pScanActive  = (long*) malloc( MaxAllocTotalChannel * sizeof(long) );
        IniSave.pScanTrigger = (long*) malloc( MaxAllocTotalChannel * sizeof(long) );
        IniSave.pScanChType  = (int*)  malloc( MaxAllocTotalChannel * sizeof(int) );
        IniSave.pScanOrder   = (int*)  malloc( MaxAllocTotalChannel * sizeof(int) );
        IniSave.pScanTitle   = (char*) malloc( MaxAllocTotalChannel * sizeof(char) * MAXTITLE );

        IniSave.pDigTitle   = (char*) malloc( MaxAllocDigPlotChannel * sizeof(char) * MAXTITLE );
        

        if (!Ini.pScanActive  || !IniSave.pScanActive  ||
            !Ini.pScanOrder   || !IniSave.pScanOrder   ||
            !Ini.pScanChType  || !IniSave.pScanChType  ||
            !Ini.pScanTitle   || !IniSave.pScanTitle   ||
            !Ini.pScanTrigger || !IniSave.pScanTrigger ||
            !Ini.pDigTitle    || !IniSave.pDigTitle    ||
            !pTMPoffset || !pDCoffset || !pPlotChType || !pPlotTitle ) {
                ErrString = "Could not allocate scan buffers or offset.";
                LogMessage( MSGLEVEL_ERR, MSGDLG_SHOW, ERROR_FATAL, NULL );
                return( 0 );
                }



        // Initialize offsets, channels done in IniSetDefaults

        for ( j = 0 ; j < MaxAllocTotalPlotChannel ; j++ ) {

                pTMPoffset[j]  = 0.0;
                pDCoffset[j]   = 0;
                pPlotChType[j] = CH_TYPE_ANALOG;
                strcpy( &pPlotTitle[j*MAXTITLE], "" );
                }


        return( 1 );
}


int CScopeApp::FreeScanArrays() {

        TRACE0("CScopeApp::FreeScanArrays\n");

        if (pTMPoffset)
                { free(pTMPoffset);  pTMPoffset = NULL; }

        if (pDCoffset)
                { free(pDCoffset);   pDCoffset  = NULL; }

        if (pPlotChType)
                { free(pPlotChType); pPlotChType  = NULL; }

        if (pPlotTitle)
                { free(pPlotTitle);  pPlotTitle  = NULL; }


        if (Ini.pScanActive)
                { free(Ini.pScanActive);  Ini.pScanActive = NULL; }

        if (Ini.pScanTrigger)
                { free(Ini.pScanTrigger); Ini.pScanTrigger = NULL; }

        if (Ini.pScanChType)
                { free(Ini.pScanChType);  Ini.pScanChType = NULL; }

        if (Ini.pScanOrder)
                { free(Ini.pScanOrder);   Ini.pScanOrder  = NULL; }

        if (Ini.pScanTitle)
                { free(Ini.pScanTitle);   Ini.pScanTitle  = NULL; }

        if (Ini.pDigTitle)
                { free(Ini.pDigTitle);    Ini.pDigTitle = NULL; }

        

        if (IniSave.pScanActive)
                { free(IniSave.pScanActive);  IniSave.pScanActive = NULL; }

        if (IniSave.pScanTrigger)
                { free(IniSave.pScanTrigger); IniSave.pScanTrigger = NULL; }

        if (IniSave.pScanChType)
                { free(IniSave.pScanChType);  IniSave.pScanChType = NULL; }

        if (IniSave.pScanOrder)
                { free(IniSave.pScanOrder);   IniSave.pScanOrder  = NULL; }

        if (IniSave.pScanTitle)
                { free(IniSave.pScanTitle);   IniSave.pScanTitle  = NULL; }


        if (IniSave.pDigTitle)
                { free(IniSave.pDigTitle);    IniSave.pDigTitle = NULL; }

        return( 1 );
}

int CScopeApp::UpdatePlotTraceInfo() {

        TRACE0("CScopeApp::UpdatePlotTraceInfo\n");

        int i, j, k, ichan;

        // First, clear off any existing plot titles

        for ( j = 0 ; j < MaxTotalPlotChannel ; j++ )
                strcpy( &pPlotTitle[j*MAXTITLE], "" );


        // Then update arrays for plot titles and trace type info
        // based on scan info.

        // Trace info, indexed by j, goes to NchannelsPlot and
        // scan order info, indexed by i, goes to Ini.Nchannels and
        // selects the regular scan info indexed by ichan which
        // goes to MaxTotalChannel.

        j = 0;

        for ( i = 0 ; i < Ini.Nchannels ; i++ ) {

                ichan = Ini.pScanOrder[i];

                if (Ini.pScanChType[ichan] == CH_TYPE_DIGITAL) {
                        for ( k = 0 ; k < MaxDigPlotChannel ; k++ ) {
                                pPlotChType[j] = Ini.pScanChType[ichan];
                                SrStringCopy( &pPlotTitle[j*MAXTITLE],
                                              &Ini.pDigTitle[k*MAXTITLE],
                                              MAXTITLE );
                                j++;
                                }
                        }
                else if (Ini.pScanChType[ichan] == CH_TYPE_GPSCOUNTER) {
                        ; // do not plot or show in list
                        }
                else { // CH_TYPE_ANALOG or CH_TYPE_GPS
                        pPlotChType[j] = Ini.pScanChType[ichan];
                        SrStringCopy( &pPlotTitle[j*MAXTITLE],
                                      &Ini.pScanTitle[ichan*MAXTITLE],
                                       MAXTITLE );
                        j++;
                        }
                
                }

        return( 1 );
}

int CScopeApp::IniFileSetShortName() {
        
        TRACE0("CScopeApp::IniFileSetShortName\n");

        char fname[_MAX_FNAME], ext[_MAX_EXT];


        if ( Ini.IniFile.IsEmpty() )
                Ini.IniFileShort = Ini.IniFile;
        else {
                _splitpath( pAPP->Ini.IniFile, NULL, NULL, fname, ext );
                pAPP->Ini.IniFileShort.Format( "%s%s", fname, ext );
                }

        return( 1 );
}
int CScopeApp::IniVerifyData() {

        TRACE0("CScopeApp::IniVerifyData\n");

        // When changing defaults, be sure to change in IniSetDefaults too

        int  i, j, ichan, maxValue, State;



        // Program setting parameter verification

        if ((Ini.DoForbidMultiple != TRUE) && (Ini.DoForbidMultiple != FALSE))
                Ini.DoForbidMultiple = TRUE;

        if ((Ini.DoAutoRun != TRUE) && (Ini.DoAutoRun != FALSE))
                Ini.DoAutoRun = FALSE;

        if ((Ini.DoExitOnError != TRUE) && (Ini.DoExitOnError != FALSE))
                Ini.DoExitOnError = TRUE;

        if ((Ini.DoInteractive != TRUE) && (Ini.DoInteractive != FALSE))
                Ini.DoInteractive = TRUE;

        if ((Ini.AtodEnable != TRUE) && (Ini.AtodEnable != FALSE))
                Ini.AtodEnable = TRUE;

        if ((Ini.DoStatusBar != TRUE) && (Ini.DoStatusBar != FALSE))
                Ini.DoStatusBar = TRUE;

        if ((Ini.DoPlotTrace != TRUE) && (Ini.DoPlotTrace != FALSE))
                Ini.DoPlotTrace = TRUE;





        // Trigger parameter verification

        if ((Ini.TrgType < TRIG_NONE) || (Ini.TrgType > TRIG_DIGEDGE))
                Ini.TrgType = TRIG_NONE;

        if ((Ini.TrgThreshold < 0) || (Ini.TrgThreshold > MaxTriggerThreshold))
                Ini.TrgThreshold = DefaultThreshold;

        if ((Ini.TrgDigLevelLine < 0) || (Ini.TrgDigLevelLine > MaxTriggerDigioLine))
                Ini.TrgDigLevelLine = 0;

        if ((Ini.TrgDigEdgeLine < 0) || (Ini.TrgDigEdgeLine > MaxTriggerDigioLine))
                Ini.TrgDigEdgeLine = 0;

        if ((Ini.TrgActiveLevel != TRIGLVL_LO) &&
            (Ini.TrgActiveLevel != TRIGLVL_HI)  )
                Ini.TrgActiveLevel = TRIGLVL_HI;

        if ((Ini.TrgActiveEdge != TRIGEDGE_FALLING) &&
            (Ini.TrgActiveEdge != TRIGEDGE_RISING)  )
                Ini.TrgActiveEdge = TRIGEDGE_RISING;

        if ((Ini.TrgActionSave != TRUE) && (Ini.TrgActionSave != FALSE))
                Ini.TrgActionSave = TRUE;

        if (Ini.TrgFilesToSave <= 0)
                Ini.TrgFilesToSave = 2;

        if ((Ini.TrgActionSignal != TRUE) && (Ini.TrgActionSignal != FALSE))
                Ini.TrgActionSignal = FALSE;

        if ((Ini.TrgLineToSignal < 0) || (Ini.TrgLineToSignal > MaxTriggerDigioLine))
                Ini.TrgLineToSignal = 0;

        if ((Ini.TrgLatchLine != TRUE) && (Ini.TrgLatchLine != FALSE))
                Ini.TrgLatchLine = FALSE;

        for ( i = 0 ;  i < MaxTotalChannel ; i++ ) {
                if ((Ini.pScanTrigger[i] != CHECKED) && (Ini.pScanTrigger[i] != UNCHECKED))
                        Ini.pScanTrigger[i] = UNCHECKED;
                }




        // Output file parameter verification

        if ((Ini.OutFileFormat < FILEFMT_NONE) ||
            (Ini.OutFileFormat > FILEFMT_DAT)  )
                Ini.OutFileFormat = FILEFMT_NONE;

        if ((Ini.OutFileNaming < FILE_ONE)    ||
            (Ini.OutFileNaming > FILE_STDOUT) )
                Ini.OutFileNaming = FILE_ONE;

        if (Ini.OutFileNaming == FILE_SEQHEX)
                Ini.OutFileNaming = FILE_SEQDEC;

        if (Ini.OutSeqStart <= 0)
                Ini.OutSeqStart = 0;

        if (Ini.OutNblock <= 0)
                Ini.OutNblock = Nplotbuffers;

        if (!SrFileExist((LPCTSTR)Ini.OutDir))
                Ini.OutDir = CurrDir;




        // Channel parameter verification

        if (Ini.DigitalEnable) {
                State        = CHECKED;
                DigitalIndex = MaxAnalogChannel;
                }
        else {
                State        = UNCHECKED;
                DigitalIndex = SRDAT_INDEXCHANNEL_NONE;
                }

        for ( i = 0 ; i < MaxDigitalChannel ; i++ ) {
                j = MaxAnalogChannel + i;
                Ini.pScanActive[j]  = State;
                Ini.pScanTrigger[j] = UNCHECKED;
                }


        if (Ini.GpsEnable) {
                State        = CHECKED;
                GpsMarkIndex = MaxAnalogChannel;
                if (Ini.DigitalEnable)
                        GpsMarkIndex += MaxDigitalChannel;
                }
        else {
                State        = UNCHECKED;
                GpsMarkIndex = SRDAT_INDEXCHANNEL_NONE;
                }

        for ( i = 0 ; i < MaxGpsChannel ; i++ ) {
                j = MaxAnalogChannel + MaxDigitalChannel + i;
                Ini.pScanActive[j]  = State;
                Ini.pScanTrigger[j] = UNCHECKED;
                }

        CounterIndex = SRDAT_INDEXCHANNEL_NONE;
        if (Ini.CounterEnable) {
                State = CHECKED;
                if (GpsMarkIndex != SRDAT_INDEXCHANNEL_NONE &&
                    MaxCounterChannel > 0)
                        CounterIndex = GpsMarkIndex + 1;
                }
        else
                State = UNCHECKED;

        for ( i = 0 ; i < MaxCounterChannel ; i++ ) {
                j = MaxAnalogChannel + MaxDigitalChannel + MaxGpsChannel + i;
                Ini.pScanActive[j]  = State;
                Ini.pScanTrigger[j] = UNCHECKED;
                }
        

        ichan = 0;
        for ( i = 0 ; i < MaxTotalChannel ; i++ ) {

                if (strlen(&Ini.pScanTitle[i*MAXTITLE]) > MAXTITLE)
                        Ini.pScanTitle[(i*MAXTITLE)+MAXTITLE-1] = '\0';

                if (Ini.pScanActive[i] == CHECKED)
                        ichan++;
                }

        if (Ini.Nchannels != ichan)
                Ini.Nchannels = ichan;
        NchannelsPlot = Ini.Nchannels;
        if (Ini.DigitalEnable)
                NchannelsPlot += (MaxDigPlotChannel-MaxDigitalChannel);
        if (Ini.GpsEnable)
                NchannelsPlot += (MaxGpsPlotChannel-MaxGpsChannel);
        if (Ini.CounterEnable)
                NchannelsPlot += (MaxCntPlotChannel-MaxCounterChannel);

        for ( i = 0 ; i < MaxDigPlotChannel ; i++ ) {

                if (strlen(&Ini.pDigTitle[i*MAXTITLE]) > MAXTITLE)
                        Ini.pDigTitle[(i*MAXTITLE)+MAXTITLE-1] = '\0';
                }
        
        if ((Ini.Nsamples<=0) || ((MaxSample!=0) && (Ini.Nsamples>MaxSample)))
                Ini.Nsamples = DefaultNsamples;




        // Display parameter verification

        if ((Ini.PltClipping != TRUE) && (Ini.PltClipping != FALSE))
                Ini.PltClipping = FALSE;

        if ((Ini.PltAutoDC != TRUE) && (Ini.PltAutoDC != FALSE))
                Ini.PltAutoDC = TRUE;

        if ((Ini.PltShiftBy < 0) || (Ini.PltShiftBy > MaxAmpShift-1))
                Ini.PltShiftBy = MaxAmpShift/2;

        if ((Ini.PltNdisplay <= 0) || (Ini.PltNdisplay > NchannelsPlot))
                Ini.PltNdisplay = DefaultNdisplay;

        maxValue = NchannelsPlot - Ini.PltNdisplay;
        if ((Ini.PltFirstTrace < 0) || (Ini.PltFirstTrace > maxValue))
                Ini.PltFirstTrace = maxValue;

        maxValue = NsamplesPlot - GetSystemMetrics(SM_CXFULLSCREEN) - 1;
        if (maxValue < 0)  maxValue = 0;
        if ((Ini.PltFirstSample < 0) || (Ini.PltFirstSample > maxValue))
                Ini.PltFirstSample = 0;

        if ((Ini.PltColorScheme < CS_MONO) || (Ini.PltColorScheme > CS_BROWN))
                Ini.PltColorScheme = DefaultColor;



        // Gps parameter verification

        if (Ini.GpsDriverName.GetLength() == 0)
                Ini.GpsDriverName = PARGPS_0;

#ifdef GPS_AVAILABLE
        if (Ini.GpsModel == GPSMODEL_UNKNOWN) {
                Ini.GpsModel = ParGpsGetDefaultModel(
                                            Ini.GpsDriverName.GetBuffer(0) );
                Ini.GpsDriverName.ReleaseBuffer();
                }
#endif
        
	if ( (Ini.GpsModel <= GPSMODEL_NONE) || (Ini.GpsModel >= GPSMODEL_MAX) )
		Ini.GpsModel = GPSMODEL_TRIMBLE;

	if ( (Ini.GpsSerialPort < 1) || (Ini.GpsSerialPort > 9) )
		Ini.GpsSerialPort = SERIAL_PORT_1;
        
        

        // Acquisition parameter verification done in the derived class...


        // Many of the other variables are assumed to be
        // checked from the derived class VerifyIniData call
        // which should complete after having called this
        // parent class version



        // Additional consistency check
        // for requesting GPS, but not writing it out.

        if ( (Ini.OutFileFormat == FILEFMT_OUT) && (Ini.GpsEnable) ) {

                ErrString  = "Only partial GPS data is saved in .OUT files.";
                TmpString  = "To save complete GPS information, please use the\n";
                TmpString += ".DAT format.  If partial GPS data is sufficient,\n";
                TmpString += "you may continue with saving to .OUT files.";
                LogMessage( MSGLEVEL_WARN, MSGDLG_SHOW, ERROR_NOT_FATAL, &TmpString );
                }


        // Additional consistency check
        // for requesting GPS, but setting the sampling rate so low
        // that a PARGPS driver software FIFO buffer would overflow

        if ( !GpsTimingCheckBufferSize( Ini.GpsEnable, Ini.Nsamples, Ini.AcqSPS ) ) {

                ErrString  = "Low sample rate would overflow a GPS buffer.";
                TmpString  = "Please increase sample rate or turn GPS off.\n";
                LogMessage( MSGLEVEL_WARN, MSGDLG_SHOW, ERROR_NOT_FATAL, &TmpString );
                }


        return( 1 );
}


int CScopeApp::IniSetDefaults() {

        TRACE0("CScopeApp::IniSetDefaults\n");

        int  i, j;
        char SigString[MAXSTR];


        // When changing defaults, be sure to change in IniVerifyData too


        // Program setting parameter defaults

        IniFileSetShortName();
        
        if ( Ini.IniFile.IsEmpty() )
                Ini.IniTitle = "";
        else
                Ini.IniTitle = Ini.IniFileShort;

        Ini.DoForbidMultiple = TRUE;
        Ini.DoAutoRun        = FALSE;
        Ini.DoExitOnError    = TRUE;
        Ini.DoInteractive    = TRUE;
        Ini.DoStatusBar      = TRUE;
        Ini.DoPlotTrace      = TRUE;
        Ini.AtodEnable       = TRUE;


        // Trigger parameter defaults

        Ini.TrgType         = TRIG_NONE;
        Ini.TrgThreshold    = DefaultThreshold;
        Ini.TrgDigLevelLine = 0;
        Ini.TrgDigEdgeLine  = 0;
        Ini.TrgActiveLevel  = TRIGLVL_HI;
        Ini.TrgActiveEdge   = TRIGEDGE_RISING;
        Ini.TrgActionSave   = TRUE;
        Ini.TrgFilesToSave  = 2;
        Ini.TrgActionSignal = FALSE;
        Ini.TrgLineToSignal = 0;
        Ini.TrgLatchLine    = FALSE;


        // Output file parameter defaults

        Ini.OutNblock     = Nplotbuffers;
        Ini.OutFileFormat = FILEFMT_NONE;
        Ini.OutFileNaming = FILE_ONE;
        Ini.OutSeqStart   = 0;
        Ini.OutDir        = CurrDir;


        // Display parameter defaults

        Ini.PltClipping    = FALSE;
        Ini.PltAutoDC      = TRUE;
        Ini.PltShiftBy     = MaxAmpShift/2;
        Ini.PltNdisplay    = DefaultNdisplay;
        Ini.PltFirstSample = 0;
        Ini.PltFirstTrace  = 0;
        Ini.PltColorScheme = DefaultColor;


        // Acquisition parameters set in the derived class...


        // Set analog channel parameter defaults

        for ( i = 0 ; i < MaxAnalogChannel ; i++ ) {

                sprintf( SigString, "CH%02d", i );
                SrStringCopy( &Ini.pScanTitle[i*MAXTITLE], SigString, MAXTITLE );
                Ini.pScanActive[i]  = CHECKED;
                Ini.pScanTrigger[i] = CHECKED;
                Ini.pScanChType[i]  = CH_TYPE_ANALOG;
                Ini.pScanOrder[i]   = i;
                }


        // Set digital channel parameter defaults

        Ini.DigitalEnable = TRUE;

        i = MaxAnalogChannel;
        SrStringCopy( &Ini.pScanTitle[i*MAXTITLE], "DGTL", MAXTITLE );
        Ini.pScanActive[i]  = CHECKED; // For Ini.DigitalEnable = TRUE
        Ini.pScanTrigger[i] = UNCHECKED;
        Ini.pScanChType[i]  = CH_TYPE_DIGITAL;
        Ini.pScanOrder[i]   = i;

        for ( j = 0 ; j < MaxDigPlotChannel ; j++ ) {
                sprintf( SigString, "DIG%1d", j );
                SrStringCopy( &Ini.pDigTitle[j*MAXTITLE], SigString, MAXTITLE );
                }

        
        // Set GPS mark channel parameter default

        i = MaxAnalogChannel + MaxDigitalChannel;
        SrStringCopy( &Ini.pScanTitle[i*MAXTITLE], "MARK", MAXTITLE );
        Ini.pScanActive[i]  = UNCHECKED;  // For Ini.GpsEnable = FALSE
        Ini.pScanTrigger[i] = UNCHECKED;
        Ini.pScanChType[i]  = CH_TYPE_GPSMARK;
        Ini.pScanOrder[i]   = i;


        // Set GPS counter channel parameter default
        
        Ini.CounterEnable = FALSE;
        i = MaxAnalogChannel + MaxDigitalChannel + MaxGpsChannel;
        SrStringCopy( &Ini.pScanTitle[i*MAXTITLE], "CNTR", MAXTITLE );
        Ini.pScanActive[i]  = UNCHECKED;  // For Ini.CounterEnable = FALSE
        Ini.pScanTrigger[i] = UNCHECKED;
        Ini.pScanChType[i]  = CH_TYPE_GPSCOUNTER;
        Ini.pScanOrder[i]   = i;


        Ini.Nsamples  = DefaultNsamples;
        Ini.Nchannels = 0;
        for ( i = 0 ; i < MaxTotalChannel ; i++ )
                if (Ini.pScanActive[i] == CHECKED)
                        Ini.Nchannels++;
        
        NchannelsPlot = Ini.Nchannels;
        if (Ini.GpsEnable)
                NchannelsPlot += (MaxGpsPlotChannel-MaxGpsChannel);
        if (Ini.DigitalEnable)
                NchannelsPlot += (MaxDigPlotChannel-MaxDigitalChannel);
        if (Ini.CounterEnable)
                NchannelsPlot += (MaxCntPlotChannel-MaxCounterChannel);
        
        Ini.PltNdisplay = NchannelsPlot;



        // Gps parameters

        Ini.GpsModel         = GPSMODEL_UNKNOWN;
        Ini.GpsDriverName    = PARGPS_0;
        Ini.GpsSerialPort    = SERIAL_PORT_1;
        Ini.GpsEnable        = FALSE;
        Ini.GpsDisplayPps    = FALSE;
        Ini.GpsDisplaySerial = FALSE;



        // Many of the other variables are assumed to be
        // set from the derived class IniSetDefaults call
        // which should complete after having called this
        // parent class version

        return( 1 );
}

int CScopeApp::IniSaveCurrent() {

        TRACE0("CScopeApp::IniSaveCurrent\n");

        int i;

        // Save program setting parameters

        IniSave.IniTitle         = Ini.IniTitle;

        IniSave.DoForbidMultiple = Ini.DoForbidMultiple;
        IniSave.DoAutoRun        = Ini.DoAutoRun;
        IniSave.DoExitOnError    = Ini.DoExitOnError;
        IniSave.DoInteractive    = Ini.DoInteractive;
        IniSave.DoStatusBar      = Ini.DoStatusBar;
        IniSave.DoPlotTrace      = Ini.DoPlotTrace;
        IniSave.AtodEnable       = Ini.AtodEnable;



        // Trigger parameter defaults

        IniSave.TrgType          = Ini.TrgType;
        IniSave.TrgThreshold     = Ini.TrgThreshold;
        IniSave.TrgDigLevelLine  = Ini.TrgDigLevelLine;
        IniSave.TrgDigEdgeLine   = Ini.TrgDigEdgeLine;
        IniSave.TrgActiveLevel   = Ini.TrgActiveLevel;
        IniSave.TrgActiveEdge    = Ini.TrgActiveEdge;
        IniSave.TrgActionSave    = Ini.TrgActionSave;
        IniSave.TrgFilesToSave   = Ini.TrgFilesToSave;
        IniSave.TrgActionSignal  = Ini.TrgActionSignal;
        IniSave.TrgLineToSignal  = Ini.TrgLineToSignal;
        IniSave.TrgLatchLine     = Ini.TrgLatchLine;


        // Save output file parameters

        IniSave.OutNblock        = Ini.OutNblock;
        IniSave.OutFileFormat    = Ini.OutFileFormat;
        IniSave.OutFileNaming    = Ini.OutFileNaming;
        IniSave.OutSeqStart      = Ini.OutSeqStart;
        IniSave.OutDir           = Ini.OutDir;


        // Save display parameters

        IniSave.PltShiftBy       = Ini.PltShiftBy;
        IniSave.PltNdisplay      = Ini.PltNdisplay;
        IniSave.PltFirstSample   = Ini.PltFirstSample;
        IniSave.PltFirstTrace    = Ini.PltFirstTrace;
        IniSave.PltColorScheme   = Ini.PltColorScheme;
        IniSave.PltClipping      = Ini.PltClipping;
        IniSave.PltAutoDC        = Ini.PltAutoDC;


        // Save acquisition parameters

        IniSave.AcqGainLog       = Ini.AcqGainLog;
        IniSave.AcqSPS           = Ini.AcqSPS;


        // Save channel parameters

        for ( i = 0 ;  i < MaxTotalChannel ; i++ ) {

                strncpy( &IniSave.pScanTitle[i*MAXTITLE], &Ini.pScanTitle[i*MAXTITLE], MAXTITLE );
                IniSave.pScanActive[i]  = Ini.pScanActive[i];
                IniSave.pScanTrigger[i] = Ini.pScanTrigger[i];
                IniSave.pScanChType[i]  = Ini.pScanChType[i];
                IniSave.pScanOrder[i]   = Ini.pScanOrder[i];
                }

        for ( i = 0 ;  i < MaxDigPlotChannel ; i++ ) {

                strncpy( &IniSave.pDigTitle[i*MAXTITLE], &Ini.pDigTitle[i*MAXTITLE], MAXTITLE );
                }

        IniSave.Nsamples         = Ini.Nsamples;
        IniSave.Nchannels        = Ini.Nchannels;
        IniSave.DigitalEnable    = Ini.DigitalEnable;


        // Save GPS parameters

        IniSave.GpsModel         = Ini.GpsModel;
        IniSave.GpsDriverName    = Ini.GpsDriverName;
        IniSave.GpsSerialPort    = Ini.GpsSerialPort;
        IniSave.GpsEnable        = Ini.GpsEnable;
        IniSave.GpsDisplayPps    = Ini.GpsDisplayPps;
        IniSave.GpsDisplaySerial = Ini.GpsDisplaySerial;


        IniSave.CounterEnable    = Ini.CounterEnable;
        

        // Many of the other variables are assumed to be
        // set from the derived class SaveCurrentIni call
        // which should complete after having called this
        // parent class version

        return( 1 );
}

int CScopeApp::IniRestoreCurrent() {

        TRACE0("CScopeApp::IniRestoreCurrent\n");

        int i;

        // Restore program setting parameters

        Ini.IniTitle         = IniSave.IniTitle;

        Ini.DoForbidMultiple = IniSave.DoForbidMultiple;
        Ini.DoAutoRun        = IniSave.DoAutoRun;
        Ini.DoExitOnError    = IniSave.DoExitOnError;
        Ini.DoInteractive    = IniSave.DoInteractive;
        Ini.DoStatusBar      = IniSave.DoStatusBar;
        Ini.DoPlotTrace      = IniSave.DoPlotTrace;
        Ini.AtodEnable       = IniSave.AtodEnable;


        // Restore trigger parameters

        Ini.TrgType          = IniSave.TrgType;
        Ini.TrgThreshold     = IniSave.TrgThreshold;
        Ini.TrgDigLevelLine  = IniSave.TrgDigLevelLine;
        Ini.TrgDigEdgeLine   = IniSave.TrgDigEdgeLine;
        Ini.TrgActiveLevel   = IniSave.TrgActiveLevel;
        Ini.TrgActiveEdge    = IniSave.TrgActiveEdge;
        Ini.TrgActionSave    = IniSave.TrgActionSave;
        Ini.TrgFilesToSave   = IniSave.TrgFilesToSave;
        Ini.TrgActionSignal  = IniSave.TrgActionSignal;
        Ini.TrgLineToSignal  = IniSave.TrgLineToSignal;
        Ini.TrgLatchLine     = IniSave.TrgLatchLine;


        // Restore output file parameters

        Ini.OutNblock        = IniSave.OutNblock;
        Ini.OutFileFormat    = IniSave.OutFileFormat;
        Ini.OutFileNaming    = IniSave.OutFileNaming;
        Ini.OutSeqStart      = IniSave.OutSeqStart;
        Ini.OutDir           = IniSave.OutDir;


        // Restore display parameters

        Ini.PltShiftBy       = IniSave.PltShiftBy;
        Ini.PltNdisplay      = IniSave.PltNdisplay;
        Ini.PltFirstSample   = IniSave.PltFirstSample;
        Ini.PltFirstTrace    = IniSave.PltFirstTrace;
        Ini.PltColorScheme   = IniSave.PltColorScheme;
        Ini.PltClipping      = IniSave.PltClipping;
        Ini.PltAutoDC        = IniSave.PltAutoDC;


        // Restore acquisition parameters

        Ini.AcqGainLog       = IniSave.AcqGainLog;
        Ini.AcqSPS           = IniSave.AcqSPS;


        // Restore channel parameters

        for ( i = 0 ;  i < MaxTotalChannel ; i++ ) {

                strncpy( &Ini.pScanTitle[i*MAXTITLE], &IniSave.pScanTitle[i*MAXTITLE], MAXTITLE );
                Ini.pScanActive[i]  = IniSave.pScanActive[i];
                Ini.pScanTrigger[i] = IniSave.pScanTrigger[i];
                Ini.pScanChType[i]  = IniSave.pScanChType[i];
                Ini.pScanOrder[i]   = IniSave.pScanOrder[i];
                }

        for ( i = 0 ;  i < MaxDigPlotChannel ; i++ ) {

                strncpy( &Ini.pDigTitle[i*MAXTITLE], &IniSave.pDigTitle[i*MAXTITLE], MAXTITLE );
                }

        Ini.Nsamples         = IniSave.Nsamples;
        Ini.Nchannels        = IniSave.Nchannels;
        Ini.DigitalEnable    = IniSave.DigitalEnable;



        // Save GPS parameters

        Ini.GpsModel         = IniSave.GpsModel;
        Ini.GpsDriverName    = IniSave.GpsDriverName;
        Ini.GpsSerialPort    = IniSave.GpsSerialPort;
        Ini.GpsEnable        = IniSave.GpsEnable;
        Ini.GpsDisplayPps    = IniSave.GpsDisplayPps;
        Ini.GpsDisplaySerial = IniSave.GpsDisplaySerial;

        Ini.CounterEnable    = IniSave.CounterEnable;



        // Many of the other variables are assumed to be
        // set from the derived class SaveCurrentIni call
        // which should complete after having called this
        // parent class version

        return( 1 );
}

int CScopeApp::IniIsChanged() {

        TRACE0("CScopeApp::IniIsChanged\n");

        int i;

        // Compare saved and current ini. Return 1 if changed, 0 if the same


        // Check program setting parameters

        if (IniSave.IniTitle         != Ini.IniTitle)         return( 1 );

        if (IniSave.DoForbidMultiple != Ini.DoForbidMultiple) return( 1 );
        if (IniSave.DoAutoRun        != Ini.DoAutoRun)        return( 1 );
        if (IniSave.DoExitOnError    != Ini.DoExitOnError)    return( 1 );
        if (IniSave.DoInteractive    != Ini.DoInteractive)    return( 1 );
        if (IniSave.DoStatusBar      != Ini.DoStatusBar)      return( 1 );
        if (IniSave.DoPlotTrace      != Ini.DoPlotTrace)      return( 1 );
        if (IniSave.AtodEnable       != Ini.AtodEnable)       return( 1 );


        // Check trigger parameters

        if (IniSave.TrgType         != Ini.TrgType)          return( 1 );
        if (IniSave.TrgThreshold    != Ini.TrgThreshold)     return( 1 );
        if (IniSave.TrgDigLevelLine != Ini.TrgDigLevelLine)  return( 1 );
        if (IniSave.TrgDigEdgeLine  != Ini.TrgDigEdgeLine)   return( 1 );
        if (IniSave.TrgActiveLevel  != Ini.TrgActiveLevel)  return( 1 );
        if (IniSave.TrgActiveEdge   != Ini.TrgActiveEdge)    return( 1 );
        if (IniSave.TrgActionSave   != Ini.TrgActionSave)    return( 1 );
        if (IniSave.TrgFilesToSave  != Ini.TrgFilesToSave)   return( 1 );
        if (IniSave.TrgActionSignal != Ini.TrgActionSignal)  return( 1 );
        if (IniSave.TrgLineToSignal != Ini.TrgLineToSignal)  return( 1 );
        if (IniSave.TrgLatchLine    != Ini.TrgLatchLine)     return( 1 );


        // Check output file parameters

        if (IniSave.OutNblock      != Ini.OutNblock)        return( 1 );
        if (IniSave.OutFileFormat  != Ini.OutFileFormat)    return( 1 );
        if (IniSave.OutFileNaming  != Ini.OutFileNaming)    return( 1 );
        if (IniSave.OutSeqStart    != Ini.OutSeqStart)      return( 1 );
        if (IniSave.OutDir         != Ini.OutDir)           return( 1 );


        // Check display parameters

        if (IniSave.PltShiftBy     != Ini.PltShiftBy)       return( 1 );
        if (IniSave.PltNdisplay    != Ini.PltNdisplay)      return( 1 );
        if (IniSave.PltFirstSample != Ini.PltFirstSample)   return( 1 );
        if (IniSave.PltFirstTrace  != Ini.PltFirstTrace)    return( 1 );
        if (IniSave.PltColorScheme != Ini.PltColorScheme)   return( 1 );
        if (IniSave.PltClipping    != Ini.PltClipping)      return( 1 );
        if (IniSave.PltAutoDC      != Ini.PltAutoDC)        return( 1 );


        // Check acquisition parameters

        if (IniSave.AcqGainLog     != Ini.AcqGainLog)       return( 1 );
        if (IniSave.AcqSPS         != Ini.AcqSPS)           return( 1 );


        // Check channel parameters

        if (IniSave.Nsamples       != Ini.Nsamples)         return( 1 );
        if (IniSave.Nchannels      != Ini.Nchannels)        return( 1 );
        if (IniSave.DigitalEnable  != Ini.DigitalEnable)  return( 1 );

        for ( i = 0 ;  i < MaxTotalChannel ; i++ ) {

                if (IniSave.pScanActive[i]  != Ini.pScanActive[i])  return( 1 );
                if (IniSave.pScanTrigger[i] != Ini.pScanTrigger[i]) return( 1 );
                if (IniSave.pScanChType[i]  != Ini.pScanChType[i])  return( 1 );
                if (IniSave.pScanOrder[i]   != Ini.pScanOrder[i])   return( 1 );
                if (0 != strncmp( &IniSave.pScanTitle[i*MAXTITLE],
                                  &Ini.pScanTitle[i*MAXTITLE],
                                  MAXTITLE ) )                     return( 1 );
                }

        for ( i = 0 ;  i < MaxDigPlotChannel ; i++ ) {

                if (0 != strncmp( &IniSave.pDigTitle[i*MAXTITLE],
                                  &Ini.pDigTitle[i*MAXTITLE],
                                  MAXTITLE ) )                     return( 1 );
                }
        



        // Check GPS parameters

        if (IniSave.GpsModel         != Ini.GpsModel)          return( 1 );
        if (IniSave.GpsDriverName    != Ini.GpsDriverName)     return( 1 );
        if (IniSave.GpsSerialPort    != Ini.GpsSerialPort)     return( 1 );
        if (IniSave.GpsEnable        != Ini.GpsEnable)         return( 1 );
        if (IniSave.GpsDisplayPps    != Ini.GpsDisplayPps)     return( 1 );
        if (IniSave.GpsDisplaySerial != Ini.GpsDisplaySerial)  return( 1 );

        if (IniSave.CounterEnable    != Ini.CounterEnable)     return( 1 );
        

        // Many of the other variables are assumed to be
        // compared in the derived class IniIsChanged call
        // which should complete after having called this
        // parent class version

        return( 0 );
}

int CScopeApp::IniInputFromFile( void ) {

        TRACE0("CScopeApp::IniInputFromFile\n");

        // Open input ini file

        if ( !parse.OpenIni( (LPCTSTR)Ini.IniFile ) ) {
                IniVerifyData();
                IniSaveCurrent();
                return( 0 );
                }



        // Set all values to their defaults

        IniSetDefaults( );


        // Set these to zero so IniReadKeyword can collect channels ok
        // Nchannels will be reset later by ComputeActiveChannels.

        IniClearChannels();


        // Parse ini file for new values

        while ( 1 ) {

                ErrCode = 0;

                parse.GetToken();

                if ( IniReadKeyword() )
                        ; // Token was recognized and processed

                else if ( parse.IsEOF() )

                        break;

                else   // Token unrecognized ...

                        parse.ErrorIni( "unknown ini syntax" );


                // An error was found and user requested cancel,
                // so restore current values and stop parsing.

                if (ErrCode) {
                        IniRestoreCurrent( );
                        break;
                        }
                }



        // Close ini file

        parse.CloseIni( );


        // Ensure that all parameter values are valid

        IniVerifyData();
        IniSaveCurrent();

        return( 1 );
}


int CScopeApp::IniOutputToFile( ) {

        TRACE0("CScopeApp::IniOutputToFile\n");

        FILE *OutFp;

        // Open output ini file

        OutFp = fopen( pAPP->Ini.IniFile, "w" );
        if ( !OutFp ) {
                ErrString.Format(
                                 "Can not open output ini file %s.",
                                 pAPP->Ini.IniFile
                                );
                LogMessage( MSGLEVEL_ERR, MSGDLG_SHOW, ERROR_NOT_FATAL, NULL );
                return( 0 );
                }


        // Print header and then keyword strings

        fprintf( OutFp, "; FILE: %s\n", pAPP->Ini.IniFile );
        fprintf( OutFp, "\n" );

        IniWriteKeyword( OutFp );


        // Close file

        fclose( OutFp );

        IniSaveCurrent();

        return( 1 );
}

int CScopeApp::IniClearChannels( void ) {

        TRACE0("CScopeApp::IniClearChannels\n");

        int i;


        // Temporarily clear number of channels and
        // the scan state for any that are optional

        NchannelsPlot = 0;
        Ini.Nchannels = 0;


        for ( i = MinTotalChannel ; i < MaxTotalChannel ; i++ ) {
                Ini.pScanActive[i]  = UNCHECKED;
                Ini.pScanTrigger[i] = UNCHECKED;
                }


        return( 1 );
}

int CScopeApp::IniAddChannel( int channel, char *title ) {

        TRACE0("CScopeApp::IniAddChannel\n");

        int i, iok;

        iok = 1;
        for ( i = 0 ; i < Ini.Nchannels ; i++ ) {
                if ( Ini.pScanOrder[i] == channel ) {
                        sprintf( pAPP->szTemp,
                                 "channel %d is defined twice",
                                 channel );
                        parse.ErrorIni( pAPP->szTemp );
                        iok = 0;
                        break;
                        }
                }
        if (iok) {
                strncpy( &Ini.pScanTitle[channel*MAXTITLE], title, MAXTITLE );
                Ini.pScanTitle[(channel*MAXTITLE)+MAXTITLE-1] = '\0';
                Ini.pScanActive[channel]                      = CHECKED;
                Ini.pScanTrigger[channel]                     = UNCHECKED;
                Ini.pScanOrder[Ini.Nchannels]                 = channel;
                if (channel < MaxAnalogChannel)
                        Ini.pScanChType[channel]             = CH_TYPE_ANALOG;
                else if (channel < MaxAnalogChannel + MaxDigitalChannel )
                        Ini.pScanChType[channel]             = CH_TYPE_DIGITAL;
                else if (channel < MaxAnalogChannel + MaxDigitalChannel + MaxGpsChannel)
                        Ini.pScanChType[channel]             = CH_TYPE_GPSMARK;
                else
                        Ini.pScanChType[channel]             = CH_TYPE_GPSCOUNTER;
                Ini.Nchannels++;
                }

        return( iok );
}


int CScopeApp::IniAddDigital( int channel, char *title ) {

        TRACE0("CScopeApp::IniAddDigital\n");

        int iok;

        iok = 1;

// WCT: Consider adding a check to prevent listing a digital channel twice
        
        strncpy( &Ini.pDigTitle[channel*MAXTITLE], title, MAXTITLE );
        Ini.pDigTitle[(channel*MAXTITLE)+MAXTITLE-1] = '\0';

        return( iok );
}


int CScopeApp::IniReadKeyword() {

        TRACE0("CScopeApp::IniReadKeyword\n");

        int           itmp, iok, TokenType;
        unsigned long ultmp;
        static int    LastChannel = -1;
        static int    LastDigital = -1;
        static int    IsChannel   = -1;


        // Check in already opened ini file for know common
        // keywords and process them to set program variables
        // Hardware specific keywords will be processed using
        // the derived class' version of this virtual function.


        iok = 1;


        // Check program setting parameters
        // IniTitle, AtodEnable
        // DoForbidMultiple, DoAutoRun, DoExitOnError, DoInteractive

        if ( parse.KeywordEquals( "Title" ) ) {

                if ( parse.GetToken() != TOKEN_STRING )
                        parse.ErrorIni( "illegal Title string");
                else
                        Ini.IniTitle = parse.TOKEN;
                }



        else if ( parse.KeywordEquals( "ForbidMultiple" ) ) {

                iok = parse.GetOnOff( &itmp, "ForbidMultiple must = {ON,OFF}" );

                if (iok)
                        Ini.DoForbidMultiple = itmp;
                }


        else if ( parse.KeywordEquals( "KeypressToRun" ) ) {

                // KeypressToRun FALSE implies DoAutoRun mode TRUE

                iok = parse.GetOnOff( &itmp, "KeypressToRun must = {ON,OFF}" );

                if (iok)
                        Ini.DoAutoRun = ( itmp ? 0 : 1 );
                }


        else if ( parse.KeywordEquals( "ExitOnError" ) ) {

                iok = parse.GetOnOff( &itmp, "ExitOnError must = {ON,OFF}" );

                if (iok)
                        Ini.DoExitOnError = itmp;
                }


        else if ( parse.KeywordEquals( "Interactive" ) ) {

                iok = parse.GetOnOff( &itmp, "Interactive must = {ON,OFF}" );

                if (iok)
                        Ini.DoInteractive = itmp;
                }


        else if ( parse.KeywordEquals( "AtodEnable" )   ||
                  parse.KeywordEquals( "SystemStatus" ) ) {

                iok = parse.GetOnOff( &itmp, "AtodEnable must = {ON,OFF}" );

                if (iok)
                        Ini.AtodEnable = itmp;
                }




        // Check output file parameters

        // Initialize output constants
        // (uses DefaultNsamples,Nplotbuffers,CurrDir):
        // OutNblock, OutFileFormat, OutFileNaming, OutSeqStart, OutDir

        else if ( parse.KeywordEquals( "OutputDirectory" ) ) {

                if ( parse.GetToken() != TOKEN_STRING )
                        parse.ErrorIni( "illegal OutputDirectory string" );
                else {
                        Ini.OutDir = parse.TOKEN;
                        if ( !SrFileExist( (LPCTSTR)Ini.OutDir ) ) {
                                sprintf(
                                        szTemp,
                                        "OutputDirectory %s must already exist",
                                        Ini.OutDir
                                        );
                                parse.ErrorIni( szTemp );
                                }
                        }
                }

        else if ( parse.KeywordEquals( "OutputFileFormat" ) ) {

                parse.GetToken();

                if ( parse.Match( "NONE" ) )
                        Ini.OutFileFormat = FILEFMT_NONE;

                else if ( parse.Match( "OUT" ) )
                        Ini.OutFileFormat = FILEFMT_OUT;

                else if ( parse.Match( "DAT" ) )
                        Ini.OutFileFormat = FILEFMT_DAT;

                else
                    parse.ErrorIni(
                        "OutputFileFormat must = {NONE,OUT,DAT}" );
                }


        else if ( parse.KeywordEquals( "OutputFileNaming" ) ||
                  parse.KeywordEquals( "OutputFileMode" )   ) {

                parse.GetToken();

                if ( parse.Match( "NONE" ) ) {
                        Ini.OutFileNaming = FILE_NONE;
                        parse.ErrorIni(
                           "OutputFileNaming = NONE obsolete, use OutputFileFormat = NONE" );
                        }

                else if ( parse.Match( "Single" ) )
                        Ini.OutFileNaming = FILE_ONE;

                else if ( parse.Match( "Sequential" ) )
                        Ini.OutFileNaming = FILE_SEQDEC;

                else if ( parse.Match( "Time" ) )
                        Ini.OutFileNaming = FILE_TIME;

                else if ( parse.Match( "StdOut" ) ) {
                        Ini.OutFileNaming = FILE_STDOUT;
                        parse.ErrorIni("OutputFileNaming = StdOut not supported");
                        }

                else
                    parse.ErrorIni(
                        "OutputFileNaming must = {Single,Sequential,Time}" );
                }


        else if ( parse.KeywordEquals( "OutputSeqStart" ) ) {

                iok = parse.GetInt(
                            &itmp,
                            "illegal starting sequence number integer value (hex values not allowed)"
                             );
                
                if (iok)
                        Ini.OutSeqStart = itmp;
                }


        else if ( parse.KeywordEquals( "OutputBlocks" ) ||
                  parse.KeywordEquals( "Blocks" ) ) {

                iok = parse.GetUlong( &ultmp, "illegal Blocks integer value" );

                if (iok)
                        Ini.OutNblock = (int)ultmp;
                }





        // Check display parameters

        // Initialize display constants (uses MaxAmpShift,DefaultNdisplay,DefaultColor)
        // DoStatusBar, DoPlotTrace, PltClipping, PltAutoDC
        // PltShiftBy, PltNdisplay, PltFirstSample,PltFirstTrace,PltColorScheme

        else if ( parse.KeywordEquals( "DisplayInfo" ) ||
                  parse.KeywordEquals( "InfoDisplay" ) ) {

                iok = parse.GetOnOff( &itmp, "DisplayInfo must = {ON,OFF}" );

                if (iok)
                        Ini.DoStatusBar = itmp;
                }


        else if ( parse.KeywordEquals( "DisplayDataTrace" ) ||
                  parse.KeywordEquals( "DisplayData" )      ||
                  parse.KeywordEquals( "DataDisplay" )      ) {

                iok = parse.GetOnOff( &itmp, "DisplayDataTrace must = {ON,OFF}" );

                if (iok)
                        Ini.DoPlotTrace = itmp;
                }


        else if ( parse.KeywordEquals( "DisplayColor" ) ) {

                parse.GetToken();

                if ( parse.Match( "MONOCHROME" ) )
                        Ini.PltColorScheme = CS_MONO;

                else if ( parse.Match( "GREEN" ) )
                        Ini.PltColorScheme = CS_GREEN;

                else if ( parse.Match( "BLUE" ) )
                        Ini.PltColorScheme = CS_BLUE;

                else if ( parse.Match( "BROWN" ) )
                        Ini.PltColorScheme = CS_BROWN;
                else
                    parse.ErrorIni(
                        "DisplayColor must = {MONOCHROME,GREEN,BLUE,BROWN}" );
                }


        else if ( parse.KeywordEquals( "DisplayClipping" ) ) {

                iok = parse.GetOnOff( &itmp, "DisplayClipping must = {ON,OFF}" );

                if (iok)
                        Ini.PltClipping = itmp;
                }


        else if ( parse.KeywordEquals( "DisplayNbits" ) ) {

                iok = parse.GetInt( &itmp, "illegal DisplayNbits integer value" );

                if ( iok && itmp >= MinAmpShift && itmp <= MaxAmpShift )
                        Ini.PltShiftBy = MaxAmpShift - itmp;
                else {
                        sprintf( pAPP->szTemp,
                                 "DisplayNbits must be %d to %d",
                                 MinAmpShift, MaxAmpShift
                               );
                        parse.ErrorIni( pAPP->szTemp );
                        }
                }


        else if ( parse.KeywordEquals( "DisplayAutoDC" ) ) {

                iok = parse.GetOnOff( &itmp, "DisplayAutoDC must = {ON,OFF}" );

                if (iok)
                        Ini.PltAutoDC = itmp;
                }


        else if ( parse.KeywordEquals( "DisplayFirstSample" ) ) {

                iok = parse.GetInt( &itmp, "illegal DisplayFirstSample integer value" );

                // True max is NsamplesPlot-TraceWidth, but that is not known yet
                if ( iok && itmp >= 0 && itmp <= MaxSample )
                        Ini.PltFirstSample = itmp;
                else {
                        sprintf( pAPP->szTemp,
                                 "DisplayFirstSample must be 0 to %d samples",
                                 MaxSample );
                        parse.ErrorIni( pAPP->szTemp );
                        }
                }


        else if ( parse.KeywordEquals( "DisplayFirstTrace" ) ) {

                iok = parse.GetInt( &itmp, "illegal DisplayFirstTrace integer value" );

                // True max is Nchannels-Ndisplay, but that is not known yet
                if ( iok && itmp >= 0 && itmp <= MaxTotalPlotChannel )
                        Ini.PltFirstTrace = itmp;
                else {
                        sprintf( pAPP->szTemp,
                                 "DisplayFirstTrace must be 0 to %d - Ndisplay channels",
                                 MaxTotalPlotChannel );
                        parse.ErrorIni( pAPP->szTemp );
                        }
                }


        else if ( parse.KeywordEquals( "DisplayNtraces" ) ) {

                iok = parse.GetInt( &itmp, "illegal DisplayNtraces integer value" );

                // True max is Nchannels, but that is not known yet
                if ( iok && itmp >= 0 && itmp <= MaxTotalPlotChannel )
                        Ini.PltNdisplay = itmp;
                else {
                        sprintf( pAPP->szTemp,
                                 "DisplayNtraces must be 0 to %d channels",
                                 MaxTotalPlotChannel );
                        parse.ErrorIni( pAPP->szTemp );
                        }
                }





        // Check for channel titles

        else if ( parse.Match( "Channel") ) {

                iok = parse.GetInt( &itmp, "illegal CHANNEL number");

                if (!iok)
                        itmp = -1;

                if ( itmp < 0 || MaxTotalPlotChannel <= itmp ) {
                        sprintf( pAPP->szTemp,
                                 "Channel number must be 0 to %d",
                                 MaxTotalPlotChannel-1
                               );
                        parse.ErrorIni( pAPP->szTemp );
                        }

                else if ( parse.GetToken() != TOKEN_EQUALS )
                        parse.ErrorIni( "missing = sign on CHANNEL");

                else if ( parse.GetToken() != TOKEN_STRING )
                        parse.ErrorIni( "illegal CHANNEL title");

                else if (NchannelsPlot >= MaxTotalPlotChannel)
                        parse.ErrorIni( "too many channels defined");

                else {
                        IniAddChannel( itmp, parse.TOKEN );
                        LastChannel = itmp;
                        IsChannel   = 1;
                        }
                }


        // Check for digital parameters

        else if ( parse.KeywordEquals( "DigitalEnable" ) ) {

                iok = parse.GetOnOff( &itmp, "DigitalEnable must = {ON,OFF}" );

                if (iok)
                        Ini.DigitalEnable = itmp;

                }


        else if ( parse.Match( "Digital") ) {

                iok = parse.GetInt( &itmp, "illegal Digital number");

                if (!iok)
                        itmp = -1;

                if ( itmp < 0 || MaxDigPlotChannel <= itmp ) {
                        sprintf( pAPP->szTemp,
                                 "Digital channel number must be 0 to %d",
                                 MaxDigPlotChannel-1
                               );
                        parse.ErrorIni( pAPP->szTemp );
                        }

                else if ( parse.GetToken() != TOKEN_EQUALS )
                        parse.ErrorIni( "missing = sign on DIGITAL");

                else if ( parse.GetToken() != TOKEN_STRING )
                        parse.ErrorIni( "illegal DIGITAL title");

                else {
                        IniAddDigital( itmp, parse.TOKEN );
                        LastDigital = itmp;
                        IsChannel   = 0;
                        }
                }




        // Check trigger parameters

        // Initialize trigger constants (uses ??,MaxTriggerThreshold,DefaultThreshold)
        // TrgType, TrgThreshold, TrgDigLevelLine, TrgDigEdgeLine,
        // TrgActiveLevel, TrgActiveEdge, TrgActionSave, TrgFilesToSave, 
        // TrgActionSignal, TrgLineToSignal, TrgLatchLine


        else if ( parse.KeywordEquals( "Trigger") ) {

                if (IsChannel == -1)
                        parse.ErrorIni( "missing Channel or Digital keyword before trigger" );

                iok = parse.GetOnOff( &itmp, "Trigger must = {ON,OFF}" );

                if (iok) {
                        if (IsChannel == 1) {
                                if (itmp)
                                        Ini.pScanTrigger[LastChannel] = CHECKED;
                                else
                                        Ini.pScanTrigger[LastChannel] = UNCHECKED;
                                }
                        else { // IsChannel == 0
                                parse.ErrorIni( "Digital channels may not be used for amplitude\ntriggering.  Use digital level or edge triggering instead." );
                                }

                        IsChannel   = -1;
                        LastChannel = -1;
                        LastDigital = -1;
                        }
                }



        else if ( parse.KeywordEquals( "TriggerType" ) ) {

                parse.GetToken();

                if ( parse.Match( "None" ) )
                        Ini.TrgType = TRIG_NONE;

                else if ( parse.Match( "Threshold" ) )
                        Ini.TrgType = TRIG_THRESHOLD;

                else if ( (MaxTriggerDigioLine > 0)          &&
                          (parse.Match( "DigitalLevel" )    ||
                           parse.Match( "DigitalSignal" ) ) )  // old style
                        Ini.TrgType = TRIG_DIGLEVEL;

                else if ( (MaxTriggerDigioLine > 0) &&
                          parse.Match( "DigitalEdge" ) )
                        Ini.TrgType = TRIG_DIGEDGE;

                else {
                        if (MaxTriggerDigioLine > 0)
                                parse.ErrorIni(
                                    "TriggerType must = {None,Threshold,DigitalLevel,DigitalEdge}" );
                        else
                                parse.ErrorIni(
                                    "TriggerType must = {None,Threshold}" );
                        }
                }


        else if ( parse.KeywordEquals( "TriggerThreshold" ) ) {

                iok = parse.GetInt( &itmp, "illegal TriggerThreshold integer value" );

                if ( iok && itmp >= 0 && itmp <= MaxTriggerThreshold )
                        Ini.TrgThreshold = itmp;
                else {
                        sprintf( pAPP->szTemp,
                                 "TriggerThreshold must be 0 to %d",
                                 MaxTriggerThreshold );
                        parse.ErrorIni( pAPP->szTemp );
                        }
                }


        else if ( parse.KeywordEquals( "TriggerDigitalLevelLine" ) ||
                  parse.KeywordEquals( "TriggerDigitalLine"      ) ) { // old style

                iok = parse.GetInt( &itmp, "illegal TriggerDigitalLevelLine integer value" );

                if ( iok && itmp >= 0 && itmp <= MaxTriggerDigioLine )
                        Ini.TrgDigLevelLine = itmp;
                else {
                        sprintf( pAPP->szTemp,
                                 "TriggerDigitalLevelLine must be 0 to %d",
                                 MaxTriggerDigioLine );
                        parse.ErrorIni( pAPP->szTemp );
                        }
                }


        else if ( parse.KeywordEquals( "TriggerDigitalEdgeLine" ) ) {

                iok = parse.GetInt( &itmp, "illegal TriggerDigitalEdgeLine integer value" );

                if ( iok && itmp >= 0 && itmp <= MaxTriggerDigioLine )
                        Ini.TrgDigEdgeLine = itmp;
                else {
                        sprintf( pAPP->szTemp,
                                 "TriggerDigitalEdgeLine must be 0 to %d",
                                 MaxTriggerDigioLine );
                        parse.ErrorIni( pAPP->szTemp );
                        }
                }


        else if ( parse.KeywordEquals( "TriggerActiveLevel" ) ) {

                parse.GetToken();

                if ( parse.Match( "High" ) )
                        Ini.TrgActiveLevel = TRIGLVL_HI;

                else if ( parse.Match( "Low" ) )
                        Ini.TrgActiveLevel = TRIGLVL_LO;
                else
                        parse.ErrorIni("TriggerLevel must = {High,Low}" );
                }


        else if ( parse.KeywordEquals( "TriggerActiveEdge" ) ) {

                parse.GetToken();

                if ( parse.Match( "Rising" ) )
                        Ini.TrgActiveEdge = TRIGEDGE_RISING;

                else if ( parse.Match( "Falling" ) )
                        Ini.TrgActiveEdge = TRIGEDGE_FALLING;
                else
                        parse.ErrorIni("TriggerEdge must = {Rising,Falling}" );
                }


        else if ( parse.KeywordEquals( "TriggerActionSave" ) ) {

                iok = parse.GetOnOff( &itmp, "TriggerActionSave must = {On,Off}" );

                if (iok)
                        Ini.TrgActionSave = itmp;
                }


        else if ( parse.KeywordEquals( "TriggerFilesToSave" ) ) {

                iok = parse.GetInt( &itmp,
                                  "illegal TriggerFilesToSave integer value" );

                if ( iok && itmp > 0 )
                        Ini.TrgFilesToSave = itmp;
                else
                        parse.ErrorIni( "TriggerFilesToSave must be > 0" );
                }


        else if ( parse.KeywordEquals( "TriggerActionSignal" ) ) {

                iok = parse.GetOnOff( &itmp, "TriggerActionSignal must = {On,Off}" );

                if (iok)
                        Ini.TrgActionSignal = itmp;
                }


        else if ( parse.KeywordEquals( "TriggerLineToSignal" ) ) {

                iok = parse.GetInt( &itmp, "illegal TriggerLineToSignal integer value" );

                if ( iok && itmp >= 0 && itmp <= MaxTriggerDigioLine )
                        Ini.TrgLineToSignal = itmp;
                else {
                        sprintf( pAPP->szTemp,
                                 "TriggerLineToSignal must be 0 to %d",
                                 MaxTriggerDigioLine );
                        parse.ErrorIni( pAPP->szTemp );
                        }
                }


        else if ( parse.KeywordEquals( "TriggerLatch" ) ) {

                iok = parse.GetOnOff( &itmp, "TriggerLatch must = {On,Off}" );

                if (iok)
                        Ini.TrgLatchLine = itmp;
                }













        // Check for GPS parameters

        else if ( parse.KeywordEquals( "GpsDriverName" ) ) {

                TokenType = parse.GetToken();
                
                if ( TokenType != TOKEN_STRING && TokenType != TOKEN_SYMBOL)
                        parse.ErrorIni( "bad GpsDriverName string");
                else
                        Ini.GpsDriverName = parse.TOKEN;
                }


        else if ( parse.KeywordEquals( "GpsModelName" )  ) {

                parse.GetToken();

                if ( parse.Match( "GARMIN" ) )

                        Ini.GpsModel = GPSMODEL_GARMIN;

                else if ( parse.Match( "TRIMBLE" ) )

                        Ini.GpsModel = GPSMODEL_TRIMBLE;

                else if ( parse.Match( "ONCORE" ) )

                        Ini.GpsModel = GPSMODEL_ONCORE;

                else if ( parse.Match( "PCTIME" ) )

                        Ini.GpsModel = GPSMODEL_PCTIME;

                else if ( parse.Match( "PARGPS" ) ) {
                        Ini.GpsModel = GPSMODEL_TRIMBLE;
                        parse.ErrorIni(
                           "GpsModel = PARGPS obsolete, use GpsModel = TRIMBLE" );
                        }

                else if ( parse.Match( "DEFAULT" ) )

                        Ini.GpsModel = GPSMODEL_UNKNOWN;

                else
                        parse.ErrorIni( "GpsModelName must = {GARMIN,TRIMBLE,ONCORE,PCTIME,DEFAULT}" );
                }



        else if ( parse.KeywordEquals( "GpsSerialPort" ) ) {

                iok = parse.GetInt( &itmp, "illegal GpsSerialPort integer value" );

                if ( iok && itmp >= 1 && itmp <= 8 )
                        Ini.GpsSerialPort = itmp;
                else
                        parse.ErrorIni( "GpsSerialPort must be 1 to 8" );
                }


        else if ( parse.KeywordEquals( "GpsEnable" ) ) {

                iok = parse.GetOnOff( &itmp, "GpsEnable must = {On,Off}" );

#if !defined( GPS_AVAILABLE )

                if ( itmp == ON ) {
                        parse.ErrorIni( "For GPS support, recompile with GPS_AVAILABLE defined." );
                        itmp = OFF;
                        }
#endif
                if (iok)
                        Ini.GpsEnable = itmp;
                }


        else if ( parse.KeywordEquals( "GpsDisplayPps" ) ) {

                iok = parse.GetOnOff( &itmp, "GpsDisplayPps must = {On,Off}" );

                if (iok)
                        Ini.GpsDisplayPps = itmp;
                }


        else if ( parse.KeywordEquals( "GpsDisplaySerial" ) ) {

                iok = parse.GetOnOff( &itmp, "GpsDisplaySerial must = {On,Off}" );

                if (iok)
                        Ini.GpsDisplaySerial = itmp;
                }




        // Counter keywords

        else if ( parse.KeywordEquals( "CounterEnable" ) ) {

                iok = parse.GetOnOff( &itmp, "CounterEnable must = {On,Off}" );

                if ( itmp == ON && MaxCounterChannel <= 0) {
                        parse.ErrorIni( "Counter channel not supported by hardware." );
                        itmp = OFF;
                        }

                if (iok)
                        Ini.CounterEnable = itmp;
                }








        // Unused options
        // These options are not used in scope at this time, but they
        // are used by the simple acquisition program simp.exe and
        // we want scope to work with simp.ini files.


        else if ( parse.KeywordEquals( "DisplayDataText" )
//                parse.KeywordEquals( "keyword2" ) ||
                  ) {
                parse.GetToken();  // Ignore them
                }





        // Common acquisition keywords are currently checked for in child class
        // AcqGainLog, AcqSPS


        // Not a recognized common keyword

        else
                return( 0 );





        // Many of the other keywords are assumed to be
        // processed from the derived class IniReadKeyword
        // call which should complete after having called this
        // parent class version


        return( 1 );
}

int CScopeApp::IniWriteKeyword( FILE *OutFp ) {

        TRACE0("CScopeApp::IniWriteKeyword\n");

        char *GainTable[] = { "1", "2", "4", "8", "16", "32", "64", "128" };
        int   i;
        char *OutputFileString, *DisplayString, *ColorString;
        char *AutoDCString, *ClippingString, *StatusBarString;
        char *InteractiveString, *AtodEnableString, *AutoRunString;
        char *ForbidMultipleString, *GainString, *OutputFmtString;
        char *TriggerTypeString, *TriggerString, *DigitalEnableString;
        char *TriggerActionSignalString, *TriggerActionSaveString;
        char *GpsEnableString, *GpsDisplayPpsString, *GpsDisplaySerialString;
        char *CounterEnableString, *ExitOnErrorString;
        char *TriggerLatchLineString, *TriggerLevelString, *TriggerEdgeString;
        char *GpsNameString;

        // Set output strings

        AutoRunString             = ( !Ini.DoAutoRun        ? "On" : "Off" );
        ExitOnErrorString         = (  Ini.DoExitOnError    ? "On" : "Off" );
        ForbidMultipleString      = (  Ini.DoForbidMultiple ? "On" : "Off" );
        InteractiveString         = (  Ini.DoInteractive    ? "On" : "Off" );
        StatusBarString           = (  Ini.DoStatusBar      ? "On" : "Off" );
        DisplayString             = (  Ini.DoPlotTrace      ? "On" : "Off" );
        AtodEnableString          = (  Ini.AtodEnable       ? "On" : "Off" );
        ClippingString            = (  Ini.PltClipping      ? "On" : "Off" );
        AutoDCString              = (  Ini.PltAutoDC        ? "On" : "Off" );
        GpsEnableString           = (  Ini.GpsEnable        ? "On" : "Off" );
        GpsDisplayPpsString       = (  Ini.GpsDisplayPps    ? "On" : "Off" );
        GpsDisplaySerialString    = (  Ini.GpsDisplaySerial ? "On" : "Off" );
        DigitalEnableString       = (  Ini.DigitalEnable    ? "On" : "Off" );
        CounterEnableString       = (  Ini.CounterEnable    ? "On" : "Off" );
        TriggerActionSignalString = (  Ini.TrgActionSignal  ? "On" : "Off" );
        TriggerActionSaveString   = (  Ini.TrgActionSave    ? "On" : "Off" );
        TriggerLatchLineString    = (  Ini.TrgLatchLine     ? "On" : "Off" );

        switch ( Ini.TrgType ) {
                case TRIG_NONE:      TriggerTypeString = "None";            break;
                case TRIG_THRESHOLD: TriggerTypeString = "Threshold";       break;
                case TRIG_DIGLEVEL:  TriggerTypeString = "DigitalLevel";    break;
                case TRIG_DIGEDGE:   TriggerTypeString = "DigitalEdge";     break;
                default:             TriggerTypeString = "Unknown";         break;
                }

        switch ( Ini.TrgActiveLevel ) {
                case TRIGLVL_HI:  TriggerLevelString = "High";        break;
                case TRIGLVL_LO:  TriggerLevelString = "Low";         break;
                default:          TriggerLevelString = "Unknown";     break;
                }

        switch ( Ini.TrgActiveEdge ) {
                case TRIGEDGE_RISING:  TriggerEdgeString = "Rising";      break;
                case TRIGEDGE_FALLING: TriggerEdgeString = "Falling";     break;
                default:               TriggerEdgeString = "Unknown";     break;
                }

        switch ( Ini.OutFileFormat ) {
                case FILEFMT_NONE:  OutputFmtString = "None";      break;
                case FILEFMT_OUT:   OutputFmtString = "Out";       break;
                case FILEFMT_DAT:   OutputFmtString = "Dat";       break;
                default:            OutputFmtString = "Unknown";   break;
                }

        switch ( Ini.OutFileNaming ) {
                case FILE_NONE:   OutputFileString = "None";                break;
                case FILE_ONE:    OutputFileString = "Single";              break;
                case FILE_SEQDEC: OutputFileString = "Sequential";          break;
                case FILE_TIME:   OutputFileString = "Time";                break;
                case FILE_STDOUT: OutputFileString = "StdOut";              break;
                case FILE_SEQHEX: // fall through to unknown case
                default:          OutputFileString = "Unknown";             break;
                }

        switch ( Ini.PltColorScheme ) {
                case CS_MONO:   ColorString = "Monochrome"; break;
                case CS_GREEN:  ColorString = "Green";      break;
                case CS_BLUE:   ColorString = "Blue";       break;
                case CS_BROWN:  ColorString = "Brown";      break;
                default:        ColorString = "Unknown";    break;
                }

        switch ( Ini.GpsModel ) {
                case GPSMODEL_GARMIN:  GpsNameString = "Garmin";   break;
                case GPSMODEL_TRIMBLE: GpsNameString = "Trimble";  break;
                case GPSMODEL_ONCORE:  GpsNameString = "Oncore";   break;
                case GPSMODEL_PCTIME:  GpsNameString = "PcTime";   break;
                default:               GpsNameString = "Unknown";  break;
                }



       if ( (MinGainLog <= Ini.AcqGainLog) && (Ini.AcqGainLog <= MaxGainLog) )
                GainString = GainTable[ Ini.AcqGainLog ];
       else
                GainString = "Unknown";



        // Print common acquisition keywords

        fprintf( OutFp, "           SamplingRate = %f\n", Ini.AcqSPS );
        fprintf( OutFp, "                   Gain = %s\n", GainString );
        fprintf( OutFp, "\n" );

        
        // Print GPS keywords

        fprintf( OutFp, "              GpsEnable = %s\n", GpsEnableString );
        fprintf( OutFp, "           GpsModelName = %s\n", GpsNameString );
        fprintf( OutFp, "          GpsDriverName = %s\n", Ini.GpsDriverName );
        fprintf( OutFp, "          GpsSerialPort = %d\n", Ini.GpsSerialPort );
        fprintf( OutFp, "          GpsDisplayPps = OFF\n" );
        fprintf( OutFp, "       GpsDisplaySerial = OFF\n" );
        fprintf( OutFp, "\n" );


        // Print Counter keywords

        fprintf( OutFp, "          CounterEnable = %s\n", CounterEnableString );
        fprintf( OutFp, "\n" );


        // Print Channel keywords

        for ( i = 0 ; i < MaxTotalChannel ; i++ ) {
                if (Ini.pScanActive[i] == CHECKED) {
                        if (Ini.pScanTrigger[i] == CHECKED)
                                TriggerString = "On";
                        else
                                TriggerString = "Off";

                        fprintf(
                                OutFp,
                                "             Channel %2d = \"%s\"   Trigger = %s\n",
                                i,
                                &Ini.pScanTitle[i*MAXTITLE],
                               TriggerString
                              );
                        }
                }
        fprintf( OutFp, "\n" );



        // Digital keywords

        fprintf( OutFp, "          DigitalEnable = %s\n", DigitalEnableString );

        if (Ini.DigitalEnable) {
                for ( i = 0 ; i < MaxDigPlotChannel ; i++ ) {
                        fprintf(
                                OutFp,
                                "             Digital %2d = \"%s\"\n",
                                i,
                                &Ini.pDigTitle[i*MAXTITLE]
                              );
                        } // end for i
                } // end if DigitalEnable
        
        fprintf( OutFp, "\n" );





        // Print Trigger keywords

        fprintf( OutFp, "            TriggerType = %s\n",  TriggerTypeString );
        fprintf( OutFp, "       TriggerThreshold = %ld\n", Ini.TrgThreshold );
        if (MaxTriggerDigioLine >= 0) {
        fprintf( OutFp, "TriggerDigitalLevelLine = %d\n",  Ini.TrgDigLevelLine );
        fprintf( OutFp, "     TriggerActiveLevel = %s\n",  TriggerLevelString );
        fprintf( OutFp, " TriggerDigitalEdgeLine = %d\n",  Ini.TrgDigEdgeLine );
        fprintf( OutFp, "      TriggerActiveEdge = %s\n",  TriggerEdgeString );
        }
        fprintf( OutFp, "      TriggerActionSave = %s\n",  TriggerActionSaveString );
        fprintf( OutFp, "     TriggerFilesToSave = %d\n",  Ini.TrgFilesToSave );
        if (MaxTriggerDigioLine >= 0) {
        fprintf( OutFp, "    TriggerActionSignal = %s\n",  TriggerActionSignalString );
        fprintf( OutFp, "    TriggerLineToSignal = %d\n",  Ini.TrgLineToSignal );
        fprintf( OutFp, "           TriggerLatch = %s\n",  TriggerLatchLineString );
        }
        fprintf( OutFp, "\n" );



        // Print Output File keywords

        fprintf( OutFp, "        OutputDirectory = \"%s\"\n", Ini.OutDir );
        fprintf( OutFp, "       OutputFileFormat = %s\n",     OutputFmtString );
        fprintf( OutFp, "       OutputFileNaming = %s\n",     OutputFileString );
        fprintf( OutFp, "         OutputSeqStart = %d\n",     Ini.OutSeqStart );
        fprintf( OutFp, "           OutputBlocks = %d\n",     Ini.OutNblock );
        fprintf( OutFp, "\n" );


        // Print Display keywords

        fprintf( OutFp, "            DisplayInfo = %s\n", StatusBarString );
        fprintf( OutFp, "       DisplayDataTrace = %s\n", DisplayString );
        fprintf( OutFp, "           DisplayColor = %s\n", ColorString );
        fprintf( OutFp, "           DisplayNbits = %d\n", MaxAmpShift - Ini.PltShiftBy );
        fprintf( OutFp, "          DisplayAutoDC = %s\n", AutoDCString );
        fprintf( OutFp, "        DisplayClipping = %s\n", ClippingString );
        fprintf( OutFp, "     DisplayFirstSample = %d\n", Ini.PltFirstSample );
        fprintf( OutFp, "      DisplayFirstTrace = %d\n", Ini.PltFirstTrace );
        fprintf( OutFp, "         DisplayNtraces = %d\n", Ini.PltNdisplay );
        fprintf( OutFp, "\n" );



        // Print program control keywords

        fprintf( OutFp, "                  Title = \"%s\"\n", Ini.IniTitle );
        fprintf( OutFp, "             AtodEnable = %s\n",     AtodEnableString );
        fprintf( OutFp, "            ExitOnError = %s\n",     ExitOnErrorString );
        fprintf( OutFp, "          KeypressToRun = %s\n",     AutoRunString );
        fprintf( OutFp, "         ForbidMultiple = %s\n",     ForbidMultipleString );
        fprintf( OutFp, "            Interactive = %s\n",     InteractiveString );
        fprintf( OutFp, "\n" );


        
        // Many of the other keywords are assumed to be
        // written from the derived class IniWriteKeyword
        // call which should complete after having called this
        // parent class version

        return( 1 );
}


int CScopeApp::AllocateDataArray( void ) {
        TRACE0("CScopeApp::AllocateDataArray\n");
        return( 0 );
}


// Virtual A/D functions

int CScopeApp::AtodMeasure( int *ival ) {
        TRACE0("CScopeApp::AtodMeasure\n");
        return( 0 );
}

int CScopeApp::AtodInit() {
        TRACE0("CScopeApp::AtodInit\n");
        return( 0 );
}

int CScopeApp::AtodStart() {
        TRACE0("CScopeApp::AtodStart\n");
        return( 0 );
}

int CScopeApp::AtodStop() {
        TRACE0("CScopeApp::AtodStop\n");
        return( 0 );
}

int CScopeApp::AtodAttachGpsTiming() {
        TRACE0("CScopeApp::AtodAttachGpsTiming\n");
        return( 0 );
}

int CScopeApp::AtodReleaseGpsTiming() {
        TRACE0("CScopeApp::AtodReleaseGpsTiming\n");
        return( 0 );
}

int CScopeApp::AtodCheckReady() {
        TRACE0("CScopeApp::AtodCheckReady\n");
        return( 0 );
}

int CScopeApp::AtodCheckOverflow() {
        TRACE0("CScopeApp::AtodCheckOverflow\n");
        return( 0 );
}

int CScopeApp::AtodOverflowRecovery() {
        TRACE0("CScopeApp::AtodOverflowRecovery\n");
        return( 0 );
}

int CScopeApp::AtodGetData() {
        TRACE0("CScopeApp::AtodGetData\n");
        return( 0 );
}

int CScopeApp::AtodDemuxData() {
        TRACE0("CScopeApp::AtodDemuxData\n");
        return( 0 );
}



// Common GPS timing unit functions

int CScopeApp::GpsTimingInit() {

        int errorcode;

       TRACE0("CScopeApp::GpsTimingInit\n");


        // Not using GPS

        if ( !Ini.GpsEnable ) {
                GpsHandle = BAD_DEVHANDLE;
                return( 1 );
                }

        errorcode = 0; // quiet compiler complaints

#ifdef GPS_AVAILABLE

        // Open the GPS driver and initialize the board

        GpsHandle = ParGpsFullOpen(
                               Ini.GpsDriverName.GetBuffer(1),
                               Ini.GpsModel,
                               Ini.GpsSerialPort,
                               NULL,
                               &errorcode
                              );

        Ini.GpsDriverName.ReleaseBuffer();
        if ( GpsHandle == BAD_DEVHANDLE ) {
                ErrString.Format( "PARGPS open failed (%s).",
                                  PARGPS_ERROR_MSG[errorcode] );
                return( 0 );
                }



        // Attach the GPS driver to the A/D driver.

        if ( !AtodAttachGpsTiming() ) {
                ParGpsClose( GpsHandle );
                GpsHandle = BAD_DEVHANDLE;
                return( 0 );
                }

        return( 1 );

#else
        return( 0 );  // Fail since GpsEnable ON but not compiled GPS_AVAILABLE

#endif // GPS_AVAILABLE



}

int CScopeApp::GpsTimingStart() {

        TRACE0("CScopeApp::GpsTimingStart\n");

        // Initialize PPS and serial data counters

        NextptPps     = 0; // current position
        NextptSerial  = 0;

        Npps          = 0; // number read while 1 analog buffer filled
        Nserial       = 0;


#ifdef GPS_AVAILABLE

        // Start the GPS if requested

        if ( Ini.GpsEnable )
                ParGpsStart( GpsHandle );
#endif

        return( 1 );
}

int CScopeApp::GpsTimingStop() {

        TRACE0("CScopeApp::GpsTimingStop\n");


#ifdef GPS_AVAILABLE

        // If the PARGPS was requested, stop it, release it from the
        // A/D, and close the PARGPS driver

        if ( Ini.GpsEnable ) {
                ParGpsStop( GpsHandle );
                AtodReleaseGpsTiming( );
                ParGpsClose( GpsHandle );
                }
#endif

        return( 1 );
}

int CScopeApp::GpsTimingGetReadyData( ) {

        int npts, errorcode;

        TRACE0("CScopeApp::GpsTimingGetReadyData\n");

#define NREMAIN_PPS    ((unsigned int)(Ngpspts-NextptPps))
#define NREMAIN_SERIAL ((unsigned int)(Ngpspts-NextptSerial))


        // Short circuit if GPS not requested

        if (!Ini.GpsEnable)
                return( 1 );

        npts = errorcode = 0; // quiet compiler complaints


#ifdef GPS_AVAILABLE

        // Read any ready GPS PPS data

        npts = ParGpsReadPpsData(
                                 GpsHandle,
                                 &pGpsPpsArray[NextptPps],
                                 NREMAIN_PPS,
                                 &errorcode
                                );
        NextptPps += npts;

        if ( GpsTimingCheckError( errorcode, "ReadPpsData" ) == -1 )
                return( 0 );



        // Read any ready GPS serial data

        npts = ParGpsReadSerialData(
                                    GpsHandle,
                                    &pGpsSerialArray[NextptSerial],
                                    NREMAIN_SERIAL,
                                    &errorcode
                                   );
        NextptSerial += npts;


        if ( GpsTimingCheckError( errorcode, "ReadSerialData" ) == -1 )
                return( 0 );


        
        if ( Ini.GpsModel == GPSMODEL_PCTIME ) {
                GpsNumSat = 0;
                }
        
        else { // Ini.GpsModel == GPSMODEL _TRIMBLE or _GARMIN or _ONCORE
                ParGpsGetCurrentTime( GpsHandle,
                                      NULL,              // double *TimeAsSecsSince1970,
                                      NULL, NULL, NULL,  // int *Yr, *Mon, *Day,
                                      NULL, NULL, NULL,  // int *Hr, *Min, *Sec,
                                      NULL,              // long *MicroSecond,
                                      &GpsNumSat,
                                      NULL, NULL, NULL,  // double *Lat, *Lon, *Alt
                                      NULL, NULL,        // int *SourceYMD, *SourceHMS,
                                      NULL, NULL,        // int *SourceNsat, *SourcePos,
                                      &errorcode
                                    );

                // Ignore error caused by having no NMEA strings as this
                // will occur any time PPS is restored after being lost
                // for awhile.
                
                if ( errorcode == PARGPS_ERROR_BAD_NMEA )
                        GpsNumSat = -1;

                // Fail on all other errors

                else if ( GpsTimingCheckError( errorcode, "ReadCurrentTime" ) == -1 )
                        return( 0 );
                }

        
#endif // GPS_AVAILABLE


        return( 1 );
}

int CScopeApp::GpsTimingCheckError( int errorcode, char *str ) {

        TRACE0("CScopeApp::GpsTimingCheckError\n");

        // Check for an error during PARGPS data read

#ifdef GPS_AVAILABLE

        if (errorcode != PARGPS_ERROR_NONE) {
                ErrString.Format(
                                 "%s failed with error # %d: %s.",
                                 str,
                                 errorcode,
                                 (LPCTSTR) PARGPS_ERROR_MSG[errorcode]
                                );
                LogMessage( MSGLEVEL_ERR, MSGDLG_SHOW, ERROR_AUTO_FATAL, NULL );
                return( -1 );
                }

#endif // GPS_AVAILABLE

        return( 1 );
}
int CScopeApp::GpsTimingCheckBufferSize( BOOL enable, int nsamples, float sps ) {
        
        TRACE0("CScopeApp::GpsTimingCheckBufferSize\n");

        // Check that buffer size for PARGPS driver software FIFO
        // would not be exceeded when GPS is requested


#ifdef GPS_AVAILABLE

        int ngpspts;

        // GPS not requested, so buffer size is not an issue

        if ( !enable )
                return( 1 );

        // Bad sampling rate given (avoid dividing by 0)

        if (sps < 0.00001)
                return( 0 );


        // GPS data comes once per second.  So, to determine the number
        // of points needed in the GPS data arrays, we compute the
        // length of time for one buffer and add 2 for good measure.
        // MAX_PARGPS_STORAGE is the total amount that can be stored in
        // the PARGPS device driver software FIFO without losing data.

        ngpspts = (int)(nsamples / sps) + 2;
        if (ngpspts > MAX_PARGPS_STORAGE)
                return( 0 );
        else
                return( 1 );
        
        
#else // GPS not available, so buffer size is not an issue
        return( 1 );
        
#endif // GPS_AVAILABLE

}

int CScopeApp::TriggerCondition( void ) {

        TRACE0("CScopeApp::TriggerCondition\n");

        // This function is called by the OnTimer function to decide
        // whether or not the currently acquired buffer of data should
        // be written out.  You can also perform A/D specific
        // triggering computations in the TriggerCondition function of
        // the derived class (CA64App, C24bApp, etc) since those
        // classes have a pointer to the raw data array.


        int i, ichan, trig, digactive;


        // Continuous recording

        if (Ini.TrgType == TRIG_NONE)
                return( 1 );


        // Amplitude threshold

        else if (Ini.TrgType == TRIG_THRESHOLD) {

        // Update in progress state based on number of files saved

          if ( (TriggerState == TRIGSTATE_INPROGRESS) &&
               (CurrTriggerFile <= 0)                 )
               TriggerState = TRIGSTATE_ARMED;

          for ( i = 0 ; i < Ini.Nchannels ; i++ ) {

              ichan = Ini.pScanOrder[i];

              // Only check channels selected for triggering

              if (Ini.pScanTrigger[ichan] == CHECKED) {
                      
                  if (Ini.pScanChType[ichan] == CH_TYPE_ANALOG)
                      trig = TriggerCheckAnalog( ichan );
                  else
                       trig = 0;

                  if (trig) {

                      // If already saving, don't count current partial file

                      if (CurrTriggerFile > 0)
                              CurrTriggerFile = Ini.TrgFilesToSave+1;
                      else
                              CurrTriggerFile = Ini.TrgFilesToSave;

                      TriggerState = TRIGSTATE_INPROGRESS;
                      return( 1 );

                      } // end if trig

                  } // end if CHECKED

              } // end for i

          } // end if TRIG_THRESHOLD



        // Digital signal on I/O line level triggered

        else if (Ini.TrgType == TRIG_DIGLEVEL) {

                // Immediate return if this feature not available

                if (Ini.TrgDigLevelLine < 0)
                        return( 0 );

                
                // Update in progress state based on number of files saved

                  if ( (TriggerState == TRIGSTATE_INPROGRESS) &&
                       (CurrTriggerFile <= 0)                 )
                       TriggerState = TRIGSTATE_ARMED;


                // Test digital I/O line

                if ( TriggerDigioCompare(Ini.TrgDigLevelLine, Ini.TrgActiveLevel) ) {

                        // If already saving, don't count current partial file

                        if (CurrTriggerFile > 0)
                                CurrTriggerFile = Ini.TrgFilesToSave+1;
                        else
                                CurrTriggerFile = Ini.TrgFilesToSave;

                        TriggerState = TRIGSTATE_INPROGRESS;
                        return( 1 );

                        } // end Digio Compare

                } // end if TRIG_DIGLEVEL



        // Digital signal on EDGE triggered I/O line (rising or falling)

        else if (Ini.TrgType == TRIG_DIGEDGE) {

                // Immediate return if this feature not available

                if (Ini.TrgDigEdgeLine < 0)
                        return( 0 );

                // Update in progress state based on number of files saved

                if ( (TriggerState == TRIGSTATE_INPROGRESS) &&
                     (CurrTriggerFile <= 0)                 )
                     TriggerState = TRIGSTATE_WAITING;

                
                // Test digital I/O line for matching the active trigger
                // state (ie rising edge is active when high, falling edge
                // is active when low)

                digactive = TriggerDigioCompare( Ini.TrgDigEdgeLine,
                                                 Ini.TrgActiveEdge );


                // Now process the match results based on the current
                // trigger state ...

                
                // We are armed and looking for trigger.  If we find one,
                // setup the number of files to save.

                if (TriggerState == TRIGSTATE_ARMED) {
                        if (digactive) { // triggered
                                TriggerState    = TRIGSTATE_INPROGRESS;
                                CurrTriggerFile = Ini.TrgFilesToSave;
                                return( 1 ); 
                                }
                        else
                                return( 0 );
                        }


                // We are currently processing a trigger.  If we are
                // almost done, we check to see if we've been re-armed.
                // This lets us skip the waiting state so we're ready to
                // trigger again immediately after the current trigger
                // finishes.  In any case, we indicate the current
                // trigger continues.
                
                else if (TriggerState == TRIGSTATE_INPROGRESS) {
                        if ((CurrTriggerFile == 1)         &&    // if last file +
                            (CurrBlock == Ini.OutNblock-1) &&    // last buffer +
                            !digactive)                          // re-armed then
                                TriggerState = TRIGSTATE_ARMED;  // skip waiting
                        return( 1 );
                        }

                
                // We are waiting to be re-armed by seeing the digital
                // line go to the inactive or non-triggered state.  With
                // rising edge triggering, a high value causes a trigger
                // and a low value re-arms.

                else { // (TriggerState == TRIGSTATE_WAITING)
                        if (!digactive)  // re-armed
                                TriggerState = TRIGSTATE_ARMED;
                        
                        return( 0 );     // Never return a trigger when waiting
                        }

                } // end if TRIG_DIGEDGE


        // Unknown trigger type

        else {
                ErrString.Format( "Unknown trigger type %d requested.",
                          Ini.TrgType );
                LogMessage( MSGLEVEL_WARN, MSGDLG_SHOW, ERROR_MAYBE_FATAL, NULL );
                return( 0 );
                }


        // If we haven't already returned from somewhere above
        // because of finding a curent trigger, return a result
        // based on whether or not we still have more "after trigger"
        // files to save.

          if (CurrTriggerFile > 0)
                  return( 1 );
          else
                  return( 0 );
}

int CScopeApp::TriggerAction( void ) {

        TRACE0("CScopeApp::TriggerAction\n");

        // See the appropriate child class for any A/D specific trigger action.
        

        // The standard trigger action is to save files

        if ( Ini.TrgActionSave == TRUE )
                return( WriteData() );


        // Otherwise, just count down files that would have been written
        // so trigger will finish as usual

        else {
                CurrBlock++;
                if (CurrBlock >= Ini.OutNblock) {
                        CurrBlock = 0;
                        if (CurrTriggerFile > 0)
                                CurrTriggerFile--;
                        }
                }
                return( 1 );
}
int CScopeApp::TriggerDone( void ) {

        TRACE0("CScopeApp::TriggerDone\n");

        // See the appropriate child class for any A/D specific trigger cleanup.
        

        // There is no default trigger cleanup that needs to be done.

        return( 1 );

}
int CScopeApp::TriggerClearLatch( void ) {

        TRACE0("CScopeApp::TriggerClearLatch\n");

        // See the appropriate child class for any A/D specific
        // trigger latch clearing.
        

        // There is no default trigger latch clearing to be done.

        return( 1 );

}
int CScopeApp::TriggerCheckAnalog( int ichan ) {

        TRACE0("CScopeApp::TriggerCheckAnalog\n");

        int  j, pt;
        long ThresholdMin, ThresholdMax;


        // Set threshold range, accounting for data format

        TriggerThresholdSetRange( ichan, &ThresholdMin, &ThresholdMax );


        // Check all points for trigger, comparing pt amplitude with threshold

        pt = ichan;
        for ( j = 0 ; j < Ini.Nsamples ; j++ ) {

                if ( TriggerThresholdCompare(pt, ThresholdMin, ThresholdMax) )
                        return( 1 );

                pt += Ini.Nchannels;
                }

        return( 0 );
}
        
int CScopeApp::TriggerThresholdSetRange( int Chan, long *ThMin, long *ThMax ) {

        TRACE0("CScopeApp::TriggerThresholdSetRange\n");

        // This function sets the min and max threshold values.
        // This is a default implementation, but the true work of this
        // virtual function should probably be done by the derived class.

        *ThMin = pDCoffset[Chan] - Ini.TrgThreshold;
        *ThMax = pDCoffset[Chan] + Ini.TrgThreshold;

        return( 1 );
}

int CScopeApp::TriggerDigitalCompare( int Pt, long TriggerMask ) {

        TRACE0("CScopeApp::TriggerDigitalCompare\n");

        // This function compares data point pt with the supplied mask
        // and returns 1 if any of the matching bits are set, indicating
        // a trigger for that digital bit.  The true work of this
        // virtual function should be done by the derived class since it
        // has a pointer to the raw data array.

        return( 1 );
}

int CScopeApp::TriggerThresholdCompare( int Pt, long ThMin, long ThMax ) {

        TRACE0("CScopeApp::TriggerThresholdCompare\n");

        // This function compares data point pt with the supplied
        // threshold and returns 1 if the absolute amplitude of the
        // point exceeds the threshold.  The true work of this virtual
        // function should be done by the derived class since it has a
        // pointer to the raw data array.

        return( 1 );
}

int CScopeApp::TriggerDigioCompare( int Line, int ActiveState ) {

        TRACE0("CScopeApp::TriggerDigioCompare\n");

        // This function compares the signal on the specified digital
        // I/O line and returns 1 if the signal matches the specified
        // state.  The true work of this virtual function should be done
        // by the derived class since it has the function to query the
        // digital I/O line.

        return( 1 );
}

int CScopeApp::HdrFill( void ) {

        TRACE0("CScopeApp::HdrFill\n");

        int  i, ichan, j;
        long cftemphi, cftemplo, cftemp2hi, cftemp2lo;


        // Fill the specified header type with appropriate default values.


        switch ( Ini.OutFileFormat ) {

          case FILEFMT_OUT:

            // Initialize the .OUT file .HDR general header info

            for ( i = 0 ; i < 32 ; i++ )
                    OutHdr.Id[i] = '\0';

            OutHdr.Channels    = Ini.Nchannels;
            OutHdr.RecordPts   = Ini.Nsamples * Ini.OutNblock;
            OutHdr.WordSize    = DataPtSize;
            OutHdr.DataCoding  = 1;

            OutHdr.SampleRate  = Ini.AcqSPS;
            OutHdr.Gain        = 1.0;
            OutHdr.FilterCoeff = 0.0;
            OutHdr.FilterScale = 1.0;


            // Fill channel list

            for ( i = 0 ; i < Ini.Nchannels ; i++ ) {

                    ichan = Ini.pScanOrder[i];
                            
// WCT: Check if this is required for A64 version
//                    OutHdr.ChannelList[i].channel = ichan;

                    OutHdr.ChannelList[i].channel = i;
                    SrStringCopy(
                                 OutHdr.ChannelList[i].title,
                                 &Ini.pScanTitle[ichan*MAXTITLE],
                                 MAXTITLE
                                );

                    } // end for i

            break;


          case FILEFMT_DAT:

            // Initialize .DAT file 4096 byte header.
            // Fill with generic defaults, then with PARxCH specific values.

            SrDatHdrInit( &DatHdr );

            for ( i = 0 ; i < SRDAT_DIM_STRING ; i++ ) {
                    DatHdr.DatProduct[i] = '\0';
                    DatHdr.DatApplication[i] = '\0';
                    }

            if (Ini.GpsEnable)
                    DatHdr.DatGpsModel = Ini.GpsModel;
            else
                    DatHdr.DatGpsModel = GPSMODEL_NONE;


            DatHdr.FileNameType      = Ini.OutFileNaming;
            if (Ini.OutFileNaming == FILE_SEQDEC ||
                Ini.OutFileNaming == FILE_SEQHEX )
                    DatHdr.FileNameInc = 1;
            else if (Ini.OutFileNaming == FILE_TIME)
                    DatHdr.FileNameInc = (long)(Ini.OutNblock *
                                                SingleBufferTime + 0.5);
            
            DatHdr.Channels       = Ini.Nchannels;
            DatHdr.ChannelsAnalog = MaxAnalogChannel;

            if (Ini.DigitalEnable  &&  MaxDigitalChannel > 0) {
                    DatHdr.ChannelsDigital  = MaxDigitalChannel;
                    DatHdr.DigitalPerChan   = MaxDigPlotChannel;
                    DatHdr.ChannelDigIndex  = DigitalIndex;
                    }
            else {
                    DatHdr.ChannelsDigital  = 0;
                    DatHdr.DigitalPerChan   = 0;
                    DatHdr.ChannelDigIndex  = SRDAT_INDEXCHANNEL_NONE;
                    }

            if (Ini.GpsEnable  &&  MaxGpsChannel > 0) {
                    DatHdr.ChannelsGps      = MaxGpsChannel;
                    DatHdr.ChannelMarkIndex = GpsMarkIndex;
                    }
            else {
                    DatHdr.ChannelsGps      = 0;
                    DatHdr.ChannelMarkIndex = SRDAT_INDEXCHANNEL_NONE;
                    }

            if (Ini.CounterEnable  &&  MaxCounterChannel > 0) {
                    DatHdr.ChannelsCounter  = MaxCounterChannel;
                    DatHdr.ChannelObcIndex  = CounterIndex;
                    }
            else {
                    DatHdr.ChannelsCounter  = 0;
                    DatHdr.ChannelObcIndex  = SRDAT_INDEXCHANNEL_NONE;
                    }



            DatHdr.ChannelPtsPerRecord = Ini.Nsamples;
            DatHdr.ChannelPtsPerFile   = Ini.Nsamples * Ini.OutNblock;

            DatHdr.AtodWordSize        = DataPtSize;
            DatHdr.AtodDataCoding      = SRDAT_DATACODING_OFFSET;
            DatHdr.AtodSamplingRate    = Ini.AcqSPS;
            DatHdr.AtodGain            = 1.0;
            DatHdr.AtodFilterCoeff     = 0.0;
            DatHdr.AtodFilterScale     = 1.0;


            // Fill channel list

            for ( i = 0 ; i < Ini.Nchannels ; i++ ) {

                    ichan = Ini.pScanOrder[i];
                    
//                  WCT: Required for A64 version?
//                  DatHdr.ChannelList[i].Channel = ichan;

                    DatHdr.ChannelList[i].Channel = i;
                    SrStringCopy(
                                 DatHdr.ChannelList[i].Title,
                                 &Ini.pScanTitle[ichan*SRDAT_DIM_TITLE],
                                 SRDAT_DIM_TITLE
                                );

                    if (Ini.pScanChType[i] == CH_TYPE_DIGITAL) {
                            for ( j = 0 ; j < MaxDigPlotChannel ; j++ ) {
                               DatHdr.DigitalList[j].Channel = j;
                               SrStringCopy(
                                       DatHdr.DigitalList[j].Title,
                                       &Ini.pDigTitle[j*MAXTITLE],
                                       SRDAT_DIM_TITLE );

                                    } // end for j

                            } // end if CH_TYPE_DIGITAL

                    } // end for i





            DatHdr.TriggerType = Ini.TrgType;
            if (Ini.TrgType == TRIG_THRESHOLD) {
                    DatHdr.TriggerParm1 = Ini.TrgThreshold;
                    DatHdr.TriggerParm2 = Ini.pScanTrigger[Ini.Nchannels-1];
                    for ( i = 2 ; i < Ini.Nchannels ; i++ ) {
                            DatHdr.TriggerParm2 *= 2; // shift left 1 bit
                            DatHdr.TriggerParm2 |= Ini.pScanTrigger[Ini.Nchannels-i];
                            }
                    }
            else if (Ini.TrgType == TRIG_DIGLEVEL) {
                    DatHdr.TriggerParm1 = Ini.TrgDigLevelLine;
                    DatHdr.TriggerParm2 = Ini.TrgActiveLevel;
                    }
            else if (Ini.TrgType == TRIG_DIGEDGE) {
                    DatHdr.TriggerParm1 = Ini.TrgDigEdgeLine;
                    DatHdr.TriggerParm2 = Ini.TrgActiveEdge;
                    }
            

            cftemphi  = cftemplo  = 0L;     // quiet compiler complaints
            cftemp2hi = cftemp2lo = 0L;

#ifdef GPS_AVAILABLE

            // Set GPS counter frequency.

            if ( Ini.GpsEnable &&
                 ParGpsGetFullCounterFrequency( GpsHandle,
                                                &cftemphi, &cftemplo,
                                                &cftemp2hi, &cftemp2lo) ) {

                    DatHdr.CounterFreqHi  = cftemphi;
                    DatHdr.CounterFreq    = cftemplo;
                    DatHdr.Counter2FreqHi = cftemp2hi;
                    DatHdr.Counter2Freq   = cftemp2lo;
			}
#endif

            break;



          default:

            // No header requested

            break;


          } // end switch

        return( 1 );
}


int CScopeApp::WriteBytes( CStdioFile *OutFile ) {
        TRACE0("CScopeApp::WriteBytes\n");
        return( 0 );
}

int CScopeApp::HdrWrite( void ) {

        TRACE0("CScopeApp::HdrWrite\n");

        CString    HdrFileName;
        CStdioFile HdrFile;
        int        realsize;



        // No output desired

        if ( (Ini.OutFileFormat == FILEFMT_NONE) ||
             (Ini.OutFileNaming == FILE_NONE)    )
                return( TRUE );



        // Write the .OUT header structure out to a binary .HDR file.
        //
        // This function only pertains to the .OUT header structure which
        // is written to a single separate .HDR file once.
        //
        // See the HdrUpdateWrite function for .DAT file headers which are
        // written at the beginning of each .DAT file.
        //
        //

        if (Ini.OutFileFormat != FILEFMT_OUT)
                return( FALSE );




        // Output the header file

        HdrFileName = Ini.OutDir;
        HdrFileName += "\\SCOPE.HDR";

        if (!HdrFile.Open( HdrFileName, CFile::modeCreate |
                                        CFile::modeReadWrite |
                                        CFile::shareExclusive, NULL ) ) {

                ErrString.Format("Could not output header file %s.",
                                                        (LPCTSTR) HdrFileName);
                LogMessage( MSGLEVEL_WARN, MSGDLG_SHOW, ERROR_MAYBE_FATAL, NULL );
                return( FALSE );
                }

        else {

                // This computation computes the real header size
                // by removing the allocated ChannelList portion
                // and then only including back the actual maximum
                // number of channels.  This allows allocating
                // enough space to service either the DSPA64 or the
                // PAR24B, but to only write out the pertinent space.

                realsize = (sizeof(OutHdr) - sizeof(OutHdr.ChannelList)) +
                           Ini.Nchannels * sizeof(struct channel_data);

                HdrFile.Write( &OutHdr, realsize );

                HdrFile.Flush();
                HdrFile.Close();

                return( TRUE );

                }

}


int CScopeApp::HdrUpdateWrite( void ) {

        TRACE0("CScopeApp::HdrUpdateWrite\n");

        CFileStatus   Status;
        struct tm     utc_time;
        LPTSTR        fn;
        int           nstr;
        LARGE_INTEGER tmp64;


        // No output desired

        if ( (Ini.OutFileFormat == FILEFMT_NONE) ||
             (Ini.OutFileNaming == FILE_NONE)    )
                return( TRUE );



        // Update a .DAT header and write it out.
        //
        // This function pertains only to .DAT headers which are written
        // at the beginning of each .DAT file.
        //
        //

        if (Ini.OutFileFormat != FILEFMT_DAT)
                return( FALSE );






        // Update the DAT header structure with current information.

        nstr = FinalFileName.GetLength();
        if (nstr < SRDAT_DIM_STRING)
                strncpy( DatHdr.FileName, FinalFileName, SRDAT_DIM_STRING-1 );
        else {
                fn = FinalFileName.GetBuffer( 1 );
                nstr = nstr - SRDAT_DIM_STRING + 1;
                strncpy( DatHdr.FileName, &fn[nstr], SRDAT_DIM_STRING-1 );
                FinalFileName.ReleaseBuffer();
                }

        // Get file "create" time.  The st_ctime value doesn't seem
        // to be properly updated when re-creating a file, so use the
        // "modify" time instead since we've just opened it for creation.

        if ( TmpFile.GetStatus( Status ) == TRUE) {
                if ( sizeof(Status.m_mtime.GetTime()) <= 4) { // returning time_t
                        DatHdr.FileCreateTimeHi    = 0L;
                        DatHdr.FileCreateTime      = (long)Status.m_mtime.GetTime();
                        }
                else {                                        // returning time64_t
                        tmp64.QuadPart = (LONGLONG)Status.m_mtime.GetTime();
                        DatHdr.FileCreateTimeHi    = tmp64.HighPart;
                        DatHdr.FileCreateTime      = tmp64.LowPart;
                        }
                }

        DatHdr.FileSeqNum                  = Ini.OutSeqStart;
        DatHdr.FileStartPtNum              = NpointsFromStart;
        DatHdr.TimePoint[0].Type           = SRDAT_TIME_TYPE_FIRSTPT;
        DatHdr.TimePoint[0].SampleFromRun  = NpointsFromStart;
        DatHdr.TimePoint[0].SampleFromFile = 0;
        DatHdr.TimePoint[0].GpsMarkIndex   = 0;


        // Split time into YMD and HMS using the "Local" function
        // since AcquireTime already has the GMT correction applied

        AcquireTime.GetLocalTm( &utc_time );

        DatHdr.TimePoint[0].Year           = utc_time.tm_year + 1900;
        DatHdr.TimePoint[0].Month          = utc_time.tm_mon + 1;
        DatHdr.TimePoint[0].Day            = utc_time.tm_mday;
        DatHdr.TimePoint[0].YMDSource      = TimeSource;
        DatHdr.TimePoint[0].Hour           = utc_time.tm_hour;
        DatHdr.TimePoint[0].Minute         = utc_time.tm_min;
        DatHdr.TimePoint[0].Second         = utc_time.tm_sec;
        DatHdr.TimePoint[0].Microsecond    = AcquireTimeUsec;
        DatHdr.TimePoint[0].HMSSource      = TimeSource;


        // Write out the updated header.

        if ( !SrDatHdrWrite( TmpFile.m_pStream, &DatHdr ) ) {
                ErrString = "Could not write updated .DAT header.";
                LogMessage( MSGLEVEL_ERR, MSGDLG_SHOW, ERROR_MAYBE_FATAL, NULL );
                return( FALSE );
                }


        return( TRUE );
}


void CScopeApp::TmpFileComputeFinalName( void ) {

        TRACE0("CScopeApp::TmpFileComputeFinalName\n");

        char *Suffix;

        FileError = 0;


        // Set file suffix based on the requested data format.

        if (Ini.OutFileFormat == FILEFMT_OUT)
                Suffix = "OUT";

        else // if (Ini.OutFileFormat == FILEFMT_DAT)
                Suffix = "DAT";



        // Compute the final file name using the requested naming mode.

        // NOTE: For FILE_TIME naming, ReadyTime should be already set
        // to the time of the first point in the file to be named.

        if (Ini.OutFileNaming == FILE_ONE)

                ShortFileName.Format( "SCOPE.%s", Suffix );

        else if (Ini.OutFileNaming == FILE_SEQDEC)

                ShortFileName.Format( "%08ld.%s", Ini.OutSeqStart, Suffix );

        else if (Ini.OutFileNaming == FILE_TIME)

                ShortFileName.Format( "%08lX.%s", ReadyTime.GetTime(), Suffix );


        else  // Ini.OutFileNaming == NONE or STDOUT

                ShortFileName.Format( "" );


        // Add directory part of name

        if ( ShortFileName.IsEmpty() )
                FinalFileName.Format( "" );
        else
                FinalFileName.Format( "%s\\%s",
                                      (LPCTSTR)Ini.OutDir,
                                      (LPCTSTR)ShortFileName );
}

int CScopeApp::TmpFileOpen( CStdioFile *TmpFile ) {

        TRACE0("CScopeApp::TmpFileOpen\n");

        int  iret;
        UINT OpenFlags;


        FileError = 0;


        // Standard out not supported.

        if ( Ini.OutFileNaming == FILE_STDOUT )
                return( 0 );



        // If this is the first block, it must be created when opening.
        // Otherwise don't truncate so we can append.

        if (CurrBlock == 0)
                OpenFlags = CFile::modeCreate;
        else
                OpenFlags = CFile::modeCreate | CFile::modeNoTruncate;



        // Open the tmp file.

        iret = TmpFile->Open( TmpFileName,
                              CFile::shareExclusive |
                              OpenFlags |
                              CFile::typeBinary |
                              CFile::modeReadWrite, NULL);

        if (!iret) {
                ErrString.Format("Could not output data file %s.",
                                                   (LPCTSTR)TmpFileName);
                LogMessage( MSGLEVEL_ERR, MSGDLG_SHOW, ERROR_MAYBE_FATAL, NULL );
                FileError = 1;
                return( 0 );
                }



        // Position at the end.

        TmpFile->SeekToEnd( );


        return( 1 );

}
void CScopeApp::TmpFileClose( CStdioFile *TmpFile ) {

        TRACE0("CScopeApp::TmpFileClose\n");

        // Standard out not supported.

        if ( Ini.OutFileNaming == FILE_STDOUT )
                return;


        // Close tmp file.

        TmpFile->Flush( );
        TmpFile->Close( );

}
int CScopeApp::TmpFileRename( void ) {

        TRACE0("CScopeApp::TmpFileRename\n");

        CFileStatus Status;


        FileError = 0;


        // Standard out not supported.

        if ( Ini.OutFileNaming == FILE_STDOUT )
                return( 1 );



        // Verify that final name is available (ie when GetStatus fails
        // because a file of that name does not exist yet).

        if ( CFile::GetStatus( FinalFileName, Status ) == TRUE) {

                ErrString.Format("Rename failed because %s already exists.",
                        (LPCTSTR)FinalFileName );
                TmpString.Format("\nCould not rename %s to \n%s.\n\n",
                        (LPCTSTR)TmpFileName, (LPCTSTR)FinalFileName );
                TmpString += "Please remove existing *.OUT or *.DAT files before running again.\n\n";
                TmpString += "Output File Naming Conventions:\n\n";
                TmpString += "'One' = name every output file scope.out (or scope.dat).  Use this\n";
                TmpString += "when a second process runs at the same time as scope, removing each\n";
                TmpString += "scope.out or scope.dat file as it is created.\n\n";
                TmpString += "'Time' = use time stamp file names with 1s resolution.  Fails at high\n";
                TmpString += "sampling rates where files are produced faster than a 1s rate.\n\n";
                TmpString += "'Sequential' = assigns sequential numbers to successive .out or\n";
                TmpString += ".dat files.  Numbering starts at the Seq Start Num on each run.  Works at\n";
                TmpString += "all sampling rates, but be sure to remove files between runs.\n";
                LogMessage( MSGLEVEL_ERR, MSGDLG_SHOW, ERROR_MAYBE_FATAL, &TmpString );
                FileError = 1;
                return( 0 );
                }



        // Rename the .tmp file.

        TRY {
                CFile::Rename( TmpFileName, FinalFileName );
                }
        CATCH( CFileException, e ) {
                ErrString.Format("Failed rename of %s because ",
                        (LPCTSTR)FinalFileName );
                if (e->m_cause == CFileException::accessDenied)
                        ErrString += "file access denied.";
                else if (e->m_cause == CFileException::fileNotFound)
                        ErrString += "file not found.";
                else if (e->m_cause == CFileException::badPath)
                        ErrString += "of bad file path.";
                TmpString  = "\nPlease remove existing files before running again.  ";
                TmpString += "You may also want to select a different output file ";
                TmpString += "naming convention from the Output Options dialog.";
                LogMessage( MSGLEVEL_ERR, MSGDLG_SHOW, ERROR_MAYBE_FATAL, &TmpString );
                FileError = 1;
                return( 0 );
                }
        END_CATCH

        return( 1 );

}
int CScopeApp::TmpFileDelete( void ) {

        TRACE0("CScopeApp::TmpFileDelete\n");

        CFileStatus Status;


        FileError = 0;


        // Standard out not supported.

        if ( Ini.OutFileNaming == FILE_STDOUT )
                return( 1 );



        // Delete the .tmp file.

        TRY {
                CFile::Remove( TmpFileName );
                }
        CATCH( CFileException, e ) {
                ErrString.Format("Failed delete of %s because ",
                        (LPCTSTR)TmpFileName );
                if (e->m_cause == CFileException::accessDenied)
                        ErrString += "file access denied.";
                else if (e->m_cause == CFileException::fileNotFound)
                        ErrString += "file not found.";
                else if (e->m_cause == CFileException::badPath)
                        ErrString += "of bad file path.";
                LogMessage( MSGLEVEL_ERR, MSGDLG_SHOW, ERROR_MAYBE_FATAL, NULL );
                FileError = 1;
                return( 0 );
                }
        END_CATCH

        return( 1 );

}
int CScopeApp::WriteData( void ) {

        TRACE0("CScopeApp::WriteData\n");

        // No output desired, but track block and file count for trigger

        if ( (Ini.OutFileFormat == FILEFMT_NONE) ||
             (Ini.OutFileNaming == FILE_NONE)    ) {

                CurrBlock++;
                if (CurrBlock >= Ini.OutNblock) {
                        CurrBlock = 0;
                        if (CurrTriggerFile > 0)
                                CurrTriggerFile--;
                        }

                return( TRUE );
                }



        // Open tmp file and get file handle.

        TmpFileName = Ini.OutDir;
        TmpFileName += "\\SCOPE.TMP";

        if ( !TmpFileOpen( &TmpFile ) )
                return( FALSE );




        // Before first buffer, update and write the header (if .DAT
        // files requested).  The final file name is now computed just
        // after the previous file was written.

        if (CurrBlock == 0)
                HdrUpdateWrite( );






        // Save analog data (subclass checks for requested format).

        WriteBytes( &TmpFile );



        // For .DAT files, save additional GPS data and add end of file record.

        if (Ini.OutFileFormat == FILEFMT_DAT) {

                if (Npps > 0)
                        if ( !SrDatRecordWrite( TmpFile.m_pStream,
                                                SRDAT_TAGID_GPSPPST,
                                                pGpsPpsArray,
                                                Npps * sizeof(PPSTDATA) ) ) {
                                ErrString = "Could not write .DAT PPS(T) data.";
                                LogMessage( MSGLEVEL_ERR, MSGDLG_SHOW, ERROR_MAYBE_FATAL, NULL );
                                return( FALSE );
                                }

                if (Nserial > 0)
                        if ( !SrDatRecordWrite( TmpFile.m_pStream,
                                                SRDAT_TAGID_GPSSERIAL,
                                                pGpsSerialArray,
                                                Nserial * sizeof(SERIALDATA) ) ) {
                                ErrString = "Could not write .DAT serial data.";
                                LogMessage( MSGLEVEL_ERR, MSGDLG_SHOW, ERROR_MAYBE_FATAL, NULL );
                                return( FALSE );
                                }


                if (CurrBlock >= Ini.OutNblock-1) // CurrBlock not inc'd yet
                        if ( !SrDatRecordWrite( TmpFile.m_pStream,
                                                SRDAT_TAGID_EOF,
                                                NULL,
                                                0 )  ) {
                                ErrString = "Could not write .DAT EOF.";
                                LogMessage( MSGLEVEL_ERR, MSGDLG_SHOW, ERROR_MAYBE_FATAL, NULL );
                                return( FALSE );
                                }

                }




        // Close tmp file

        TmpFileClose( &TmpFile );




        // Increment the number of blocks processed.

        CurrBlock++;



        // Rename file depending on chosen naming scheme

        if (CurrBlock >= Ini.OutNblock) {

                if ( !TmpFileRename( ) ) {

                        // If scope can't rename the tmp file, it was
                        // deleted as part of cleaning up.  This was
                        // changed on 03/25/2005 so the user will have a
                        // chance to recover the data manually when scope
                        // stops due to the error.
                        //
                        // WCT: There may be some situations when
                        // running automatically with DoExitOnError off
                        // where this is not the best behavior.  This code
                        // should be changed again if those situtations can
                        // be clearly identified.
                        //
                        // int SaveFileError 
                        // SaveFileError = FileError;
                        // TmpFileDelete();
                        // FileError = SaveFileError;
                        
                        CurrBlock = 0;
                        return( FALSE );
                        }

                CurrBlock = 0;
                if (CurrTriggerFile > 0)
                        CurrTriggerFile--; // Files output since trigger
                Ini.OutSeqStart++;         // Next output file name number


                // Prepare the final filename for the NEXT file, using the
                // ReadyTime of the last buffer in the file just written
                // as an approximation of the start time of the first point
                // in the next file.  This is a change from versions of the
                // code before 6/30/03 that computed the file name using
                // the time at the END of the first buffer in the next file.

                TmpFileComputeFinalName( );
                }



        return( TRUE );

}


int CScopeApp::ShowHelp( int titleID, int textID, CWnd *caller ) {

        int iret;

        CTopicDlg dlg(titleID, textID, m_pMainWnd);
        iret = dlg.DoModal();

        if (caller)
                caller->SetFocus();

        return( iret );
}

int CScopeApp::LoadHelpTitles( CListBox *lbx ) {

        int pos;

        // Load general help titles at default location

        pos = AddResourceString(lbx, IDR_HELP_MAIN_TITLE,   IDR_HELP_MAIN_TEXT   );
        pos = AddResourceString(lbx, IDR_HELP_STATUS_TITLE, IDR_HELP_STATUS_TEXT    );
        pos = AddResourceString(lbx, IDR_HELP_INI_TITLE,    IDR_HELP_INI_TEXT    );
        pos = AddResourceString(lbx, IDR_HELP_HARD_TITLE,   IDR_HELP_HARD_TEXT   );
        pos = AddResourceString(lbx, IDR_HELP_CHAN_TITLE,   IDR_HELP_CHAN_TEXT   );
        pos = AddResourceString(lbx, IDR_HELP_ACQ_TITLE,    IDR_HELP_ACQ_TEXT    );
        pos = AddResourceString(lbx, IDR_HELP_OUTPUT_TITLE, IDR_HELP_OUTPUT_TEXT );
        pos = AddResourceString(lbx, IDR_HELP_TRIG_TITLE,   IDR_HELP_TRIG_TEXT   );
        pos = AddResourceString(lbx, IDR_HELP_DISP_TITLE,   IDR_HELP_DISP_TEXT   );
        pos = AddResourceString(lbx, IDR_HELP_GPS_TITLE,    IDR_HELP_GPS_TEXT    );

        return( pos );
}

int CScopeApp::AddResourceString( CListBox *listbox,
                                  int titleID, int textID, int pos /*=-1*/ ) {

        CString htitle;

        htitle = (LPCSTR)::LockResource(
                         ::LoadResource( NULL,
                         ::FindResource( NULL,
                            (LPCTSTR)titleID, RT_RCDATA)) );

        if (pos == -1)
                pos = listbox->AddString( htitle );
        else
                pos = listbox->InsertString( pos, htitle );

        listbox->SetItemData( pos, MAKELONG(titleID,textID) );

        return( pos );
}

void CScopeApp::WriteLogFile( CString *LogMessage ) {

        CString    ErrFileName;
        CStdioFile ErrFile;
        
        // Open and write message to the log file.

        ErrFileName.Format( "%s\\SCOPE.ERR", Ini.OutDir );
        if ( ErrFile.Open( ErrFileName,
                              CFile::modeCreate |
                              CFile::modeNoTruncate |
                              CFile::modeWrite |
                              CFile::shareDenyNone |
                              CFile::typeText, NULL ) ) {

                ErrFile.SeekToEnd();
                ErrFile.WriteString( (LPCTSTR)(*LogMessage) );
                ErrFile.Flush();
                ErrFile.Close();
                }
}

void CScopeApp::LogMessage( int MsgLevel, BOOL ShowDialog, int Fatal, CString *ExtraMsg ) {

        char *datestr, dbuffer[16];
        char *timestr, tbuffer[16];
        char *levelstr;
        time_t     current_time;
        struct tm *local_time;
        CString    MsgString;
        BOOL       KillProgram;

        CScopeWindow *pWnd = (CScopeWindow*)m_pMainWnd;
        

        // Get the date and time.

        time( &current_time );
        local_time = localtime( &current_time );

        strftime( dbuffer, 16, "%m/%d/%y", local_time );
        strftime( tbuffer, 16, "%H:%M:%S", local_time );

        datestr = dbuffer;
        timestr = tbuffer;

        
        // Prepare message string

        if (MsgLevel == MSGLEVEL_MSG)
                levelstr = "Message";
        else if (MsgLevel == MSGLEVEL_WARN)
                levelstr = "Warning";
        else if (MsgLevel == MSGLEVEL_ERR)
                levelstr = "Error";
        else
                levelstr = "Note";

        MsgString.Format( "%s %s %-7s = %s\n",
                          datestr, timestr, levelstr, ErrString );


        WriteLogFile( &MsgString );
        

        // If requested (and not running automatically with no keypress),
        // also write the message to a dialog box.  Scope should stop
        // before displaying the dialog since the user might take a
        // long time to respond and we don't want to build up lots of
        // other errors in the meantime.

        if ( ShowDialog && !Ini.DoAutoRun ) {
                if (ExtraMsg)
                        MsgString.Format( "%s occurred on %s at %s\n\n%s\n%s\n",
                                          levelstr, datestr, timestr,
                                          ErrString, *ExtraMsg );
                else
                        MsgString.Format( "%s occurred on %s at %s\n\n%s\n",
                                          levelstr, datestr, timestr, ErrString );
                
                if (IsRunning && pWnd)
                        pWnd->OnStopProgram();
                AfxMessageBox( (LPCTSTR)MsgString );
                if (pWnd)
                        pWnd->Invalidate();
                }



        // The program needs to be killed, in the following situations:
        //   1) the code requests it by sending in Fatal as ERROR_FATAL
        //   2) we are running automatically with no keypress (so there
        //      is no user to press Ctrl+R for a restart) and the code
        //      requests it by sending in Fatal as ERROR_AUTO_FATAL
        //   3) we are running automatically with no keypress and the 
        //      user requested exiting on errors from the ini file and 
        //      this is an error (eg we are at ERR level for this message)
        // It should NOT be killed in the following cases:
        //   1) the code requests it by sending in Fatal as ERROR_NOT_FATAL
        //   2) we are running automatically and the user requested
        //      not to exit on errors
        //   3) we are running with keypresses required (eg not
        //      automatically)
        //   4) we are in any case not covered by all the above
        //

        if ( Fatal == ERROR_FATAL )
                KillProgram = TRUE;
        
        else if ( Fatal == ERROR_NOT_FATAL  )
                KillProgram = FALSE;
        
        else if ( Fatal == ERROR_AUTO_FATAL  ) {
                if ( Ini.DoAutoRun )
                        KillProgram = TRUE;
                else
                        KillProgram = FALSE;
                }

        else if ( Fatal == ERROR_MAYBE_FATAL  ) {
                if ( Ini.DoAutoRun && Ini.DoExitOnError )
                        KillProgram = TRUE;
                else
                        KillProgram = FALSE;
                }
        else
                KillProgram = FALSE;


        // If error is fatal, then destroy window and quit

        if ( KillProgram ) {
                
                ErrCode = 1;

                if (IsRunning && pWnd)
                        pWnd->OnStopProgram();

                MsgString.Format( "%s %s Message = Ending scope on error.\n",
                          datestr, timestr );
                WriteLogFile( &MsgString );
                
                if (m_pMainWnd != NULL)
                        m_pMainWnd->DestroyWindow();
                else
                        AfxPostQuitMessage( 0 );

                }

}



//***********************************************************************
//
// CLASS CScopeWindow
//
// This class, derived from CFrameWnd, contains the generic main window
// portion of the Scope data acquisition program.  It provides the
// graphics drawing, menu operations, and other main window functions
// for the CScopeApp class which deals with acquisition, files, and
// ini.  Window functions specific to the A64, 24B, and 8CH versions are
// taken care of in the derived classes CA64Win, C24bWin, and C8chWin.
//

BEGIN_MESSAGE_MAP( CScopeWindow, CFrameWnd )
        //{{AFX_MSG_MAP(CScopeWindow)
        ON_WM_CREATE()
        ON_WM_CLOSE()
        ON_WM_SIZE()
        ON_WM_PAINT()
        ON_WM_HSCROLL()
        ON_WM_VSCROLL()
        ON_WM_DESTROY()
        ON_WM_KEYDOWN()
        ON_WM_KEYUP()
        ON_WM_MOUSEMOVE()
        ON_WM_TIMER()
        ON_COMMAND(IDM_PROGRAM_RUN,         OnRunProgram)
        ON_COMMAND(IDM_PROGRAM_STOP,        OnStopProgram)
        ON_COMMAND(IDM_PROGRAM_EXIT,        OnClose)
        ON_COMMAND(IDM_OPTIONS_INIREAD,     OnIniRead)
        ON_COMMAND(IDM_OPTIONS_INIWRITE,    OnIniWrite)
        ON_COMMAND(IDM_OPTIONS_HARDWARE,    OnHardwareDialog)
        ON_COMMAND(IDM_OPTIONS_CHANNEL,     OnChannelDialog)
        ON_COMMAND(IDM_OPTIONS_ACQUIRE,     OnAcquireDialog)
        ON_COMMAND(IDM_OPTIONS_TRIGGER,     OnTriggerDialog)
        ON_COMMAND(IDM_OPTIONS_OUTPUT,      OnOutputDialog)
        ON_COMMAND(IDM_OPTIONS_DISPLAY,     OnDisplayDialog)
        ON_COMMAND(IDM_OPTIONS_GPS,         OnGpsDialog)
        ON_COMMAND(IDM_OPTIONS_CLIP,        OnSetClip)
        ON_COMMAND(IDM_OPTIONS_AUTODC,      OnSetAutoDC)
        ON_COMMAND(IDM_OPTIONS_LATCH,       OnSetTriggerLatch)
        ON_COMMAND(IDM_OPTIONS_UPDATEDC,    OnSetDC)
        ON_COMMAND( ID_VIEW_STATUS_BAR,     OnSetStatusBar)
        ON_COMMAND(IDM_VIEW_PLOTTRACE,      OnSetPlotTrace)
        ON_COMMAND(IDM_HELP_CONTENTS,       OnContentsDialog)
        ON_COMMAND(IDM_HELP_ABOUT,          OnAboutDialog)
        ON_COMMAND(IDM_HELP_DEBUG,          OnDebug)
        ON_UPDATE_COMMAND_UI(IDM_PROGRAM_STOP,        OnUpdateMenuRunning)
        ON_UPDATE_COMMAND_UI(IDM_PROGRAM_RUN,         OnUpdateMenuStopped)
        ON_UPDATE_COMMAND_UI(IDM_OPTIONS_INIREAD,     OnUpdateMenuStopped)
        ON_UPDATE_COMMAND_UI(IDM_OPTIONS_INIWRITE,    OnUpdateMenuStopped)
        ON_UPDATE_COMMAND_UI(IDM_OPTIONS_HARDWARE,    OnUpdateMenuStopped)
        ON_UPDATE_COMMAND_UI(IDM_OPTIONS_CHANNEL,     OnUpdateMenuStopped)
        ON_UPDATE_COMMAND_UI(IDM_OPTIONS_ACQUIRE,     OnUpdateMenuStopped)
        ON_UPDATE_COMMAND_UI(IDM_OPTIONS_TRIGGER,     OnUpdateMenuStopped)
        ON_UPDATE_COMMAND_UI(IDM_OPTIONS_OUTPUT,      OnUpdateMenuStopped)
        ON_UPDATE_COMMAND_UI(IDM_OPTIONS_DISPLAY,     OnUpdateMenuStopped)
        ON_UPDATE_COMMAND_UI(IDM_OPTIONS_GPS,         OnUpdateMenuStopped)
        ON_UPDATE_COMMAND_UI(IDM_OPTIONS_CLIP,        OnUpdateClipCheck)
        ON_UPDATE_COMMAND_UI(IDM_OPTIONS_LATCH,       OnUpdateTriggerLatch)
        ON_UPDATE_COMMAND_UI(IDM_OPTIONS_AUTODC,      OnUpdateAutoDCCheck)
        ON_UPDATE_COMMAND_UI(IDM_VIEW_PLOTTRACE,      OnUpdatePlotTraceCheck)
        ON_UPDATE_COMMAND_UI(IDM_HELP_CONTENTS,       OnUpdateMenuStopped)
        ON_UPDATE_COMMAND_UI(IDM_HELP_ABOUT,          OnUpdateMenuStopped)
        //}}AFX_MSG_MAP
        ON_MESSAGE(WM_ENTERSIZEMOVE, OnEnterSizeMove)
        ON_MESSAGE(WM_EXITSIZEMOVE,  OnExitSizeMove)
END_MESSAGE_MAP()

static UINT indicators[] = { ID_SEPARATOR,       // status line indicator
                             IDS_PANE_COORDS,    // for mouse coords
                             IDS_PANE_INFO       // for sps, bits, nsat, etc
                           };


CScopeWindow::CScopeWindow() {

        TRACE0("CScopeWindow::CScopeWindow\n");


        // Initialize some variables

        pYclipmin  = pYclipmax = pYlast = pYaxis = NULL;
        pPtArray   = pOldArray = NULL;
        CurrFont   = NULL;
        PenBack    = PenAxis = PenTitle = PenTrace = NULL;
        IsDragging = FALSE;


        // Initialize dialog positions

        InitDialogPositions();


        // Create and load icon, menus, accelerators, etc

        LoadFrame( IDR_SCOPE, WS_OVERLAPPEDWINDOW | WS_VSCROLL | WS_HSCROLL );


        TRACE0("CScopeWindow::CScopeWindow - END\n");
}

CScopeWindow::~CScopeWindow() {

        TRACE0("CScopeWindow::~CScopeWindow\n");
}

BOOL CScopeWindow::PreCreateWindow(CREATESTRUCT& cs) {

        TRACE0("CScopeWindow::PreCreateWindow\n");

        // Modify default size

        cs.x  = GUTTER;
        cs.y  = GUTTER;
        cs.cx = pAPP->DefaultScreenWidth  - GUTTER - cs.x;
        cs.cy = pAPP->DefaultScreenHeight - GUTTER - cs.y - HSTARTBAR;

        return( CFrameWnd::PreCreateWindow(cs) );

}

void CScopeWindow::PostNcDestroy() {

        TRACE0("CScopeWindow::PostNcDestroy\n");

        // do nothing at this level, let child class delete "this"
}

int CScopeWindow::OnCreate(LPCREATESTRUCT lpCreateStruct) {

        TRACE0("CScopeWindow::OnCreate\n");

        int nsize, yppi, apoint, PaneSize;
        UINT PaneId, PaneStyle;


        // Create parent class

        if (CFrameWnd::OnCreate(lpCreateStruct) == -1) {
                pAPP->ErrString = "CFrameWnd creation failed.";
                pAPP->LogMessage( MSGLEVEL_ERR, MSGDLG_SHOW, ERROR_FATAL, NULL );
                return -1;
                }

        // Add status bar

        if (!m_wndStatusBar.Create(this) ||
                !m_wndStatusBar.SetIndicators(indicators,
                  sizeof(indicators)/sizeof(UINT))) {

                pAPP->ErrString = "Status bar creation failed.";
                pAPP->LogMessage( MSGLEVEL_ERR, MSGDLG_SHOW, ERROR_FATAL, NULL );
                return -1;      // fail to create
                }


        // Make first pane a fixed size

        m_wndStatusBar.GetPaneInfo( PANE_MESSAGE, PaneId, PaneStyle, PaneSize );
        PaneStyle = (PaneStyle & ~SBPS_STRETCH);
        m_wndStatusBar.SetPaneInfo( PANE_MESSAGE, PaneId, PaneStyle, PaneSize );


        // Make second pane a fixed size too

        PaneStyle = SBPS_NORMAL;
        m_wndStatusBar.SetPaneInfo( PANE_COORDS, IDS_PANE_COORDS, PaneStyle, 120 );
        m_wndStatusBar.SetPaneText( PANE_COORDS, "( , )", TRUE );


        // Make last pane stretchy

        PaneStyle = SBPS_NORMAL | SBPS_STRETCH;
        m_wndStatusBar.SetPaneInfo( PANE_INFO, IDS_PANE_INFO, PaneStyle, 400 );
        m_wndStatusBar.SetPaneText( PANE_INFO, "Scope info", TRUE );


        // Set default colors, create pens

        SetColors();

        CreatePens();



        // Compute some important point sizes (2,8,20) as pixels

        CClientDC dc(this);

        yppi = dc.GetDeviceCaps( LOGPIXELSY );  // Y Pixels Per logical Inch

        minPoint = (yppi *  2) / 72;
        midPoint = (yppi *  8) / 72;
        maxPoint = (yppi * 20) / 72;

        apoint   = (yppi * 16) / 72 - 1;




        // Set up a font

        font.lfHeight         = 0;      // apoint;
        font.lfWidth          = 0;
        font.lfEscapement     = 0;
        font.lfOrientation    = 0;
        font.lfWeight         = FW_NORMAL; // tried FW_BOLD;
        font.lfItalic         = FALSE;
        font.lfUnderline      = FALSE;
        font.lfStrikeOut      = FALSE;
        font.lfCharSet        = ANSI_CHARSET;
        font.lfOutPrecision   = OUT_DEFAULT_PRECIS;
        font.lfClipPrecision  = CLIP_DEFAULT_PRECIS;
        font.lfQuality        = DEFAULT_QUALITY;
        font.lfPitchAndFamily = FIXED_PITCH | FF_MODERN; // was VARIABLE_PITCH | FF_SWISS;

        SrStringCopy(font.lfFaceName, "CourierNew", LF_FACESIZE); // was Arial





        // Initialize plot arrays and variables

        LastPtsPlot = 0;
        pPtArray    = NULL;
        pOldArray   = NULL;




        // Allocate vertical plot position buffers

        nsize = pAPP->MaxAllocTotalPlotChannel * sizeof( int );

        pYlast    = (int*) malloc( nsize );
        pYaxis    = (int*) malloc( nsize );
        pYclipmin = (int*) malloc( nsize );
        pYclipmax = (int*) malloc( nsize );


        if (!pYlast || !pYaxis || !pYclipmin || !pYclipmax) {

                pAPP->ErrString = "Could not allocate vertical position buffers.";
                pAPP->LogMessage( MSGLEVEL_ERR, MSGDLG_SHOW, ERROR_FATAL, NULL );
                return -1;  // failed create
                }





        // If status bar is to be off, call OnBarCheck to toggle it
        // since the status bar starts off visible.  
        // (See UpdateStatusBar for a way to set it rather than toggle it)

        if (!pAPP->Ini.DoStatusBar)
                CFrameWnd::OnBarCheck(AFX_IDW_STATUS_BAR);



        // Compute plot positions (and allocate plot buffers if needed)

        if (!ComputePlotPositions())
                return -1;      // failed create



        // Set scroll bar ranges

        SetScrollBar( SB_BOTH );

        controlkey = FALSE;



        // Set title and pane string

        SetWindowTitle( );
        SetPaneString( );


        // Write start up message to log, but don't show dialog

        pAPP->ErrString = "Beginning scope.";
        pAPP->LogMessage( MSGLEVEL_MSG, MSGDLG_NO_SHOW, ERROR_NOT_FATAL, NULL );

        return 0;
}

void CScopeWindow::OnSize(UINT nType, int cx, int cy) {

        TRACE0("CScopeWindow::OnSize\n");

        CFrameWnd::OnSize(nType, cx, cy);

        if (!ComputePlotPositions())
                return;

}

void CScopeWindow::OnHScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar) {

        TRACE0("CScopeWindow::OnHScroll\n");

        int first, line, page, maxstart;

        first    = pAPP->Ini.PltFirstSample;
        maxstart = pAPP->NsamplesPlot - TraceWidth - 1;
        page     = TraceWidth;
        line     = maxstart / 20;

        if (line < 5)
                line = 5;
        if (maxstart < 0)
                maxstart = 0;


        switch (nSBCode) {

                case SB_PAGEUP:
                        first -= page;
                        break;

                case SB_PAGEDOWN:
                        first += page;
                        break;

                case SB_LINEUP:
                        first -= line;
                        break;

                case SB_LINEDOWN:
                        first += line;
                        break;

                case SB_TOP:
                        first = 0;
                        break;

                case SB_BOTTOM:
                        first = maxstart;
                        break;

                case SB_THUMBPOSITION:
                        first = nPos;
                        break;
                }


        // Ensure range is valid

        if (first < 0)
                first = 0;

        if (first > maxstart)
                first = maxstart;




        // Update if changed

        if (first != pAPP->Ini.PltFirstSample) {

                SetScrollPos( SB_HORZ, first, TRUE );

                pAPP->Ini.PltFirstSample = first;

                Invalidate();

                }


}

void CScopeWindow::OnVScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar) {

        TRACE0("CScopeWindow::OnVScroll\n");

        int first, maxstart;

        first    = pAPP->Ini.PltFirstTrace;
        maxstart = pAPP->NchannelsPlot - pAPP->Ini.PltNdisplay;

        switch (nSBCode) {

                case SB_PAGEUP:
                        first -= pAPP->Ini.PltNdisplay;
                        break;

                case SB_PAGEDOWN:
                        first += pAPP->Ini.PltNdisplay;
                        break;

                case SB_LINEUP:
                        first--;
                        break;

                case SB_LINEDOWN:
                        first++;
                        break;

                case SB_TOP:
                        first = 0;
                        break;

                case SB_BOTTOM:
                        first = maxstart;
                        break;

                case SB_THUMBPOSITION:
                        first = nPos;
                        break;
                }


        // Ensure range is valid

        if (first < 0)
                first = 0;

        if (first > maxstart)
                first = maxstart;



        // Update if changed

        if (first != pAPP->Ini.PltFirstTrace) {

                SetScrollPos( SB_VERT, first, TRUE );

                pAPP->Ini.PltFirstTrace = first;
                pAPP->ResetPaint        = TRUE;

                Invalidate();

                }


}

void CScopeWindow::OnPaint() {

        TRACE0("CScopeWindow::OnPaint\n");

        // No paint when minimized

        if ( IsIconic() )
                return;


        // Delay paint when dragging for a move or resize

        if ( IsDragging ) {
                PaintDeferred = TRUE;
                return;
                }


        // Only listen to DoPlotTrace flag when running

        if (!pAPP->Ini.DoPlotTrace && pAPP->IsRunning)
                return;



        // Redraw plot

        CPaintDC dc(this); // device context for painting

        DrawBackground( &dc );

        DrawTitles( &dc );

        DrawScales( &dc );

        DrawAxes( &dc );




        // Check for overflow

        if (pAPP->IsRunning && pAPP->AtodCheckOverflow())
                return;



        // Redraw traces

        pAPP->ResetPaint = TRUE;

        if (pAPP->IsAllocated)
                DrawTraces( &dc );



        // Do not call CFrameWnd::OnPaint() for painting messages
}

LRESULT CScopeWindow::OnEnterSizeMove( WPARAM wParam, LPARAM lParam ) {

        TRACE0("CScopeWindow::OnEnterSizeMove\n");

        // Prevent any painting while dragging for move/resize

        IsDragging = TRUE;
        return LRESULT(0);
}


LRESULT CScopeWindow::OnExitSizeMove( WPARAM wParam, LPARAM lParam ) {

        TRACE0("CScopeWindow::OnExitSizeMove\n");

        // Drag for move/resize is done, so allow painting and
        // force a paint if any have been skipped during the drag.

        IsDragging = FALSE;
        if (PaintDeferred) {
                PaintDeferred = FALSE;
                Invalidate(FALSE);
        }
        return LRESULT(0);
}


void CScopeWindow::SetClipLimits( int height, int traceHalf ) {

        TRACE0("CScopeWindow::SetClipLimits\n");

        int i;

        if (pAPP->Ini.PltClipping)
                for ( i = 0 ; i < pAPP->MaxTotalPlotChannel ; i++ ) {
                        pYclipmin[i] = pYaxis[i] - traceHalf;
                        pYclipmax[i] = pYaxis[i] + traceHalf;
                        }

        else
                for ( i = 0 ; i < pAPP->MaxTotalPlotChannel ; i++ ) {
                        pYclipmin[i] = 0;
                        pYclipmax[i] = height-1;
                        }
}


void CScopeWindow::DrawBackground( CDC *pDC ) {

        TRACE0("CScopeWindow::DrawBackground\n");

        int height, width;

        GetSize( &height, &width );

        pDC->FillSolidRect( 0, 0, width+1, height+1, ColorBack );
}


void CScopeWindow::DrawTitles( CDC *pDC ) {

        TRACE0("CScopeWindow::DrawTitles\n");

        int i, ch2, itrace;


        // clear title area


        // draw titles

        pDC->SelectObject( CurrFont );
        pDC->SetTextColor( ColorTitle );
        pDC->SetBkMode(TRANSPARENT);

        ch2    = cyChar/2;
        itrace = pAPP->Ini.PltFirstTrace;

        for ( i = 0 ;  i < pAPP->Ini.PltNdisplay ; i++  ) {

                pDC->TextOut( HMARGIN, pYaxis[i]-ch2,
                             &(pAPP->pPlotTitle[itrace*MAXTITLE]) );
                itrace++;
                }

}


void CScopeWindow::DrawScales( CDC *pDC ) {

        TRACE0("CScopeWindow::DrawScales\n");

        int i, j, k, xpos, nthick, nax, yax[2], sign[2];
        int tic, xinc, ticsm, xincsm, nsmall;


        pDC->SelectObject( PenAxis );


        // Draw extra top and bottom axes with tics

        yax[0] = YScaleTop;     sign[0] =  1;
        yax[1] = YScaleBottom;  sign[1] = -1;
        nax=2;
        
        nthick   = 3;
        nsmall   = 4;
        
        xinc     = 100;
        xincsm   = xinc / (nsmall+1);
        
        tic      = VTICK;
        ticsm    = tic  / 2;


        for ( i = 0 ; i < nax ; i++ ) {

                // Draw line

                pDC->MoveTo( TraceStart, yax[i] );
                pDC->LineTo( TraceEnd+1, yax[i] );


                // Draw tics

                xpos = TraceStart;
                while ( xpos < TraceEnd+1 ) {

                        // Large tic

                        for ( j = 0 ; j < nthick ; j++ ) {

                                pDC->MoveTo( xpos+j, yax[i] );
                                pDC->LineTo( xpos+j, yax[i]+sign[i]*tic );
                                
                                } // end for j < nthick

                        // Small tics

                        for ( k = 0 ; k < nsmall ; k++ ) {

                                xpos += xincsm;

                                for ( j = 0 ; j < nthick ; j++ ) {

                                        pDC->MoveTo( xpos+j, yax[i] );
                                        pDC->LineTo( xpos+j, yax[i]+sign[i]*ticsm );

                                        } // end for j < nthick

                                } // end for k < nsmall
                        
                        xpos += xincsm;

                        } // end while

        } // end for i > nax

}


void CScopeWindow::DrawAxes( CDC *pDC ) {

        TRACE0("CScopeWindow::DrawAxes\n");

        int i;


        pDC->SelectObject( PenAxis );


        // Draw horizontal trace axis lines
        
        for ( i = 0 ; i < pAPP->Ini.PltNdisplay ; i++ ) {

                pDC->MoveTo( TraceStart, pYaxis[i] );
                pDC->LineTo( TraceEnd+1, pYaxis[i] );

                }
}


void CScopeWindow::DrawTraces( CDC *pDC ) {

        TRACE0("CScopeWindow::DrawTraces\n");

        int npoints, j, k, kend, height, width;



        // Correct number of points based on TraceWidth and max plot points

        kend = pAPP->Ini.PltFirstSample + TraceWidth;

        if (kend > pAPP->NsamplesPlot)
                kend = pAPP->NsamplesPlot;

        npoints = kend - pAPP->Ini.PltFirstSample;




        // Fill plot arrays from data arrays with clipping

        FillTraces( npoints );




        // Blankout and draw new traces

        k = 0;

        if (pAPP->ResetPaint == TRUE) {

                pAPP->ResetPaint = FALSE;

                GetSize( &height, &width );

                pDC->FillSolidRect(TraceStart,0,TraceWidth,height,ColorBack);

                DrawScales( pDC );

                DrawAxes( pDC );

                pDC->SelectObject( PenTrace );

                for ( j = 0 ; j < pAPP->Ini.PltNdisplay ; j++ ) {

                        pDC->Polyline( &(pPtArray[k]), npoints );

                        k += npoints;

                        }
                }



        // Or erase old and draw new traces

        else {

                for ( j = 0 ; j < pAPP->Ini.PltNdisplay ; j++ ) {

                        pDC->SelectObject( PenBack );
                        pDC->Polyline( &(pOldArray[k]), npoints );


                        pDC->SelectObject( PenAxis );
                        pDC->MoveTo( TraceStart, pYaxis[j] );
                        pDC->LineTo( TraceEnd+1, pYaxis[j] );


                        pDC->SelectObject( PenTrace );
                        pDC->Polyline( &(pPtArray[k]), npoints );

                        k += npoints;

                        }

                DrawScales( pDC );

                }

}

void CScopeWindow::FillTraces( int npoints ) {

        TRACE0("CScopeWindow::FillTraces\n");

        int kstart, jstart, jend, jtrace, tracehalf;
        int i, j, k, kdata, joff, xpos, xval;
        int ycurr, yprev, yval, yscaled, yax, yclipmin, yclipmax;



        // Prepare point array with all points (some new and some old)

        k      = 0;
        jtrace = pAPP->Ini.PltFirstTrace;
        kstart = pAPP->Iplotbuff * pAPP->Ini.Nsamples + pAPP->Ini.PltFirstSample;

        if (kstart > pAPP->NsamplesPlot)
                kstart -= pAPP->NsamplesPlot;

        tracehalf = TraceHeight / 2;

        for ( j = 0 ; j < pAPP->Ini.PltNdisplay ; j++ ) {

                xpos     = TraceStart;
                yprev    = 0;
                yax      = pYaxis[j];
                yclipmin = pYclipmin[j];
                yclipmax = pYclipmax[j];

                joff     = pAPP->pDCoffset[jtrace];
                jstart   = jtrace*pAPP->NsamplesPlot;
                jend     = jstart + pAPP->NsamplesPlot;
                kdata    = jstart + kstart;



                for ( i = 0 ; i < npoints ; i++ ) {

                    // Fill old trace points

                    pOldArray[k].x = pPtArray[k].x;
                    pOldArray[k].y = pPtArray[k].y;

                    // Process analog channels
                    
                    // Note: GpsCounter channel should never be turned on,
                    // since we prohibit it when setting up the plot array.

                    if ( (pAPP->pPlotChType[jtrace] == CH_TYPE_ANALOG) ||
                         (pAPP->pPlotChType[jtrace] == CH_TYPE_GPSCOUNTER) ) {
                        
                        // Remove DC offset

                        yval = (pAPP->pDemuxArray[kdata] - joff);


                        // Scale, rounding towards zero as per DOS scope

                        if (yval >= 0)
                                yscaled = yax - (yval >> pAPP->Ini.PltShiftBy);
                        else
                                yscaled = yax + (-yval >> pAPP->Ini.PltShiftBy);


                        // Set x value
                        
                        xval = xpos;
                        
                        } // end analog and gps counter channels

                    
                    // Process digital and GPS mark channels

                    else { // CH_TYPE_DIGITAL || CH_TYPE_GPSMARK

                            // Set digital y value to zero or half height

                            if (pAPP->pDemuxArray[kdata] == 0) {
                                    ycurr   = 0;
                                    yscaled = yax;
                                    }
                            else {
                                    ycurr   = 1;
                                    yscaled = yax - tracehalf;
                                    }


                            // Set or correct x values to avoid sloping lines

                            if (ycurr == yprev) {      // flat section
                                    xval = xpos;
                                    }
                            else if (ycurr > yprev) {  // step up
                                    xval = xpos;
                                    if (i != 0)
                                            pPtArray[k-1].x = xpos;
                                    }
                            else { // ycurr < yprev    // step down
                                    xval = xpos-1;
                                    }

                            yprev = ycurr;
                        
                        } // end digital and GPS mark channels



                    // Clip

                    if (yscaled < yclipmin)
                            yscaled = yclipmin;

                    else if (yscaled > yclipmax)
                            yscaled = yclipmax;



                    // Fill new trace points

                    pPtArray[k].x = xval;
                    pPtArray[k].y = yscaled;


                    // Move to next point

                    xpos++;
                    k++;
                    kdata++;
                    if (kdata >= jend)
                            kdata = jstart;

                    } // end for i < npoints

                jtrace++;

                } // end for j < PltNdisplay

}


void CScopeWindow::OnDestroy() {

        TRACE0("CScopeWindow::OnDestroy\n");


        // Clean up pens

        DeletePens();



        // Clean up last font

        DeleteFont();



        // Clean up plot arrays

        DeletePlotArrays();



        // Clean up vertical plot position buffers

        DeleteVerticalBuffers();





        CFrameWnd::OnDestroy();
}


void CScopeWindow::OnMouseMove( UINT nFlags, CPoint point ) {
        
        CString PaneString;
        char    *d;
        int     i, x, t, y, traceHalf;

        traceHalf = TraceHeight / 2;

        // Determine which trace we are closest to
        
        t = pAPP->Ini.PltNdisplay - 1;
        
        for ( i = 0 ; i < pAPP->Ini.PltNdisplay ; i++ ) {
                if ( point.y < pYaxis[i] - traceHalf ) {
                        t = i - 1;
                        break;
                        }
                }

        if ( t < 0 )
                t = 0;


        // Determine x, y, trace

        x  = point.x - TraceStart;
        if ( x < 0 )
                x = 0;
        else if ( x > TraceWidth )
                x = TraceWidth;
        
        d  = ">";
        y  = pYaxis[t] - point.y;
        if (y < 0) {
                d = "<";
                y = -y;
                }
        
        t += pAPP->Ini.PltFirstTrace;
        
        PaneString.Format( "( %04d, %03d %1s %-2d )", x, y, d, t );

        m_wndStatusBar.SetPaneText( PANE_COORDS, (LPCTSTR)PaneString, TRUE );
        
}

void CScopeWindow::OnTimer(UINT nIDEvent) {

        TRACE0("CScopeWindow::OnTimer\n");

        int        readyResult;
        time_t     ebtime;  // this is a long
        double     microtime;
        SYSTEMTIME systime;


        // Check overflow

        if ( pAPP->AtodCheckOverflow() ) {
                pAPP->ErrString = "A/D Overflow occurred.";
                pAPP->LogMessage( MSGLEVEL_ERR, MSGDLG_SHOW, ERROR_MAYBE_FATAL, NULL );
                if (pAPP->Ini.DoAutoRun && !pAPP->Ini.DoExitOnError)
                        pAPP->AtodOverflowRecovery();
                return;
                }





        // Check ready buffer and process if ready

        readyResult = pAPP->AtodCheckReady();

        if (readyResult > 0) {

                // Save the time when this complete buffer was ready
                // (which is approximately the start of the NEXT buffer).

                GetSystemTime( &systime );  // in UTC
                pAPP->ReadyTime = systime;  // "local" now has UTC correction


                pAPP->AtodGetData();
                pAPP->AtodDemuxData();   // demux data and compute offset

                
                // increment plot buff index for just read data

                if (pAPP->Iplotbuff >= pAPP->Nplotbuffers-1)
                        pAPP->Iplotbuff = 0;
                else
                        pAPP->Iplotbuff++;


                // write data if trigger condition is satisfied

                if (pAPP->TriggerCondition() == 1) {

                        pAPP->IsTriggered = TRUE;

                        if (!pAPP->TriggerAction()) { // default is WriteData()
                                if (!pAPP->Ini.DoAutoRun || pAPP->Ini.DoExitOnError) {
                                        OnStopProgram();
                                        return;
                                        }
                                } // end if Trigger Action
                        } // end if Trigger Condition

                // do cleanup if changing from triggered to non-triggered

                else {
                        if (pAPP->IsTriggered) //  this is now really WasTriggered
                                pAPP->TriggerDone();

                        pAPP->IsTriggered = FALSE;
                        }





                // check overflow

                if ( pAPP->AtodCheckOverflow() ) {
                        pAPP->ErrString = "A/D Overflow occurred after write.";
                        pAPP->LogMessage( MSGLEVEL_ERR, MSGDLG_SHOW, ERROR_MAYBE_FATAL, NULL );
                        if (pAPP->Ini.DoAutoRun && !pAPP->Ini.DoExitOnError)
                                pAPP->AtodOverflowRecovery();
                        return;
                        }



                // draw traces

                if (pAPP->Ini.DoPlotTrace) {
                        CClientDC dc(this);
                        DrawTraces( &dc );
                        SetWindowTitle();
                        SetPaneString();
                        }


                // check overflow

                if ( pAPP->AtodCheckOverflow() ) {
                        pAPP->ErrString = "A/D Overflow occurred after draw.";
                        pAPP->LogMessage( MSGLEVEL_ERR, MSGDLG_SHOW, ERROR_MAYBE_FATAL, NULL );
                        if (pAPP->Ini.DoAutoRun && !pAPP->Ini.DoExitOnError)
                                pAPP->AtodOverflowRecovery();
                        return;
                        }



                // Roughly update the next buffer time.  For a more accurate
                // value, use the GPS data to compute time.

                pAPP->NpointsFromStart  += pAPP->Ini.Nsamples;
                pAPP->TimeSource         = SRDAT_SOURCE_ACQ;
                pAPP->ElapsedBufferTime += pAPP->SingleBufferTime;
                ebtime                   = (time_t)pAPP->ElapsedBufferTime;
                microtime                = (pAPP->ElapsedBufferTime - ebtime) *
                                           USPERSEC;
                pAPP->AcquireTimeUsec    = pAPP->StartTimeUsec + (long)microtime;
                
                while (pAPP->AcquireTimeUsec > USPERSEC) {
                        pAPP->AcquireTimeUsec -= USPERSEC;
                        ebtime++;
                        }
                
                pAPP->AcquireTime        = pAPP->StartTime + ebtime;

                } // end if readyResult > 0 


        // Data error detected when calling AtodCheckReady
        
        else if (readyResult == -1) {
                pAPP->ErrString = "Data error requiring acquisition restart.";
                pAPP->LogMessage( MSGLEVEL_ERR, MSGDLG_SHOW, ERROR_MAYBE_FATAL, NULL );
                if (pAPP->Ini.DoAutoRun && !pAPP->Ini.DoExitOnError)
                        pAPP->AtodOverflowRecovery();
                return;
                }





        CFrameWnd::OnTimer(nIDEvent);  // calls default which calls defwinproc
}



void CScopeWindow::OnRunProgram() {

        TRACE0("CScopeWindow::OnRunProgram\n");

        int        interval;
        SYSTEMTIME systime;


        // Initialize A/D (and GPS) board

        if ( ! pAPP->AtodInit() ) {

                pAPP->IsAllocated = FALSE;      // stops trace draw
                pAPP->LogMessage( MSGLEVEL_ERR, MSGDLG_SHOW, ERROR_AUTO_FATAL, &pAPP->TmpString );
                return;

                }




        // Allocate A/D (and GPS) buffers

        if ( ! pAPP->AllocateBuffers() )
                return;




        // Compute timer interval

        if (pAPP->Ini.AcqSPS > 0.0001)
                interval = (int)(MSPERSEC*pAPP->Ini.Nsamples/
                                                (pAPP->Ini.AcqSPS*TICKINBUFF));
        else
                interval = 20;



        // Set counters for number of blocks and trigger files,
        // number of points, and time for one buffer

        pAPP->NpointsFromStart = 0;
        pAPP->CurrBlock        = 0;
        pAPP->CurrTriggerFile  = 0;
        if (pAPP->Ini.TrgType == TRIG_DIGEDGE)
                pAPP->TriggerState     = TRIGSTATE_WAITING;
        else
                pAPP->TriggerState     = TRIGSTATE_ARMED;

        if (pAPP->Ini.AcqSPS > 0.0001)
                pAPP->SingleBufferTime = pAPP->Ini.Nsamples / pAPP->Ini.AcqSPS;
        else
                pAPP->SingleBufferTime = 0.0L;




        // Output header

        pAPP->HdrFill();
        pAPP->HdrWrite();



        // Begin A/D (and GPS) execution

        pAPP->AtodStart();

        GetSystemTime( &systime );  // in UTC


        // Get approximate start time from the PC and set first file name

        pAPP->TimeSource        = SRDAT_SOURCE_ACQ;
        pAPP->StartTime         = systime; // "local" now has UTC correction
        pAPP->StartTimeUsec     = (systime.wMilliseconds * 1000);
        pAPP->AcquireTime       = pAPP->StartTime;
        pAPP->AcquireTimeUsec   = pAPP->StartTimeUsec;
        pAPP->ReadyTime         = pAPP->StartTime;
        pAPP->ElapsedBufferTime = 0.0L;
        pAPP->TmpFileComputeFinalName( );


        // Enable timer

        if ( ! SetTimer( IDT_TIMER, interval, NULL ) ) {

                pAPP->AtodStop();
                pAPP->ErrString = "Not enough timers.  Close some other apps.";
                pAPP->LogMessage( MSGLEVEL_ERR, MSGDLG_SHOW, ERROR_AUTO_FATAL, NULL );
                return;
                }


        pAPP->IsRunning = TRUE;


        // Set title and pane string

        SetWindowTitle( );
        SetPaneString( );
}


void CScopeWindow::OnStopProgram() {

        TRACE0("CScopeWindow::OnStopProgram\n");

        if (pAPP->IsRunning) {

                pAPP->AtodStop();
                KillTimer(IDT_TIMER);

                pAPP->IsRunning = FALSE;
                SetWindowTitle( );
                SetPaneString( );
                }
}

void CScopeWindow::OnKeyUp(UINT nChar, UINT nRepCnt, UINT nFlags) {

        TRACE0("CScopeWindow::OnKeyUp\n");


        // Keep track of control key

        if (nChar == VK_CONTROL)

                controlkey = FALSE;




        // Default handling

        CFrameWnd::OnKeyUp(nChar, nRepCnt, nFlags);
}

void CScopeWindow::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags) {

        TRACE0("CScopeWindow::OnKeyDown\n");

        int scancode;




        // Keep track of control key

        if (nChar == VK_CONTROL)

                controlkey = TRUE;



        // Translate scan codes into a reliable virtkey value

        scancode = nFlags & 0x00FF;

        if ( (scancode == 0x0C) || (scancode == 0x33) ) // -_ or ,< keys
                nChar = VK_SUBTRACT;

        if ( (scancode == 0x0D) || (scancode == 0x34) ) // =+ or .> keys
                nChar = VK_ADD;




        switch (nChar) {

            // Increase amplitude (decrease shift)

            case 0xBB:          // =+ key
            case 0xBE:          // .> key
            case VK_ADD:        // numpad + key
            case VK_UP:         // up arrow key

                pAPP->Ini.PltShiftBy -= nRepCnt;

                if (pAPP->Ini.PltShiftBy < pAPP->MinAmpShift)
                        pAPP->Ini.PltShiftBy = pAPP->MinAmpShift;

                UpdateTitleTraces();

                break;




            // Decrease amplitude (increase shift)

            case 0xBD:          // -_ key
            case 0xBC:          // ,< key
            case VK_SUBTRACT:   // numpad - key
            case VK_DOWN:       // down arrow key

                pAPP->Ini.PltShiftBy += nRepCnt;

                if (pAPP->Ini.PltShiftBy > pAPP->MaxAmpShift-1)
                        pAPP->Ini.PltShiftBy = pAPP->MaxAmpShift-1;

                UpdateTitleTraces();

                break;





            // Process arrow keys for scrolling

// This was old use of arrow keys, now the up and down arrows are
// for changing amplitude
//
//            case VK_UP:
//                SendMessage(WM_VSCROLL, MAKELONG(SB_LINEUP, 0), 0L);
//                break;
//            case VK_DOWN:
//                SendMessage(WM_VSCROLL, MAKELONG(SB_LINEDOWN, 0), 0L);
//                break;

            case VK_LEFT:
                SendMessage(WM_HSCROLL, MAKELONG(SB_LINEUP, 0), 0L);
                break;

            case VK_RIGHT:
                SendMessage(WM_HSCROLL, MAKELONG(SB_LINEDOWN, 0), 0L);
                break;



            // Process page up/down keys for scrolling

            case VK_PRIOR:    //PAGEUP key

                if (controlkey)
                        SendMessage(WM_HSCROLL, MAKELONG(SB_PAGEUP, 0), 0L);
                else
                        SendMessage(WM_VSCROLL, MAKELONG(SB_PAGEUP, 0), 0L);
                break;

            case VK_NEXT:     // PAGEDOWN key

                if (controlkey)
                        SendMessage(WM_HSCROLL, MAKELONG(SB_PAGEDOWN, 0), 0L);
                else
                        SendMessage(WM_VSCROLL, MAKELONG(SB_PAGEDOWN, 0), 0L);
                break;




            // Process home/end keys for scrolling

            case VK_HOME:

                if (controlkey)
                        SendMessage(WM_VSCROLL, MAKELONG(SB_TOP, 0), 0L);
                else
                        SendMessage(WM_HSCROLL, MAKELONG(SB_TOP, 0), 0L);
                break;

            case VK_END:

                if (controlkey)
                        SendMessage(WM_VSCROLL, MAKELONG(SB_BOTTOM, 0), 0L);
                else
                        SendMessage(WM_HSCROLL, MAKELONG(SB_BOTTOM, 0), 0L);
                break;



         }




        // Default handling

        CFrameWnd::OnKeyDown(nChar, nRepCnt, nFlags);
}

void CScopeWindow::OnClose() {

        TRACE0("CScopeWindow::OnClose\n");

        int result;

        if (pAPP->IsRunning) {
                pAPP->ErrString = "Please stop before exiting.";
                pAPP->LogMessage( MSGLEVEL_WARN, MSGDLG_SHOW, ERROR_NOT_FATAL, NULL );
                return;
                }


        if (pAPP->IniIsChanged()) {
                result = AfxMessageBox("Options have been changed.\nDo you want to save them in a new .ini file before exiting?\n", MB_YESNOCANCEL);

                if (result == IDCANCEL)
                        return;  // return without closing
                
                if (result == IDYES) {
                        OnIniWrite();
                        return;
                        }
                }


        // Write exit message to log, but don't show dialog

        pAPP->ErrString = "Ending scope.";
        pAPP->LogMessage( MSGLEVEL_MSG, MSGDLG_NO_SHOW, ERROR_NOT_FATAL, NULL );

        CFrameWnd::OnClose();
}

void CScopeWindow::OnIniRead() {

        TRACE0("CScopeWindow::OnIniRead\n");
        
        CString szFilter = "Initialization Files (*.ini)|*.ini|All Files (*.*)|*.*||";

        CFileDialog dlg( TRUE, NULL, NULL,
                         OFN_FILEMUSTEXIST | OFN_HIDEREADONLY | OFN_LONGNAMES,
                         szFilter );


        if (dlg.DoModal() != IDOK)
                return;

        pAPP->Ini.IniFile = dlg.GetPathName();
        pAPP->IniFileSetShortName();



        // Open and parse the .ini file.

        pAPP->IniInputFromFile();


        // Update for possible changes

        pAPP->ResetDC = TRUE;
        pAPP->ComputeActiveChannels();
        if (!pAPP->ComputeBufferSizes()) {
                AfxMessageBox( "Failed compute buffer sizes." );
                return;
                }
        pAPP->UpdatePlotTraceInfo();

        CClientDC dc(this);
        ComputeMaxTitleLength( &dc );
        UpdateStatusBar();
        UpdateColorScheme();
        if (!ComputePlotPositions()) {
                AfxMessageBox( "Failed compute plot positions." );
                return;
                }
        SetScrollBar( SB_BOTH );
        Invalidate();
        SetWindowTitle();
        SetPaneString();
}

void CScopeWindow::OnIniWrite() {

        TRACE0("CScopeWindow::OnIniWrite\n");

        CString szFilter = "Initialization Files (*.ini)|*.ini|All Files (*.*)|*.*||";

        CFileDialog dlg( FALSE, NULL, NULL,
                         OFN_OVERWRITEPROMPT | OFN_HIDEREADONLY | OFN_LONGNAMES,
                         szFilter );

        if (dlg.DoModal() != IDOK)
                return;

        pAPP->Ini.IniFile = dlg.GetPathName();
        pAPP->IniFileSetShortName();

        pAPP->IniOutputToFile();

        SetWindowTitle();
        SetPaneString();
}


void CScopeWindow::OnHardwareDialog() {
        TRACE0("CScopeWindow::OnHardwareDialog (virtual!)\n");

        CString msg;
        msg  = "ERROR: Hardware dialog must be provided by the\n";
        msg += "       derived class.  Please contact technical\n";
        msg += "       support for help with program modifications\n";
        AfxMessageBox( msg );
}

void CScopeWindow::OnChannelDialog() {
        TRACE0("CScopeWindow::OnChannelDialog (virtual!)\n");

        CString msg;
        msg  = "ERROR: Channel dialog must be provided by the\n";
        msg += "       derived class.  Please contact technical\n";
        msg += "       support for help with program modifications\n";
        AfxMessageBox( msg );
}

void CScopeWindow::OnAcquireDialog() {
        TRACE0("CScopeWindow::OnAcquireDialog (virtual!)\n");

        CString msg;
        msg  = "ERROR: Acquire dialog must be provided by the\n";
        msg += "       derived class.  Please contact technical\n";
        msg += "       support for help with program modifications\n";
        AfxMessageBox( msg );
}

void CScopeWindow::OnTriggerDialog() {

        TRACE0("CScopeWindow::OnTriggerDialog\n");

        BOOL oldTrgActionSignal, oldTrgActionSave, oldTrgLatchLine;
        int  oldTrgLineToSignal, oldTrgFilesToSave, oldTrgType;
        int  oldTrgDigLevelLine, oldTrgDigEdgeLine;
        int  oldTrgActiveLevel, oldTrgActiveEdge;
        long oldTrgThreshold;

        CTriggerDlg dlg( this );

        oldTrgType         = pAPP->Ini.TrgType;
        oldTrgThreshold    = pAPP->Ini.TrgThreshold;
        oldTrgDigLevelLine = pAPP->Ini.TrgDigLevelLine;
        oldTrgDigEdgeLine  = pAPP->Ini.TrgDigEdgeLine;
        oldTrgActiveLevel  = pAPP->Ini.TrgActiveLevel;
        oldTrgActiveEdge   = pAPP->Ini.TrgActiveEdge;
        oldTrgActionSave   = pAPP->Ini.TrgActionSave;
        oldTrgFilesToSave  = pAPP->Ini.TrgFilesToSave;
        oldTrgActionSignal = pAPP->Ini.TrgActionSignal;
        oldTrgLineToSignal = pAPP->Ini.TrgLineToSignal;
        oldTrgLatchLine    = pAPP->Ini.TrgLatchLine;

        if (dlg.DoModal() == IDOK) {

                // Update title and pane
                
                SetWindowTitle();
                SetPaneString();

                }

        controlkey = FALSE;
}

void CScopeWindow::OnOutputDialog() {

        TRACE0("CScopeWindow::OnOutputDialog\n");

        int oldNsamples;

        oldNsamples = pAPP->Ini.Nsamples;

        COutputDlg dlg( pAPP->ROnsamples, this );

        if (dlg.DoModal() == IDOK) {

                if (oldNsamples < pAPP->Ini.Nsamples)
                        pAPP->IsAllocated = FALSE;

                if (!pAPP->ComputeBufferSizes())
                        return;

                SetScrollBar( SB_HORZ );


                // Consistancy check for GPS and output format

                if ( (pAPP->Ini.OutFileFormat == FILEFMT_OUT) &&
                     (pAPP->Ini.GpsEnable) ) {

                        pAPP->ErrString  = "Only partial GPS data is saved in .OUT files.\n";
                        pAPP->ErrString += "To save complete GPS information, please use the\n";
                        pAPP->ErrString += ".DAT format.  If partial GPS data is sufficient,\n";
                        pAPP->ErrString += "you may continue with saving to .OUT files.";
                        AfxMessageBox( pAPP->ErrString );
                        }
                }

        controlkey = FALSE;
}

void CScopeWindow::OnDisplayDialog() {

        TRACE0("CScopeWindow::OnDisplayDialog\n");

        int         oldColorScheme, oldNdisplay, oldClipping;
        CDisplayDlg dlg(this);

        oldColorScheme = pAPP->Ini.PltColorScheme;
        oldNdisplay    = pAPP->Ini.PltNdisplay;
        oldClipping    = pAPP->Ini.PltClipping;

        if (dlg.DoModal() == IDOK) {

                if (oldColorScheme != pAPP->Ini.PltColorScheme)
                        UpdateColorScheme();
                if (oldNdisplay != pAPP->Ini.PltNdisplay) {
                        if (!pAPP->ComputeBufferSizes())
                                return;
                        if (!ComputePlotPositions())
                                return;
                        }
                if (oldClipping != pAPP->Ini.PltClipping) {
                        if (!ComputePlotPositions())
                                return;
                        }

                SetScrollBar( SB_BOTH );

                Invalidate();                    // cause repaint
                SetWindowTitle();
                SetPaneString();
                }

        controlkey = FALSE;
}

void CScopeWindow::OnGpsDialog() {

        TRACE0("CScopeWindow::OnGpsDialog\n");

#ifdef GPS_AVAILABLE

        int i, j, maxFirst;
        int oldGpsEnable, oldGpsModel, oldNchannels, oldAllPlotChannels;

        oldGpsEnable = pAPP->Ini.GpsEnable;
        oldGpsModel  = pAPP->Ini.GpsModel;
        oldNchannels = pAPP->Ini.Nchannels;

        CGpsDlg dlg( this );

        if (dlg.DoModal() == IDOK) {

                // If GpsEnable setting has changed, we must update
                // the number of active channels.  If the user was displaying
                // ALL the channels, then we must update that number too.

            if (oldGpsEnable != pAPP->Ini.GpsEnable) {

                if (pAPP->Ini.GpsEnable) {
                    oldAllPlotChannels = pAPP->MaxAnaPlotChannel;
                    pAPP->GpsMarkIndex = pAPP->MaxAnalogChannel;
                    if (pAPP->Ini.DigitalEnable) {
                            oldAllPlotChannels += pAPP->MaxDigPlotChannel; // old count w/o GPS
                            pAPP->GpsMarkIndex += pAPP->MaxDigitalChannel; // new GPS output index
                            }

                    // Turn on gps mark channel and force its name to be "MARK"
                    j = pAPP->MaxAnalogChannel + pAPP->MaxDigitalChannel + 0;
                    pAPP->Ini.pScanActive[j]  = CHECKED;
                    pAPP->Ini.pScanTrigger[j] = UNCHECKED;
                    pAPP->Ini.pScanOrder[pAPP->Ini.Nchannels] = j;
                    SrStringCopy( &pAPP->Ini.pScanTitle[j*MAXTITLE], "MARK", MAXTITLE );
                    pAPP->pDCoffset[pAPP->NchannelsPlot] = 0L;
                    pAPP->Ini.Nchannels++;
                    pAPP->NchannelsPlot++;
                            
                    // Turn on gps counter channel (but not for plotting)
                    if (pAPP->MaxCounterChannel > 0) {
                            j = pAPP->MaxAnalogChannel +
                                pAPP->MaxDigitalChannel + pAPP->MaxGpsChannel;
                            pAPP->Ini.pScanActive[j]  = CHECKED;
                            pAPP->Ini.pScanTrigger[j] = UNCHECKED;
                            pAPP->Ini.pScanOrder[pAPP->Ini.Nchannels] = j;
                            pAPP->Ini.Nchannels++;
                            pAPP->CounterIndex = pAPP->GpsMarkIndex + 1;
                            }                            
                    
                    if (pAPP->Ini.PltNdisplay == oldAllPlotChannels)
                            pAPP->Ini.PltNdisplay += pAPP->MaxGpsPlotChannel;
                    
                    // This calls ComputeBufferSizes and then re-allocates
                    // the data and demux arrays if necessary

                    if (!pAPP->AllocateBuffers())
                            return;

                    } // end if Ini.GpsEnable

                else { // (!pAPP->Ini.GpsEnable)
                    oldAllPlotChannels = pAPP->MaxAnaPlotChannel + pAPP->MaxGpsPlotChannel;
                    if (pAPP->Ini.DigitalEnable)
                            oldAllPlotChannels += pAPP->MaxDigPlotChannel; // old count with GPS
                    pAPP->GpsMarkIndex = SRDAT_INDEXCHANNEL_NONE;


                    // Turn off gps mark channel
                    j = pAPP->MaxAnalogChannel + pAPP->MaxDigitalChannel + 0;
                    pAPP->Ini.pScanActive[j]  = UNCHECKED;
                    pAPP->Ini.pScanTrigger[j] = UNCHECKED;
                    if (pAPP->Ini.Nchannels > 1) {
                            pAPP->Ini.Nchannels--;
                            pAPP->NchannelsPlot--;
                            }

                    // Turn off gps counter channel (but was never plotted)
                    pAPP->CounterIndex = SRDAT_INDEXCHANNEL_NONE;
                    if (pAPP->MaxCounterChannel > 0) {
                            j = pAPP->MaxAnalogChannel +
                                pAPP->MaxDigitalChannel + pAPP->MaxGpsChannel;
                            pAPP->Ini.pScanActive[j]  = UNCHECKED;
                            pAPP->Ini.pScanTrigger[j] = UNCHECKED;
                            if (pAPP->Ini.Nchannels > 1)
                                    pAPP->Ini.Nchannels--;
                            }

                    if ((pAPP->Ini.Nchannels   != oldNchannels)   &&
                        (pAPP->Ini.PltNdisplay == oldAllPlotChannels) ) {
                            for ( i = 0 ; i < pAPP->MaxGpsPlotChannel ; i++ ) {
                                    if (pAPP->Ini.PltNdisplay > 1)
                                            pAPP->Ini.PltNdisplay--;
                                    } // end for i
                            } // end if Ini.Nchannels != oldNchannels
                    
                    } // end else !pAPP->Ini.GpsEnable


                // Update first channel if needed

                 maxFirst = pAPP->NchannelsPlot - pAPP->Ini.PltNdisplay;
                 if (pAPP->Ini.PltFirstTrace > maxFirst)
                         pAPP->Ini.PltFirstTrace = maxFirst;


                 // This updates plot titles and trace types

                 pAPP->UpdatePlotTraceInfo();


                // Update plot if number of channels changed

                if (oldNchannels != pAPP->Ini.Nchannels) {
                        if (!pAPP->ComputeBufferSizes())
                                return;
                        if (!ComputePlotPositions())
                                return;

                        SetScrollBar( SB_BOTH );

                        pAPP->ResetPaint = TRUE;
                        Invalidate();           // cause repaint
                        SetWindowTitle();
                        SetPaneString();
                        }

                } // end if oldGpsEnable

            // Consistancy check for GPS and output format

            if ( (pAPP->Ini.OutFileFormat == FILEFMT_OUT) &&
                 (pAPP->Ini.GpsEnable) ) {

                    pAPP->ErrString  = "Only partial GPS data is saved in .OUT files.\n";
                    pAPP->ErrString += "To save complete GPS information, please use the\n";
                    pAPP->ErrString += ".DAT format.  If partial GPS data is sufficient,\n";
                    pAPP->ErrString += "you may continue with saving to .OUT files.";
                    AfxMessageBox( pAPP->ErrString );
                    }

            } // end DoModal

#else // ! GPS_AVAILABLE

        pAPP->ErrString  = "GPS support currently unavailable.\n";
        pAPP->ErrString += "For GPS support, you must:\n\n";
        pAPP->ErrString += "* have a PARxCH with firmware 3.0 or greater\n";
        pAPP->ErrString += "* have the GPS hardware and software installed\n";
        pAPP->ErrString += "* recompile with GPS_AVAILABLE defined (see makeall.bat)\n";
        pAPP->ErrString += "* reload the recompiled device driver\n";
        AfxMessageBox( pAPP->ErrString );

#endif
        controlkey = FALSE;
}

void CScopeWindow::OnSetClip() {

        TRACE0("CScopeWindow::OnSetClip\n");

        int height, width;

        pAPP->Ini.PltClipping = !pAPP->Ini.PltClipping;

        GetSize( &height, &width );

        SetClipLimits( height, (pYaxis[1]-pYaxis[0])/2 );

//      Used to call UpdateTitleTraces, but that didn't do
//      the trace draw when running.  This caused bad redraw
//      behavior when clipping was changed via the pulldown
//      menu and held longer than one buffer.

        SetWindowTitle();
        SetPaneString();
        if (pAPP->IsAllocated) {
                CClientDC dc(this);
                DrawTraces( &dc );
                }
}

void CScopeWindow::OnSetTriggerLatch() {

        TRACE0("CScopeWindow::OnSetTriggerLatch\n");

        pAPP->Ini.TrgLatchLine = !pAPP->Ini.TrgLatchLine;

        if ( pAPP->Ini.TrgType != TRIG_NONE  &&  // triggering
             pAPP->Ini.TrgLatchLine == FALSE &&  // no latch
             pAPP->IsRunning )                   // running
                pAPP->TriggerClearLatch();


        // Update title and pane

        SetWindowTitle();
        SetPaneString();
}


void CScopeWindow::OnSetDC() {

        TRACE0("CScopeWindow::OnSetDC\n");

        pAPP->ResetDC = TRUE;

        MessageBeep(MB_ICONASTERISK);


       // One beep is enough for Win2K/XP

#if defined( SROS_WIN9X ) || defined( SROS_WIN98 )
        MessageBeep(MB_ICONASTERISK);
        MessageBeep(MB_ICONASTERISK);
#endif
}

void CScopeWindow::OnSetAutoDC() {

        TRACE0("CScopeWindow::OnSetAutoDC\n");

        pAPP->Ini.PltAutoDC = !pAPP->Ini.PltAutoDC;
        pAPP->ResetDC = TRUE;

        SetWindowTitle();
        SetPaneString();
}


void CScopeWindow::OnSetStatusBar() {

        TRACE0("CScopeWindow::OnSetStatusBar\n");

        pAPP->Ini.DoStatusBar = !pAPP->Ini.DoStatusBar;

        UpdateStatusBar();
}

void CScopeWindow::OnSetPlotTrace() {

        TRACE0("CScopeWindow::OnSetPlotTrace\n");

        pAPP->Ini.DoPlotTrace = !pAPP->Ini.DoPlotTrace;

        UpdatePlotTrace();
}

void CScopeWindow::UpdateStatusBar() {

        TRACE0("CScopeWindow::UpdateStatusBar\n");

        CControlBar* pBar;

        // Update status bar

        pBar = GetControlBar( AFX_IDW_STATUS_BAR );
        if (pBar != NULL)
                ShowControlBar( pBar, pAPP->Ini.DoStatusBar, FALSE );

        if (!ComputePlotPositions())
                return;
        Invalidate();
}

void CScopeWindow::UpdatePlotTrace() {

        TRACE0("CScopeWindow::UpdatePlotTrace\n");

        // Update things that depend on plot trace setting

        SetWindowTitle();
        SetPaneString();

        if (pAPP->Ini.DoPlotTrace)
                Invalidate();
}

void CScopeWindow::UpdateTitleTraces() {

        TRACE0("CScopeWindow::UpdateTitleTraces\n");

        // Update title

        SetWindowTitle();
        SetPaneString();


        // Update traces if not running

        if (!pAPP->IsRunning && pAPP->IsAllocated) {
                CClientDC dc(this);
                DrawTraces( &dc );
                }
}


void CScopeWindow::UpdateColorScheme() {

        TRACE0("CScopeWindow::UpdateColorScheme\n");

        DeletePens();   // Remove old pens
        SetColors();    // Set colors according to current scheme
        CreatePens();   // Create new pens
}


void CScopeWindow::SetColors() {

        TRACE0("CScopeWindow::SetColors\n");

        switch (pAPP->Ini.PltColorScheme) {

        case CS_MONO:

                ColorBack  = RGB(255, 255, 255);        // White
                ColorAxis  = RGB(224, 224, 224);        // White Dim
                ColorTitle = RGB( 96,  96,  96);        // Gray Dark
                ColorTrace = RGB(  0,   0,   0);        // Black
                break;


        case CS_GREEN:

                ColorBack  = RGB(  0,   0,   0);        // Black
                ColorAxis  = RGB(128, 128, 128);        // Gray
                ColorTitle = RGB(160, 255, 160);        // Green Light
                ColorTrace = RGB(  0, 255,   0);        // Green
                break;


        case CS_BLUE:

                ColorBack  = RGB(  0,   0, 128);        // Blue Dark
                ColorAxis  = RGB(128, 128, 128);        // Gray
                ColorTitle = RGB(255, 255, 255);        // White
                ColorTrace = RGB(  0, 255, 255);        // Cyan
                break;


        case CS_BROWN:

                ColorBack  = RGB(128,   0,   0);        // Brown
                ColorAxis  = RGB(128, 128, 128);        // Gray
                ColorTitle = RGB(255, 192, 160);        // Light Orange
                ColorTrace = RGB(255, 255, 255);        // White
                break;


        default:
                ColorBack  = RGB(  0,   0,   0);        // Black
                ColorAxis  = RGB(128, 128, 128);        // Gray
                ColorTitle = RGB(  0, 128,   0);        // Green Dark
                ColorTrace = RGB(  0, 255,   0);        // Green
                break;

        }
}

void CScopeWindow::CreatePens() {

        TRACE0("CScopeWindow::CreatePens\n");

        PenBack  = new CPen( PS_SOLID, 1, ColorBack  );
        PenAxis  = new CPen( PS_SOLID, 1, ColorAxis  );
        PenTitle = new CPen( PS_SOLID, 1, ColorTitle );
        PenTrace = new CPen( PS_SOLID, 1, ColorTrace );
}

void CScopeWindow::DeletePens() {

        TRACE0("CScopeWindow::DeletePens\n");

        // Make sure pens are not selected

        CClientDC dc(this);
        dc.SelectObject(GetStockObject(BLACK_PEN));


        // Delete them (destructor calls DeleteObject)

        delete PenBack;   PenBack  = NULL;
        delete PenAxis;   PenAxis  = NULL;
        delete PenTitle;  PenTitle = NULL;
        delete PenTrace;  PenTrace = NULL;
}

void CScopeWindow::CreateFont() {

        TRACE0("CScopeWindow::CreateFont\n");

        if (CurrFont != NULL)
                DeleteFont();


        // Compensate for including Internal Leading here

        if (font.lfHeight > 0)
                font.lfHeight = -font.lfHeight;




        CurrFont = new CFont();
        CurrFont->CreateFontIndirect( &font );

        CurrFont->GetLogFont( &font );



        // Uncompensate for including Internal Leading here

        if (font.lfHeight < 0)
                font.lfHeight = -font.lfHeight;
}

void CScopeWindow::DeleteFont() {

        TRACE0("CScopeWindow::DeleteFont\n");

        // Make sure font is not selected

        CClientDC dc(this);
        dc.SelectObject(GetStockObject(SYSTEM_FIXED_FONT));


        // Delete it (destructor calls DeleteObject)

        delete CurrFont;        CurrFont = NULL;

}

void CScopeWindow::DeletePlotArrays( ) {

        TRACE0("CScopeWindow::DeletePlotArrays\n");

        if (pPtArray)
                { free(pPtArray);  pPtArray  = NULL; }

        if (pOldArray)
                { free(pOldArray); pOldArray = NULL; }
}

void CScopeWindow::DeleteVerticalBuffers( ) {

        TRACE0("CScopeWindow::DeleteVerticalBuffers\n");

        if (pYlast)
                { free(pYlast);    pYlast    = NULL; }

        if (pYaxis)
                { free(pYaxis);    pYaxis    = NULL; }

        if (pYclipmin)
                { free(pYclipmin); pYclipmin = NULL; }

        if (pYclipmax)
                { free(pYclipmax); pYclipmax = NULL; }
}

void CScopeWindow::UpdateTextSize( CDC *pDC ) {

        TRACE0("CScopeWindow::UpdateTextSize\n");

        TEXTMETRIC tm;


        // Update text dimensions

        pDC->SelectObject( CurrFont );
        pDC->GetTextMetrics( &tm );

        cxChar = tm.tmAveCharWidth;
        cyChar = tm.tmHeight + tm.tmExternalLeading;
}


void CScopeWindow::OnContentsDialog() {

        TRACE0("CScopeWindow::OnContentsDialog\n");

        CContentsDlg dlg(this);
        dlg.DoModal();

        controlkey = FALSE;
}

void CScopeWindow::OnAboutDialog() {

        TRACE0("CScopeWindow::OnAboutDialog\n");

        CAboutDlg dlg(this);
        dlg.DoModal();

        controlkey = FALSE;
}

void CScopeWindow::OnDebug() {

        TRACE0("CScopeWindow::OnDebug\n");

        CString msg;

        CClientDC dc(this);

        int lpy  = dc.GetDeviceCaps( LOGPIXELSY );
        int vres = dc.GetDeviceCaps( VERTRES );

        msg.Format("LOGPIXELSY = %d, VERTRES = %d\n",lpy,vres);
        AfxMessageBox(msg);
}


void CScopeWindow::OnUpdateMenuRunning(CCmdUI* pCmdUI) {

        TRACE0("CScopeWindow::OnUpdateMenuRunning\n");

        pCmdUI->Enable( pAPP->IsRunning );
}

void CScopeWindow::OnUpdateMenuStopped(CCmdUI* pCmdUI) {

        TRACE0("CScopeWindow::OnUpdateMenuStopped\n");

        pCmdUI->Enable( !pAPP->IsRunning );
}

void CScopeWindow::OnUpdateClipCheck(CCmdUI* pCmdUI) {

        TRACE0("CScopeWindow::OnUpdateClipCheck\n");

        pCmdUI->SetCheck( pAPP->Ini.PltClipping );
}

void CScopeWindow::OnUpdateAutoDCCheck(CCmdUI* pCmdUI) {

        TRACE0("CScopeWindow::OnUpdateAutoDCCheck\n");

        pCmdUI->SetCheck( pAPP->Ini.PltAutoDC );
}

void CScopeWindow::OnUpdateTriggerLatch(CCmdUI* pCmdUI) {

        TRACE0("CScopeWindow::OnUpdateTriggerLatch\n");

        pCmdUI->SetCheck( pAPP->Ini.TrgLatchLine );
}

void CScopeWindow::OnUpdatePlotTraceCheck(CCmdUI* pCmdUI) {

        TRACE0("CScopeWindow::OnUpdatePlotTraceCheck\n");

        pCmdUI->SetCheck( pAPP->Ini.DoPlotTrace );
}

void CScopeWindow::SetPaneString( void ) {

        TRACE0("CScopeWindow::SetPaneString\n");

        CString PaneString;


        // Sampling rate

        PaneString.Format( "%9.3f Hz", pAPP->Ini.AcqSPS );



        // Auto DC offset

        if (pAPP->Ini.PltAutoDC == TRUE )
                PaneString += "  AutoDC";



        // Clipping

        if (pAPP->Ini.PltClipping == TRUE )
                PaneString += "  Clip";


        // Triggering

        if ((pAPP->Ini.TrgType != TRIG_NONE) ) {
                
                if (pAPP->Ini.TrgLatchLine == TRUE )
                        PaneString += "  Latch";
                
                if (pAPP->TriggerState == TRIGSTATE_INPROGRESS)
                        PaneString += "  Triggered";
                
                else if ( (pAPP->TriggerState == TRIGSTATE_ARMED)  ||
                          (pAPP->TriggerState == TRIGSTATE_WAITING) ) {
                        
                        if (pAPP->Ini.TrgType == TRIG_THRESHOLD)
                                PaneString += "  Threshold trigger ?";
                        
                        else if (pAPP->Ini.TrgType == TRIG_DIGLEVEL) {
                                if (pAPP->Ini.TrgActiveLevel == TRIGLVL_HI)
                                        PaneString += "  High level trigger ?";
                                else
                                        PaneString += "  Low level trigger ?";
                                }
                        
                        else if (pAPP->Ini.TrgType == TRIG_DIGEDGE) {
                                if (pAPP->Ini.TrgActiveEdge == TRIGEDGE_RISING)
                                        PaneString += "  Rising edge trigger ?";
                                else
                                        PaneString += "  Falling edge trigger ?";
                                }

                        } // end ARMED or WAITING

                } // end !TRIG_NONE


        // Gps satellites

        if (pAPP->Ini.GpsEnable) {
                if (pAPP->GpsNumSat == -1 || !pAPP->IsRunning )
                        sprintf( pAPP->szTemp, "  ? Sat" );
                else
                        sprintf( pAPP->szTemp, "  %d Sat",
                                 pAPP->GpsNumSat );
                PaneString += pAPP->szTemp;
                }


        // Add data file name (or file error)

        if ( pAPP->FileError )
                PaneString += "  ** FILE ERROR **";
        
        else if ( pAPP->Ini.OutFileFormat == FILEFMT_NONE )
                PaneString += "  No file";
        
        else if ( !pAPP->ShortFileName.IsEmpty() ) {
                sprintf( pAPP->szTemp, "  %s", pAPP->ShortFileName );
                PaneString += pAPP->szTemp;
                }



        // Now, finally write out the pane string we've built up

        m_wndStatusBar.SetPaneText( PANE_INFO, (LPCTSTR)PaneString, TRUE );
}

void CScopeWindow::SetScrollBar( int nBar ) {

        TRACE0("CScopeWindow::SetScrollBar\n");

        SCROLLINFO sif;

        sif.cbSize = sizeof(SCROLLINFO);
        sif.fMask  = SIF_ALL;


        if ((nBar == SB_VERT) || (nBar == SB_BOTH)) {

                sif.nMin  = 0;
                sif.nMax  = pAPP->NchannelsPlot - 1;
                sif.nPage = pAPP->Ini.PltNdisplay;
                sif.nPos  = pAPP->Ini.PltFirstTrace;

                SetScrollInfo( SB_VERT, &sif, TRUE );
                }


        if ((nBar == SB_HORZ) || (nBar == SB_BOTH)) {

                sif.nMin  = 0;
                sif.nMax  = pAPP->NsamplesPlot - 1;
                sif.nPage = TraceWidth;
                sif.nPos  = pAPP->Ini.PltFirstSample;

                SetScrollInfo( SB_HORZ, &sif, TRUE );
                }

}

void CScopeWindow::SetWindowTitle() {

        TRACE0("CScopeWindow::SetWindowTitle\n");

        CString title;
        CString tmpstr;


        // Fill in app name and maybe ini name

        if ( pAPP->Ini.IniFileShort.IsEmpty() )
                title = pAPP->AppTitle;
        else
                title.Format( "%s %s", pAPP->AppTitle, pAPP->Ini.IniFileShort );
        

        // Add optional title

        if (!pAPP->Ini.IniTitle.IsEmpty()) {
                tmpstr.Format(" (%s)", pAPP->Ini.IniTitle);
                title += tmpstr;
                }


        // Indicate program run/stop info

        if (pAPP->IsRunning == TRUE)
                title += " - Running (Ctrl+S to stop)";
        else
                title += " - Stopped (Ctrl+R to run)";



        // Showing bits

        tmpstr.Format( " - Showing %d bits",
                       pAPP->MaxAmpShift - pAPP->Ini.PltShiftBy );
        title += tmpstr;



        // Indicate trigger in process

        if ((pAPP->Ini.TrgType != TRIG_NONE) && pAPP->IsTriggered)
                title += " - TRIGGERED";



        SetWindowText( title );
}

void CScopeWindow::GetSize( int *height, int *width ) {

        TRACE0("CScopeWindow::GetSize\n");

        RECT r, rs;

        GetClientRect(&r);

        if (m_wndStatusBar.IsWindowVisible())
                m_wndStatusBar.GetClientRect( &rs );
        else
                rs.bottom = 0;


        *height = r.bottom - rs.bottom;
        *width  = r.right;
}

int CScopeWindow::ChangeFontSize( int newsize ) {

        TRACE0("CScopeWindow::ChangeFontSize\n");


        // Enforce reasonable sizes

        if (newsize < minPoint)
                newsize = minPoint;
        if (newsize > maxPoint)
                newsize = maxPoint;



        // Test for no change

        if (font.lfHeight == newsize)
                return( 0 );



        // Create font with new size

        font.lfWidth  = 0;
        font.lfHeight = newsize;


// Leave font the same no matter what size we're asking for
//        
//        if (newsize < midPoint)
//                SrStringCopy(font.lfFaceName,"Small Fonts", LF_FACESIZE);
//        else
//                SrStringCopy(font.lfFaceName,"CourierNew", LF_FACESIZE); // was Arial


        CreateFont();




        // Refigure text size and title length

        CClientDC dc(this);

        UpdateTextSize( &dc );

        ComputeMaxTitleLength( &dc );


        return( 1 );
}

int CScopeWindow::ComputeMaxTitleLength( CDC *pDC ) {

        TRACE0("CScopeWindow::ComputeMaxTitleLength\n");

        int       i, ncount;
        CSize     sz(0,0);
        LPCTSTR   pstr;

        MaxTitleLength = 0;

        pDC->SelectObject( CurrFont );


        for ( i = 0 ; i < pAPP->MaxTotalPlotChannel ; i++ ) {

                pstr   = &(pAPP->pPlotTitle[i*MAXTITLE]);
                ncount = strlen( pstr );

                sz     = pDC->GetTextExtent( pstr, ncount );

                if (MaxTitleLength < sz.cx)
                        MaxTitleLength = sz.cx;
                }

        return( MaxTitleLength );
}

int CScopeWindow::ComputePlotPositions( ) {

        TRACE0("CScopeWindow::ComputePlotPositions\n");

        int    totalHeight, traceHalf, evenHeight, halfExcess;
        int    height, width, sbextra, maxFirstSample;
        int    i, j, xpos, yax;
        POINT *pArray;

        SCROLLINFO sif;


        // Always check the return value of this function and return
        // immediately if it fails.  This is because failure means that
        // the allocate plot buffer function has failed and the destroy
        // window function may have been called so the entire CScopeWindow
        // is deleted.  You can't access any class variables at that point,
        // so all functions should just unwind with returns.




        // Get size of window

        GetSize( &height, &width );



        // Always compute totalHeight assuming that horizontal scroll bar
        // is visible to avoid race condition where size without bar
        // increases font which reduces width so that bar is needed, but
        // size with bar has smaller font that increases width so bar
        // is not needed, etc

        GetScrollInfo( SB_HORZ, &sif );

        if ((int)sif.nPage >= sif.nMax)
                sbextra = GetSystemMetrics( SM_CYHSCROLL );
        else
                sbextra = 0;



        // Compute trace height

        totalHeight = height - 2 * VMARGIN - 2 * VTICK - sbextra;
        TraceHeight = (int)((float)totalHeight / pAPP->Ini.PltNdisplay);


        ChangeFontSize( TraceHeight/2 );  // was TraceHeight


        // Compute scale positions

        YScaleTop    = 0      + VMARGIN;
        YScaleBottom = height - VMARGIN;


        // Compute trace width

        TraceStart = HMARGIN + MaxTitleLength + HMARGIN;
        TraceEnd   = width - HMARGIN;
        TraceWidth = TraceEnd - TraceStart;

        if (TraceWidth < 1)
                TraceWidth = 1;



        // Compute trace height positions

        traceHalf   = TraceHeight / 2;
        evenHeight  = TraceHeight * pAPP->Ini.PltNdisplay;
        halfExcess  = (totalHeight - evenHeight + sbextra) / 2;

        pYaxis[0] = VMARGIN + VTICK + halfExcess + traceHalf;
        pYlast[0] = pYaxis[0];

        for ( i = 1 ; i < pAPP->MaxTotalPlotChannel ; i++ ) {
                pYaxis[i] = pYaxis[i - 1] + TraceHeight;
                pYlast[i] = pYaxis[i];
                }


        // Update clipping

        SetClipLimits( height, traceHalf );




        // If needed, reallocate plot buffers

        if (!AllocatePlotBuffers())
                return( 0 );



        // Reset horizontal scroll bar range

        maxFirstSample = pAPP->NsamplesPlot - TraceWidth - 1;
        if (maxFirstSample < 0)
                maxFirstSample = 0;
        if (pAPP->Ini.PltFirstSample > maxFirstSample)
                pAPP->Ini.PltFirstSample = maxFirstSample;

        SetScrollBar( SB_HORZ );



        // Fill up polyline positions

        pAPP->ResetPaint = TRUE;

        pArray = pPtArray;

        for ( j = 0 ; j < pAPP->NchannelsPlot; j++ ) {

                xpos = TraceStart;
                yax  = pYaxis[j];

                for ( i = 0 ; i < pAPP->NsamplesPlot ; i++ ) {

                        pArray->x = xpos++;
                        pArray->y = yax;

                        pArray++;
                        }
                }

        // Need to replot

        Invalidate( FALSE );

        return( 1 );
}


int CScopeWindow::AllocatePlotBuffers() {

        TRACE0("CScopeWindow::AllocatePlotBuffers\n");


        // Only reallocate if need to increase size

        if (LastPtsPlot < pAPP->NbufferptsPlot) {

                LastPtsPlot = pAPP->NbufferptsPlot;


                // Clean up any previous plot arrays

                DeletePlotArrays( );


                // Allocate current arrays

                pPtArray  = (POINT*) malloc(pAPP->NbufferptsPlot * sizeof(POINT));
                pOldArray = (POINT*) malloc(pAPP->NbufferptsPlot * sizeof(POINT));

                if (!pPtArray || !pOldArray) {
                        pAPP->ErrString = "Can not allocate plot buffers.";
                        pAPP->LogMessage( MSGLEVEL_ERR, MSGDLG_SHOW, ERROR_FATAL, NULL );
                        return( 0 );
                        }
                }


        // Fail with NULL arrays even if haven't reallocated
        if (!pPtArray || !pOldArray)
                return( 0 );

        return( 1 );
}


void CScopeWindow::InitDialogPositions( void ) {

        TRACE0("CScopeWindow::InitDialogPositions\n");


        // These dialogs vary by specific A/D system

        HardwarePosTop  = DLG_POS_INIT;
        HardwarePosLeft = DLG_POS_INIT;

        ChannelPosTop   = DLG_POS_INIT;
        ChannelPosLeft  = DLG_POS_INIT;

        AcquirePosTop   = DLG_POS_INIT;
        AcquirePosLeft  = DLG_POS_INIT;



        // These dialogs are identical for all A/D systems

        TriggerPosTop   = DLG_POS_INIT;
        TriggerPosLeft  = DLG_POS_INIT;

        OutputPosTop    = DLG_POS_INIT;
        OutputPosLeft   = DLG_POS_INIT;

        DisplayPosTop   = DLG_POS_INIT;
        DisplayPosLeft  = DLG_POS_INIT;

        GpsPosTop       = DLG_POS_INIT;
        GpsPosLeft      = DLG_POS_INIT;

        ContentsPosTop  = DLG_POS_INIT;
        ContentsPosLeft = DLG_POS_INIT;

        TopicPosTop     = DLG_POS_INIT;
        TopicPosLeft    = DLG_POS_INIT;
        TopicPosBottom  = DLG_POS_INIT;
        TopicPosRight   = DLG_POS_INIT;

}

void CScopeWindow::GetDialogPosition( RECT *r, int nID ) {

        TRACE0("CScopeWindow::GetDialogPosition\n");

        int height, width;


        // Keep original size

        height = r->bottom - r->top;
        width  = r->right  - r->left;


        // Get saved dialog positions

        switch ( nID ) {


                // These dialogs vary by specific A/D system

                case IDD_HARDWARE:
                        r->top    = HardwarePosTop;
                        r->left   = HardwarePosLeft;
                        r->bottom = r->top  + height;
                        r->right  = r->left + width;
                        break;

                case IDD_CHANNEL:
                        r->top    = ChannelPosTop;
                        r->left   = ChannelPosLeft;
                        r->bottom = r->top  + height;
                        r->right  = r->left + width;
                        break;

                case IDD_ACQUIRE:
                        r->top    = AcquirePosTop;
                        r->left   = AcquirePosLeft;
                        r->bottom = r->top  + height;
                        r->right  = r->left + width;
                        break;



                // These dialogs are identical for all A/D systems

                case IDD_TRIGGER:
                        r->top    = TriggerPosTop;
                        r->left   = TriggerPosLeft;
                        r->bottom = r->top  + height;
                        r->right  = r->left + width;
                        break;

                case IDD_OUTPUT:
                        r->top    = OutputPosTop;
                        r->left   = OutputPosLeft;
                        r->bottom = r->top  + height;
                        r->right  = r->left + width;
                        break;

                case IDD_DISPLAY:
                        r->top    = DisplayPosTop;
                        r->left   = DisplayPosLeft;
                        r->bottom = r->top  + height;
                        r->right  = r->left + width;
                        break;

                case IDD_GPS:
                        r->top    = GpsPosTop;
                        r->left   = GpsPosLeft;
                        r->bottom = r->top  + height;
                        r->right  = r->left + width;
                        break;

                case IDD_HELPCONTENTS:
                        r->top    = ContentsPosTop;
                        r->left   = ContentsPosLeft;
                        r->bottom = r->top  + height;
                        r->right  = r->left + width;
                        break;

                case IDD_HELPTOPIC:
                        r->top    = TopicPosTop;
                        r->left   = TopicPosLeft;
                        r->bottom = TopicPosBottom;
                        r->right  = TopicPosRight;
                        break;


                default:
                        r->top    = DLG_POS_INIT;
                        r->left   = DLG_POS_INIT;
                        r->bottom = DLG_POS_INIT;
                        r->right  = DLG_POS_INIT;
                        break;
                }
}


void CScopeWindow::SaveDialogPosition( RECT *r, int nID ) {

        TRACE0("CScopeWindow::SaveDialogPosition\n");


        // Save dialog positions

        switch ( nID ) {

                // These dialogs vary by specific A/D system

                case IDD_HARDWARE:
                        HardwarePosTop  = r->top;
                        HardwarePosLeft = r->left;
                        break;

                case IDD_CHANNEL:
                        ChannelPosTop   = r->top;
                        ChannelPosLeft  = r->left;
                        break;

                case IDD_ACQUIRE:
                        AcquirePosTop   = r->top;
                        AcquirePosLeft  = r->left;
                        break;



                // These dialogs are identical for all A/D systems

                case IDD_TRIGGER:
                        TriggerPosTop   = r->top;
                        TriggerPosLeft  = r->left;
                        break;

                case IDD_OUTPUT:
                        OutputPosTop    = r->top;
                        OutputPosLeft   = r->left;
                        break;

                case IDD_DISPLAY:
                        DisplayPosTop   = r->top;
                        DisplayPosLeft  = r->left;
                        break;

                case IDD_GPS:
                        GpsPosTop       = r->top;
                        GpsPosLeft      = r->left;
                        break;

                case IDD_HELPCONTENTS:
                        ContentsPosTop  = r->top;
                        ContentsPosLeft = r->left;
                        break;

                case IDD_HELPTOPIC:
                        TopicPosTop     = r->top;
                        TopicPosLeft    = r->left;
                        TopicPosBottom  = r->bottom;
                        TopicPosRight   = r->right;
                        break;

                }
}





//***********************************************************************
//
// CLASS CScopeDialog
//
// This class, derived from CDialog, contains a number of validation and
// convenience functions available to all the scope dialogs.  Basically,
// it is just an extension of the CDialog class
//

BEGIN_MESSAGE_MAP(CScopeDialog, CDialog)
        //{{AFX_MSG_MAP(CScopeDialog)
        ON_WM_ACTIVATEAPP()
        //}}AFX_MSG_MAP
END_MESSAGE_MAP()




CScopeDialog::CScopeDialog(UINT nIdTemplate, CWnd* pParent /*=NULL*/)
             : CDialog(nIdTemplate, pParent) {

        TRACE0("CScopeDialog::CScopeDialog\n");

        DialogId   = nIdTemplate;
        CancelId   = IDCANCEL;
        DoValidate = TRUE;
}

BOOL CScopeDialog::OnInitDialog( int OkButtonId ) {

        TRACE0("CScopeDialog::OnInitDialog\n");

        CDialog::OnInitDialog();


        // Return to saved position, if any

        RestorePosition();


        // Disable OK button if not in interactive mode

        if ( !pAPP->Ini.DoInteractive && (OkButtonId > 0) ) {
                CWnd *pWnd = GetDlgItem( OkButtonId );
                if (pWnd)
                        pWnd->EnableWindow( FALSE );
                }


        return TRUE;  // return TRUE unless you set the focus to a control
                      // EXCEPTION: OCX Property Pages should return FALSE
}

void CScopeDialog::OnOK() {

        TRACE0("CScopeDialog::OnOK\n");

        RememberPosition();

        CDialog::OnOK();
}

void CScopeDialog::OnCancel() {

        TRACE0("CScopeDialog::OnCancel\n");

        RememberPosition();

        CDialog::OnCancel();

        DoValidate = FALSE;
}

void CScopeDialog::OnActivateApp( BOOL bActive, HTASK hTask ) {

        TRACE0("CScopeDialog::OnActivateApp\n");

        CDialog::OnActivateApp( bActive, hTask );

        DoValidate = bActive;
}

LONG CScopeDialog::OnEditLostFocus( UINT nID, LONG xxx ) {

        TRACE1("CScopeDialog::OnEditLostFocus (%u)\n", nID);

        MessageBeep( 0 );

        return( 1 );
}

int CScopeDialog::CheckFocus() {

        TRACE0("CScopeDialog::CheckFocus\n");

        CWnd *pWnd;
        int   iret;

        // Call this function from inside button click routines.
        // Even though the click routine is being executed, the
        // previous control may still have the focus.  If it was
        // an edit control, it needs to go through the "losing
        // focus" validation function.

        pWnd = GetFocus();

        if (pWnd)
                iret = OnEditLostFocus(pWnd->GetDlgCtrlID(), 0L );
        else
                iret = 1;


        return( iret );
}

int CScopeDialog::ValidateAsInt( int nID, int minval, int maxval, int *theval, char *txt ) {

        TRACE0("CScopeDialog::ValidateAsInt\n");

        BOOL good;
        int  oldval, newval;


        oldval = *theval;

        newval = (int)GetDlgItemInt(nID, &good);

        if ((good) && (newval >= minval) && (newval <= maxval) ) {
                *theval = newval;
                return( TRUE );
                }
        else {
                pAPP->ErrString.Format("%s must be an integer between %d and %d.", (LPCTSTR)txt, minval, maxval);
                AfxMessageBox( pAPP->ErrString );
                SetDlgItemInt(nID, oldval);
                GotoDlgCtrl( GetDlgItem(nID) );
                return( FALSE );
                }
}

int CScopeDialog::ValidateAsLong( int nID, long minval, long maxval, long *theval, char *txt ) {

        TRACE0("CScopeDialog::ValidateAsLong\n");

        BOOL good;
        long  oldval, newval;


        oldval = *theval;

        newval = (long)GetDlgItemInt(nID, &good);

        if ((good) && (newval >= minval) && (newval <= maxval) ) {
                *theval = newval;
                return( TRUE );
                }
        else {
                pAPP->ErrString.Format("%s must be a long integer between %d and %d.", (LPCTSTR)txt, minval, maxval);
                AfxMessageBox( pAPP->ErrString );
                SetDlgItemInt(nID, (int)oldval);
                GotoDlgCtrl( GetDlgItem(nID) );
                return( FALSE );
                }
}

int CScopeDialog::ValidateAsHex( int nID, int minval, int maxval, int *theval, char *txt ) {

        TRACE0("CScopeDialog::ValidateAsHex\n");

        CString str;
        BOOL    good;
        int     oldval, newval;


        oldval = *theval;

        GetDlgItemText( nID, str );

        newval = strtol( str, NULL, 0 );

        // Overflow or underflow in conversion to int
        
        if ( (newval == LONG_MAX) || (newval == LONG_MIN) )
                good = FALSE;

        // Found zero or can't convert

        else if ( newval == 0 ) {
                str.TrimLeft();
                if ( (str.Compare("0") == 0) || (str.Compare("0x0") == 0) )
                        good = TRUE;
                else
                        good = FALSE;
                }

        // All is good

        else
                good = TRUE;

        if ((good) && (newval >= minval) && (newval <= maxval) ) {
                *theval = newval;
                return( TRUE );
                }
        else {
                pAPP->ErrString.Format("%s must be a hex number between 0x%X and 0x%X.", (LPCTSTR)txt, minval, maxval);
                AfxMessageBox( pAPP->ErrString );
                SetDlgItemHex( nID, oldval );
                GotoDlgCtrl( GetDlgItem(nID) );
                return( FALSE );
                }
}


int CScopeDialog::ValidateAsFloat( int nID, float minval, float maxval, float *theval, char *txt ) {

        TRACE0("CScopeDialog::ValidateAsFloat\n");

        CString btntxt;
        BOOL    good;
        float   oldval, newval;


        oldval = *theval;

        GetDlgItemText(nID, btntxt);
        newval = (float)strtod(btntxt,NULL);
        if (newval != 0.0)
                good = TRUE;
        else
                good = FALSE;

        if ((good) && (newval >= minval) && (newval <= maxval) ) {
                *theval = newval;
                return( TRUE );
                }
        else {
                pAPP->ErrString.Format("%s must be a float between %f and %f.", (LPCTSTR)txt, minval, maxval);
                AfxMessageBox( pAPP->ErrString );
                btntxt.Format("%f", oldval);
                SetDlgItemText(nID, btntxt);
                GotoDlgCtrl( GetDlgItem(nID) );
                return( FALSE );
                }
}


int CScopeDialog::SetDlgItemHex( int nID, int val ) {

        TRACE0("CScopeDialog::SetDlgItemHex\n");

        CString txt;
        txt.Format("0x%X", val);
        SetDlgItemText(nID, txt);

        return(TRUE);
}


int CScopeDialog::SetDlgItemFloat( int nID, float val ) {

        TRACE0("CScopeDialog::SetDlgItemFloat\n");

        CString txt;

        txt.Format("%.3f", val);
        SetDlgItemText( nID, txt );

        return(TRUE);
}


float CScopeDialog::GetDlgItemFloat(int nID, BOOL *lpTrans ) {

        CString txt;
        float   newval;

        GetDlgItemText(nID, txt);
        newval = (float)strtod(txt,NULL);


        // Fill error flag if desired

        if (lpTrans != NULL) {
                if (newval == 0.0)
                        *lpTrans = FALSE;
                else
                        *lpTrans = TRUE;
                }

        return( newval );
}


BOOL CScopeDialog::SetTextReadOnly( int nID ) {

        TRACE0("CScopeDialog::SetTextReadOnly\n");

        CEdit *pTxt = (CEdit*)GetDlgItem( nID );
        if (pTxt)
                return( pTxt->SetReadOnly( TRUE ) );
        else
                return( FALSE );
}

void CScopeDialog::RestorePosition() {

        TRACE0("CScopeDialog::RestorePosition\n");

        CScopeWindow *pWnd = (CScopeWindow *)m_pParentWnd;

        if (!pWnd)
                return;


        // Get current positions

        RECT r;

        GetWindowRect( &r );



        // Update with saved positions

        pWnd->GetDialogPosition( &r, DialogId );



        // If no uninitialized positions, then move dialog

        if ((r.top    != DLG_POS_INIT) &&
            (r.left   != DLG_POS_INIT) &&
            (r.bottom != DLG_POS_INIT) &&
            (r.right  != DLG_POS_INIT) )

                MoveWindow( &r );
}

void CScopeDialog::RememberPosition() {

        TRACE0("CScopeDialog::RememberPosition\n");


        CScopeWindow *pWnd = (CScopeWindow *)m_pParentWnd;

        if (!pWnd)
                return;



        // Get current position and have main window save it.

        RECT r;

        GetWindowRect( &r );

        pWnd->SaveDialogPosition( &r, DialogId );
}




//***********************************************************************
//
// CLASS CAboutDlg
//
// This class, derived from CDialog, displays version information about
// the current program.
//

BEGIN_MESSAGE_MAP(CAboutDlg, CDialog)
        //{{AFX_MSG_MAP(CAboutDlg)
        ON_BN_CLICKED(IDC_BTN_X_OK,  OnOK)
        //}}AFX_MSG_MAP
END_MESSAGE_MAP()




CAboutDlg::CAboutDlg(CWnd* pParent /*=NULL*/)
          : CDialog(CAboutDlg::IDD, pParent) {

        TRACE0("CAboutDlg::CAboutDlg\n");
}

BOOL CAboutDlg::OnInitDialog() {

        TRACE0("CAboutDlg::OnInitDialog\n");

        CDialog::OnInitDialog();


        // Set dialog title based on app title

        sprintf( pAPP->szTemp, "About %s", pAPP->AppTitle );

        SetWindowText( pAPP->szTemp );



        // Set version text based on app title

        CStatic *pWnd = (CStatic *) GetDlgItem( IDC_LAB_X_VERSION );

        if (pWnd) {

                sprintf( pAPP->szTemp, "%s Rev %s",
                         pAPP->AppTitle, pAPP->AppRev );

                pWnd->SetWindowText( pAPP->szTemp );

                }


        return TRUE;  // return TRUE unless you set the focus to a control
                      // EXCEPTION: OCX Property Pages should return FALSE
}

void CAboutDlg::OnOK() {

        TRACE0("CAboutDlg::OnOK\n");

        CDialog::OnOK();
}





//***********************************************************************
//
// CLASS CTriggerDlg
//
// This class, derived from CScopeDialog, presents and controls the Trigger
// dialog.  This dialog allows the user to specify various trigger options
// such as the number of traces to be displayed, the scaling applied
// to those traces, and the overall color scheme.

BEGIN_MESSAGE_MAP(CTriggerDlg, CScopeDialog)
        //{{AFX_MSG_MAP(CTriggerDlg)
        ON_MESSAGE(WM_EDITLOSTFOCUS,           OnEditLostFocus)
        ON_BN_CLICKED(IDC_BTN_T_OK,            OnOK)
        ON_BN_CLICKED(IDC_BTN_T_CANCEL,        OnCancel)
        ON_BN_CLICKED(IDC_BTN_T_HELP,          OnHelp)
        ON_BN_CLICKED(IDC_OPT_T_TRIGGERNONE,   OnOptTriggerType)
        ON_BN_CLICKED(IDC_OPT_T_TRIGGERTHRESH, OnOptTriggerType)
        ON_BN_CLICKED(IDC_OPT_T_TRIGGERDIGLEVEL, OnOptTriggerType)
        ON_BN_CLICKED(IDC_OPT_T_TRIGGERDIGEDGE,  OnOptTriggerType)
        ON_BN_CLICKED(IDC_OPT_T_LEVELLOW,        OnOptTriggerLevel)
        ON_BN_CLICKED(IDC_OPT_T_LEVELHIGH,       OnOptTriggerLevel)
        ON_BN_CLICKED(IDC_OPT_T_EDGEFALL,        OnOptTriggerEdge)
        ON_BN_CLICKED(IDC_OPT_T_EDGERISE,        OnOptTriggerEdge)
        ON_BN_CLICKED(IDC_BTN_T_ALLON,         OnBtnAllOn)
        ON_BN_CLICKED(IDC_BTN_T_ALLOFF,        OnBtnAllOff)
        ON_BN_CLICKED(IDC_CHK_T_ACTIONSAVE,    OnChkActionSave)
        ON_BN_CLICKED(IDC_CHK_T_ACTIONSIGNAL,  OnChkActionSignal)
        ON_EN_KILLFOCUS(IDC_TXT_T_THRESHOLD,   OnKillfocusTxtTriggerThreshold)
        ON_EN_KILLFOCUS(IDC_TXT_T_DIGLEVELLINE,  OnKillfocusTxtTriggerDigLevelLine)
        ON_EN_KILLFOCUS(IDC_TXT_T_DIGEDGELINE,   OnKillfocusTxtTriggerDigEdgeLine)
        ON_EN_KILLFOCUS(IDC_TXT_T_SAVEFILES,   OnKillfocusTxtTriggerFilesToSave)
        ON_EN_KILLFOCUS(IDC_TXT_T_SIGNALLINE,  OnKillfocusTxtTriggerLineToSignal)
        //}}AFX_MSG_MAP
//      ON_CONTROL_RANGE(BN_CLICKED, IDC_OPT_T_TRIGGERNONE, IDC_OPT_T_TRIGGERDIGEDGE, OnOptTriggertype)
END_MESSAGE_MAP()




CTriggerDlg::CTriggerDlg(CWnd* pParent /*=NULL*/)
              : CScopeDialog(CTriggerDlg::IDD, pParent) {

        TRACE0("CTriggerDlg::CTriggerDlg\n");

        int i;

        //{{AFX_DATA_INIT(CTriggerDlg)
        m_TriggerType         = 0;
        m_TriggerThreshold    = 0;
        m_TriggerDigLevelLine = 0;
        m_TriggerDigEdgeLine  = 0;
        m_TriggerActiveLevel  = 0;
        m_TriggerActiveEdge   = 0;
        m_TriggerActionSave   = TRUE;
        m_TriggerFilesToSave  = 0;
        m_TriggerTimeToSave   = 0;
        m_TriggerActionSignal = FALSE;
        m_TriggerLineToSignal = 0;
        m_TriggerLatchLine    = FALSE;
        m_HasDigio            = FALSE;
        //}}AFX_DATA_INIT

        for ( i = 0 ; i < 8 ; i++ )
                m_AnalogChannelExist[i] = TRUE;

}


CTriggerDlg::~CTriggerDlg() {
        TRACE0("CTriggerDlg::~CTriggerDlg\n");

        // Nothing to clean up
}

void CTriggerDlg::DoDataExchange(CDataExchange* pDX) {

        CScopeDialog::DoDataExchange(pDX);

        //{{AFX_DATA_MAP(CTriggerDlg)
        DDX_Control(pDX, IDC_CHK_T_CH00,            m_CHK_TriggerChan[0]);
        DDX_Control(pDX, IDC_CHK_T_CH01,            m_CHK_TriggerChan[1]);
        DDX_Control(pDX, IDC_CHK_T_CH02,            m_CHK_TriggerChan[2]);
        DDX_Control(pDX, IDC_CHK_T_CH03,            m_CHK_TriggerChan[3]);
        DDX_Control(pDX, IDC_CHK_T_CH04,            m_CHK_TriggerChan[4]);
        DDX_Control(pDX, IDC_CHK_T_CH05,            m_CHK_TriggerChan[5]);
        DDX_Control(pDX, IDC_CHK_T_CH06,            m_CHK_TriggerChan[6]);
        DDX_Control(pDX, IDC_CHK_T_CH07,            m_CHK_TriggerChan[7]);
        DDX_Control(pDX, IDC_CHK_T_ACTIONSAVE,    m_CHK_TriggerActionSave);
        DDX_Control(pDX, IDC_CHK_T_ACTIONSIGNAL,  m_CHK_TriggerActionSignal);
        DDX_Control(pDX, IDC_CHK_T_LATCHLINE,     m_CHK_TriggerLatchLine);
        DDX_Control(pDX, IDC_TXT_T_THRESHOLD,     m_TXT_TriggerThreshold);
        DDX_Control(pDX, IDC_TXT_T_DIGLEVELLINE,    m_TXT_TriggerDigLevelLine);
        DDX_Control(pDX, IDC_TXT_T_DIGEDGELINE,     m_TXT_TriggerDigEdgeLine);
        DDX_Control(pDX, IDC_TXT_T_SAVEFILES,     m_TXT_TriggerFilesToSave);
        DDX_Control(pDX, IDC_TXT_T_SIGNALLINE,    m_TXT_TriggerLineToSignal);
        DDX_Control(pDX, IDC_OPT_T_TRIGGERNONE,   m_OPT_TriggerNone);
        DDX_Control(pDX, IDC_OPT_T_TRIGGERTHRESH, m_OPT_TriggerThresh);
        DDX_Control(pDX, IDC_OPT_T_TRIGGERDIGLEVEL, m_OPT_TriggerDigLevel);
        DDX_Control(pDX, IDC_OPT_T_TRIGGERDIGEDGE,  m_OPT_TriggerDigEdge);
        DDX_Control(pDX, IDC_OPT_T_LEVELHIGH,       m_OPT_TriggerLevelHigh);
        DDX_Control(pDX, IDC_OPT_T_LEVELLOW,        m_OPT_TriggerLevelLow);
        DDX_Control(pDX, IDC_OPT_T_EDGERISE,        m_OPT_TriggerEdgeRise);
        DDX_Control(pDX, IDC_OPT_T_EDGEFALL,        m_OPT_TriggerEdgeFall);
        DDX_Text(pDX, IDC_TXT_T_THRESHOLD,        m_TriggerThreshold);
        DDX_Text(pDX, IDC_TXT_T_DIGLEVELLINE,       m_TriggerDigLevelLine);
        DDX_Text(pDX, IDC_TXT_T_DIGEDGELINE,        m_TriggerDigEdgeLine);
        DDX_Text(pDX, IDC_TXT_T_SAVEFILES,        m_TriggerFilesToSave);
        DDX_Text(pDX, IDC_TXT_T_SIGNALLINE,       m_TriggerLineToSignal);
        //}}AFX_DATA_MAP
}


BOOL CTriggerDlg::OnInitDialog() {

        // Initialize member variables from application globals

        m_TriggerType         = pAPP->Ini.TrgType;
        m_TriggerThreshold    = pAPP->Ini.TrgThreshold;
        m_TriggerDigLevelLine = pAPP->Ini.TrgDigLevelLine;
        m_TriggerDigEdgeLine  = pAPP->Ini.TrgDigEdgeLine;
        m_TriggerActiveLevel  = pAPP->Ini.TrgActiveLevel;
        m_TriggerActiveEdge   = pAPP->Ini.TrgActiveEdge;
        m_TriggerActionSave   = pAPP->Ini.TrgActionSave;
        m_TriggerFilesToSave  = pAPP->Ini.TrgFilesToSave;
        m_TriggerActionSignal = pAPP->Ini.TrgActionSignal;
        m_TriggerLineToSignal = pAPP->Ini.TrgLineToSignal;
        m_TriggerLatchLine    = pAPP->Ini.TrgLatchLine;

        if (pAPP->MaxTriggerDigioLine < 0)
                m_HasDigio = FALSE;
        else
                m_HasDigio = TRUE;

        // Initialize base class and initialize text boxes (via DDX)

        CScopeDialog::OnInitDialog( IDC_BTN_T_OK );


        // Only enable digital trigger type and action if the
        // hardware supports it

        GetDlgItem(IDC_OPT_T_TRIGGERDIGLEVEL)->EnableWindow(m_HasDigio);
        GetDlgItem(IDC_OPT_T_TRIGGERDIGEDGE)->EnableWindow(m_HasDigio);
        GetDlgItem(IDC_CHK_T_ACTIONSIGNAL)->EnableWindow(m_HasDigio);



        // Initialize radio buttons and check boxes

        SetTriggerChannelCheckBoxes();
        SetTriggerLevelRadio();
        SetTriggerEdgeRadio();
        SetTriggerTypeRadio();
        OnOptTriggerType();

        
        m_CHK_TriggerActionSignal.SetCheck( m_TriggerActionSignal );
        m_CHK_TriggerActionSave.SetCheck( m_TriggerActionSave );
        m_CHK_TriggerLatchLine.SetCheck( m_TriggerLatchLine );

        OnChkActionSignal();
        OnChkActionSave();
        // No action required for LatchLine check


        // Update number of seconds to save

        ComputeTriggerTimeToSave();


        // Disable OK button if not in interactive mode

        if ( !pAPP->Ini.DoInteractive ) {
                CWnd *pWnd = GetDlgItem( IDC_BTN_T_OK );
                if (pWnd)
                        pWnd->EnableWindow( FALSE );
                }


        return TRUE;  // return TRUE unless you set the focus to a control
                      // EXCEPTION: OCX Property Pages should return FALSE
}

void CTriggerDlg::SetTriggerTypeRadio( ) {

        TRACE0("CTriggerDlg::SetTriggerTypeRadio\n");

        int ibtn;

        if ( (m_TriggerType == TRIG_DIGLEVEL) && (!m_HasDigio))
                m_TriggerType = TRIG_NONE;

        if ( (m_TriggerType == TRIG_DIGEDGE) && (!m_HasDigio))
                m_TriggerType = TRIG_NONE;

        ibtn = IDC_OPT_T_TRIGGERNONE + m_TriggerType;
        CheckRadioButton( IDC_OPT_T_TRIGGERNONE, IDC_OPT_T_TRIGGERDIGEDGE, ibtn );
}

void CTriggerDlg::SetTriggerLevelRadio( ) {

        TRACE0("CTriggerDlg::SetTriggerLevelRadio\n");

        int ibtn;

        if ( m_TriggerActiveLevel == TRIGLVL_LO )
                ibtn = IDC_OPT_T_LEVELLOW;
        else
                ibtn = IDC_OPT_T_LEVELHIGH;
        
        CheckRadioButton( IDC_OPT_T_LEVELLOW, IDC_OPT_T_LEVELHIGH, ibtn );
}

void CTriggerDlg::SetTriggerEdgeRadio( ) {

        TRACE0("CTriggerDlg::SetTriggerEdgeRadio\n");

        int ibtn;

        if ( m_TriggerActiveEdge == TRIGEDGE_FALLING )
                ibtn = IDC_OPT_T_EDGEFALL;
        else
                ibtn = IDC_OPT_T_EDGERISE;
        
        CheckRadioButton( IDC_OPT_T_EDGEFALL, IDC_OPT_T_EDGERISE, ibtn );
}

void CTriggerDlg::SetTriggerChannelCheckBoxes( ) {

        TRACE0("CTriggerDlg::SetTriggerChannelCheckBoxes\n");

        int i;

        // Enable the appropriate number of check boxes for PAR 1,4,8 CH
        
        for ( i = 0 ; i < 8 ; i++ ) {
                pAPP->TmpString.Format( "Ch%1d - %s", i, &pAPP->Ini.pScanTitle[i*MAXTITLE] );
                m_CHK_TriggerChan[i].SetWindowText( pAPP->TmpString );
                if (i < pAPP->MaxAnalogChannel) {
                        m_AnalogChannelExist[i] = TRUE;
                        m_CHK_TriggerChan[i].EnableWindow( TRUE );
                        m_CHK_TriggerChan[i].SetCheck( pAPP->Ini.pScanTrigger[i] );
                        }
                else {
                        m_AnalogChannelExist[i] = FALSE;
                        m_CHK_TriggerChan[i].EnableWindow( FALSE );
                        // Don't set check default since these don't exist
                        }
                }

}

void CTriggerDlg::OnCancel() {

        TRACE0("CTriggerDlg::OnCancel\n");

        CScopeDialog::OnCancel();
}

void CTriggerDlg::OnOK() {

        int i;

        TRACE0("CTriggerDlg::OnOK\n");

        if (!CheckFocus())
                return;

        CScopeDialog::OnOK();


        // Retrived check state of check boxes

        m_TriggerActionSignal = m_CHK_TriggerActionSignal.GetCheck( );
        m_TriggerActionSave   = m_CHK_TriggerActionSave.GetCheck( );
        m_TriggerLatchLine    = m_CHK_TriggerLatchLine.GetCheck( );



        // Update globals

        pAPP->Ini.TrgType         = m_TriggerType;
        pAPP->Ini.TrgThreshold    = m_TriggerThreshold;
        pAPP->Ini.TrgDigLevelLine = m_TriggerDigLevelLine;
        pAPP->Ini.TrgDigEdgeLine  = m_TriggerDigEdgeLine;
        pAPP->Ini.TrgActiveLevel  = m_TriggerActiveLevel;
        pAPP->Ini.TrgActiveEdge   = m_TriggerActiveEdge;
        pAPP->Ini.TrgActionSave   = m_TriggerActionSave;
        pAPP->Ini.TrgFilesToSave  = m_TriggerFilesToSave;
        pAPP->Ini.TrgActionSignal = m_TriggerActionSignal;
        pAPP->Ini.TrgLineToSignal = m_TriggerLineToSignal;
        pAPP->Ini.TrgLatchLine    = m_TriggerLatchLine;


        // Update analog trigger channels

        for ( i = 0 ; i < pAPP->MaxAnalogChannel ; i++ )
                pAPP->Ini.pScanTrigger[i] = m_CHK_TriggerChan[i].GetCheck();

}

void CTriggerDlg::OnHelp() {

        TRACE0("CTriggerDlg::OnHelp\n");

        if (!CheckFocus())
                return;


        pAPP->ShowHelp( IDR_HELP_TRIG_TITLE, IDR_HELP_TRIG_TEXT, this );
}

LONG CTriggerDlg::OnEditLostFocus( UINT nID, LONG xxx ) {

        TRACE1("CTriggerDlg::OnEditLostFocus (%u)\n", nID);

        BOOL good;


        if (!DoValidate) {
                return( 1 );
                }


        // Don't allow validation when already doing validation

        DoValidate = FALSE;

        // Validate various edit controls

        switch ( nID ) {

                case IDC_TXT_T_SAVEFILES:

                        good = ValidateAsInt( IDC_TXT_T_SAVEFILES,
                                              1,
                                              999,
                                              &m_TriggerFilesToSave,
                                              "Number of files to save" );
                        if (good)
                                ComputeTriggerTimeToSave();

                        break;




                case IDC_TXT_T_SIGNALLINE:

                        good = ValidateAsInt( IDC_TXT_T_SIGNALLINE,
                                              0,
                                              pAPP->MaxTriggerDigioLine,
                                              &m_TriggerLineToSignal,
                                              "Digital output line to signal" );
                        break;




                case IDC_TXT_T_THRESHOLD:

                        good = ValidateAsLong( IDC_TXT_T_THRESHOLD,
                                              0,
                                              pAPP->MaxTriggerThreshold,
                                              &m_TriggerThreshold,
                                              "Trigger amplitude threshold" );
                        break;


                case IDC_TXT_T_DIGLEVELLINE:

                        good = ValidateAsInt( IDC_TXT_T_DIGLEVELLINE,
                                              0,
                                              pAPP->MaxTriggerDigioLine,
                                              &m_TriggerDigLevelLine,
                                              "Trigger digital level line" );
                        break;



                case IDC_TXT_T_DIGEDGELINE:

                        good = ValidateAsInt( IDC_TXT_T_DIGEDGELINE,
                                              0,
                                              pAPP->MaxTriggerDigioLine,
                                              &m_TriggerDigEdgeLine,
                                              "Trigger digital edge line" );
                        break;



                default:
                        good = TRUE;
                        break;

                }


        // Re-allow validation

        DoValidate = TRUE;


        if (good)
                return( 1 );
        else
                return( 0 );
}


void CTriggerDlg::OnBtnAllOn() {
        TRACE0("CTriggerDlg::OnBtnAllOn\n");

        if (!CheckFocus())
                return;

        SetAllState( CHECKED );
}

void CTriggerDlg::OnBtnAllOff() {
        TRACE0("CTriggerDlg::OnBtnAllOff\n");

        if (!CheckFocus())
                return;

        SetAllState( UNCHECKED );
}

void CTriggerDlg::SetAllState( int TriggerState ) {

        int i;

        // Update analog trigger channels

        for ( i = 0 ; i < pAPP->MaxAnalogChannel ; i++ )
                m_CHK_TriggerChan[i].SetCheck( TriggerState );


        // WCT: In the future, if channels can be deactivated, we
        // may want to add a check like the following to the update
        // loops above.
        // if ( pAPP->Ini.pScanActive[i] == CHECKED )

}

void CTriggerDlg::OnChkActionSave() {
        TRACE1("CTriggerDlg::OnChkActionSave (%d)\n", IsDlgButtonChecked(IDC_CHK_T_ACTIONSAVE));

        int state;

        // Quick exit for continuous recording since
        // these controls should be disabled

        if (m_TriggerType == TRIG_NONE)
                return;


        // Control whether num files and seconds fields are enabled

        if ( IsDlgButtonChecked(IDC_CHK_T_ACTIONSAVE) )
                state = TRUE;
        else
                state = FALSE;

        GetDlgItem(IDC_LAB_T_SAVEFILES)->EnableWindow( state );
        GetDlgItem(IDC_TXT_T_SAVEFILES)->EnableWindow( state );

        GetDlgItem(IDC_LAB_T_SAVESEC   )->EnableWindow( state );
        GetDlgItem(IDC_LAB_T_SAVESECNUM)->EnableWindow( state );
        
}

void CTriggerDlg::OnChkActionSignal() {
        TRACE1("CTriggerDlg::OnChkActionSignal (%d)\n", IsDlgButtonChecked(IDC_CHK_T_ACTIONSIGNAL));

        int state;

        // Quick exit for continuous recording since
        // these controls should be disabled

        if (m_TriggerType == TRIG_NONE)
                return;


        // Control whether digital line field is enabled

        if ( IsDlgButtonChecked(IDC_CHK_T_ACTIONSIGNAL) )
                state = TRUE;
        else
                state = FALSE;

        GetDlgItem(IDC_LAB_T_SIGNALLINE)->EnableWindow( state );
        GetDlgItem(IDC_TXT_T_SIGNALLINE)->EnableWindow( state );

        GetDlgItem(IDC_CHK_T_LATCHLINE )->EnableWindow( state );
        
}

void CTriggerDlg::OnKillfocusTxtTriggerFilesToSave() {

        TRACE0("CTriggerDlg::OnKillfocusTxtTriggerFilesToSave\n");

        PostMessage( WM_EDITLOSTFOCUS, IDC_TXT_T_SAVEFILES, 0 );
}

void CTriggerDlg::OnKillfocusTxtTriggerLineToSignal() {

        TRACE0("CTriggerDlg::OnKillfocusTxtTriggerLineToSignal\n");

        PostMessage( WM_EDITLOSTFOCUS, IDC_TXT_T_SIGNALLINE, 0 );
}

void CTriggerDlg::OnKillfocusTxtTriggerThreshold() {

        TRACE0("CTriggerDlg::OnKillfocusTxtTriggerThreshold\n");

        PostMessage( WM_EDITLOSTFOCUS, IDC_TXT_T_THRESHOLD, 0 );
}

void CTriggerDlg::OnKillfocusTxtTriggerDigLevelLine() {

        TRACE0("CTriggerDlg::OnKillfocusTxtTriggerDigLevelLine\n");

        PostMessage( WM_EDITLOSTFOCUS, IDC_TXT_T_DIGLEVELLINE, 0 );
}

void CTriggerDlg::OnKillfocusTxtTriggerDigEdgeLine() {

        TRACE0("CTriggerDlg::OnKillfocusTxtTriggerDigEdgeLine\n");

        PostMessage( WM_EDITLOSTFOCUS, IDC_TXT_T_DIGEDGELINE, 0 );
}

void CTriggerDlg::ComputeTriggerTimeToSave( void ) {

        TRACE0("CTriggerDlg::ComputeTriggerTimeToSave\n");

        int nsec;

        nsec = m_TriggerFilesToSave * pAPP->Ini.OutNblock * pAPP->Ini.Nsamples;
        nsec = (int)((float)nsec / pAPP->Ini.AcqSPS);

        SetDlgItemInt( IDC_LAB_T_SAVESECNUM, nsec );
}

void CTriggerDlg::OnOptTriggerType( void ) {

        BOOL TriggerState, ThreshState, LevelState, EdgeState;
        int  i, chkbtn;

        // Determine which trigger type radio button is checked

        chkbtn = GetCheckedRadioButton( IDC_OPT_T_TRIGGERNONE, IDC_OPT_T_TRIGGERDIGEDGE );
        m_TriggerType = chkbtn - IDC_OPT_T_TRIGGERNONE;


        // Control whether trigger output action fields are enabled

        if (chkbtn == IDC_OPT_T_TRIGGERNONE)
                TriggerState = FALSE;
        else
                TriggerState = TRUE;

        // Bulk turn on/off

        GetDlgItem(IDC_FRA_T_TRIGGERACTION)->EnableWindow( TriggerState );
        GetDlgItem(IDC_CHK_T_ACTIONSIGNAL)->EnableWindow( TriggerState );
        GetDlgItem(IDC_CHK_T_ACTIONSAVE)->EnableWindow( TriggerState );
        GetDlgItem(IDC_LAB_T_SIGNALLINE)->EnableWindow( TriggerState );
        GetDlgItem(IDC_TXT_T_SIGNALLINE)->EnableWindow( TriggerState );
        GetDlgItem(IDC_CHK_T_LATCHLINE)->EnableWindow( TriggerState );
        GetDlgItem(IDC_LAB_T_SAVEFILES)->EnableWindow( TriggerState );
        GetDlgItem(IDC_TXT_T_SAVEFILES)->EnableWindow( TriggerState );
        GetDlgItem(IDC_LAB_T_SAVESEC)->EnableWindow( TriggerState );
        GetDlgItem(IDC_LAB_T_SAVESECNUM)->EnableWindow( TriggerState );

        // Fine tuning turn on/off

        if (TriggerState) {
                OnChkActionSignal();
                OnChkActionSave();
                }


        // Control whether threshold and all trigger channel fields are enabled

        if (chkbtn == IDC_OPT_T_TRIGGERTHRESH)
                ThreshState = TRUE;
        else
                ThreshState = FALSE;

        // Bulk turn on/off

        GetDlgItem(IDC_LAB_T_THRESHOLD)->EnableWindow( ThreshState );
        GetDlgItem(IDC_TXT_T_THRESHOLD)->EnableWindow( ThreshState );
        GetDlgItem(IDC_FRA_T_CHANNEL)->EnableWindow( ThreshState );
        GetDlgItem(IDC_BTN_T_ALLOFF)->EnableWindow( ThreshState );
        GetDlgItem(IDC_BTN_T_ALLON)->EnableWindow( ThreshState );

        for ( i = 0 ; i < 8 ; i++ )
                m_CHK_TriggerChan[i].EnableWindow( ThreshState &
                                                   m_AnalogChannelExist[i] );


        // Control whether dig level line and high/low fields are enabled

        if (chkbtn == IDC_OPT_T_TRIGGERDIGLEVEL)
                LevelState = TRUE;
        else
                LevelState = FALSE;

        GetDlgItem(IDC_LAB_T_DIGLEVELLINE)->EnableWindow( LevelState );
        GetDlgItem(IDC_TXT_T_DIGLEVELLINE)->EnableWindow( LevelState );
        GetDlgItem(IDC_OPT_T_LEVELHIGH)->EnableWindow( LevelState );
        GetDlgItem(IDC_OPT_T_LEVELLOW)->EnableWindow( LevelState );
        


        // Control whether dig edge line and rise/fall fields are enabled

        if (chkbtn == IDC_OPT_T_TRIGGERDIGEDGE)
                EdgeState = TRUE;
        else
                EdgeState = FALSE;

        GetDlgItem(IDC_LAB_T_DIGEDGELINE)->EnableWindow( EdgeState );
        GetDlgItem(IDC_TXT_T_DIGEDGELINE)->EnableWindow( EdgeState );
        GetDlgItem(IDC_OPT_T_EDGEFALL)->EnableWindow( EdgeState );
        GetDlgItem(IDC_OPT_T_EDGERISE)->EnableWindow( EdgeState );

}

void CTriggerDlg::OnOptTriggerLevel( void ) {

        int  chkbtn;

        // Determine which trigger level radio button is checked

        chkbtn = GetCheckedRadioButton( IDC_OPT_T_LEVELLOW, IDC_OPT_T_LEVELHIGH );
        if (chkbtn == IDC_OPT_T_LEVELLOW)
                m_TriggerActiveLevel = TRIGLVL_LO;
        else
                m_TriggerActiveLevel = TRIGLVL_HI;
                
}
void CTriggerDlg::OnOptTriggerEdge( void ) {

        int  chkbtn;

        // Determine which trigger edge radio button is checked

        chkbtn = GetCheckedRadioButton( IDC_OPT_T_EDGEFALL, IDC_OPT_T_EDGERISE );
        if (chkbtn == IDC_OPT_T_EDGEFALL)
                m_TriggerActiveEdge = TRIGEDGE_FALLING;
        else
                m_TriggerActiveEdge = TRIGEDGE_RISING;
                
}







//***********************************************************************
//
// CLASS COutputDlg
//
// This class, derived from CScopeDialog, presents and controls the Output
// file dialog.  This dialog allows the user to specify the naming
// convention and directory for the output binary data files.  A binary
// header file called SCOPE.HDR is also written to that directory.  The
// user also has some limited control over the size of the resulting files.
//

BEGIN_MESSAGE_MAP(COutputDlg, CScopeDialog)
        //{{AFX_MSG_MAP(COutputDlg)
        ON_MESSAGE(WM_EDITLOSTFOCUS,          OnEditLostFocus)
        ON_BN_CLICKED(IDC_BTN_F_OK,           OnOK)
        ON_BN_CLICKED(IDC_BTN_F_CANCEL,       OnCancel)
        ON_BN_CLICKED(IDC_BTN_F_HELP,         OnHelp)
        ON_BN_CLICKED(IDC_OPT_F_FILEFMT_NONE, OnOptFileFmt)
        ON_BN_CLICKED(IDC_OPT_F_FILEFMT_OUT,  OnOptFileFmt)
        ON_BN_CLICKED(IDC_OPT_F_FILEFMT_DAT,  OnOptFileFmt)
        ON_BN_CLICKED(IDC_OPT_F_FILE_ONE,     OnOptFileMode)
        ON_BN_CLICKED(IDC_OPT_F_FILE_SEQ,     OnOptFileMode)
        ON_BN_CLICKED(IDC_OPT_F_FILE_TIME,    OnOptFileMode)
        ON_EN_KILLFOCUS(IDC_TXT_F_NBLOCK,     OnKillfocusTxtNblock)
        ON_EN_KILLFOCUS(IDC_TXT_F_NSAMPLE,    OnKillfocusTxtNsample)
        ON_EN_KILLFOCUS(IDC_TXT_F_SEQSTART,   OnKillfocusTxtSeqStart)
        ON_EN_KILLFOCUS(IDC_TXT_F_WORKDIR,    OnKillfocusTxtWorkDir)
        //}}AFX_MSG_MAP
END_MESSAGE_MAP()




COutputDlg::COutputDlg(CWnd* pParent /*=NULL*/)
              : CScopeDialog(COutputDlg::IDD, pParent) {

        TRACE0("COutputDlg::COutputDlg\n");

        InitData();

}


COutputDlg::COutputDlg(BOOL ROnsamples, CWnd* pParent /*=NULL*/)
              : CScopeDialog(COutputDlg::IDD, pParent) {

        TRACE0("COutputDlg::COutputDlg\n");

        InitData();

        m_ROnsamples = ROnsamples;
}


void COutputDlg::DoDataExchange(CDataExchange* pDX) {

        TRACE0("COutputDlg::DoDataExchange\n");

        CScopeDialog::DoDataExchange(pDX);

        //{{AFX_DATA_MAP(COutputDlg)
        DDX_Text(pDX, IDC_TXT_F_NBLOCK,   m_Nblock);
        DDX_Text(pDX, IDC_TXT_F_NSAMPLE,  m_Nsamples);
        DDX_Text(pDX, IDC_TXT_F_SEQSTART, m_SeqStart);
        DDX_Text(pDX, IDC_TXT_F_WORKDIR,  m_WorkDir);
        //}}AFX_DATA_MAP
}

void COutputDlg::InitData(void) {

        TRACE0("COutputDlg::InitData\n");

        //{{AFX_DATA_INIT(COutputDlg)
        m_ROnsamples = FALSE;
        m_Nblock     = 0;
        m_Nsamples   = 0;
        m_FileFmt    = 0;
        m_FileName   = 0;
        m_SeqStart   = 0;
        m_WorkDir    = _T("");
        //}}AFX_DATA_INIT
}

BOOL COutputDlg::OnInitDialog() {

        int chkbtn;

        TRACE0("COutputDlg::OnInitDialog\n");


        // Initialize all member variables from application globals

        m_Nsamples  = pAPP->Ini.Nsamples;
        m_Nblock    = pAPP->Ini.OutNblock;
        m_FileFmt   = pAPP->Ini.OutFileFormat;
        m_FileName  = pAPP->Ini.OutFileNaming;
        m_SeqStart  = pAPP->Ini.OutSeqStart;
        m_WorkDir   = pAPP->Ini.OutDir;

        CScopeDialog::OnInitDialog( IDC_BTN_F_OK );



        // Maybe set Nsamples field to be read only

        if (m_ROnsamples)
                SetTextReadOnly( IDC_TXT_F_NSAMPLE );




        // Update filesize

        ComputeFileSize();



        // Set format radio button

        CheckRadioButton( IDC_OPT_F_FILEFMT_NONE, IDC_OPT_F_FILEFMT_DAT,
                                        (IDC_OPT_F_FILEFMT_NONE + m_FileFmt) );

        OnOptFileFmt();



        // Set naming convention radio button

        chkbtn = IDC_OPT_F_FILE_ONE + m_FileName-1;

        CheckRadioButton( IDC_OPT_F_FILE_ONE, IDC_OPT_F_FILE_TIME, chkbtn );

        OnOptFileMode();



        return TRUE;  // return TRUE unless you set the focus to a control
                      // EXCEPTION: OCX Property Pages should return FALSE
}


void COutputDlg::OnCancel() {

        TRACE0("COutputDlg::OnCancel\n");

        CScopeDialog::OnCancel();
}

void COutputDlg::OnOK() {

        TRACE0("COutputDlg::OnOK\n");


        if (!CheckFocus())
                return;


        CScopeDialog::OnOK();


        // Update globals

        pAPP->Ini.Nsamples      = m_Nsamples;
        pAPP->Ini.OutFileFormat = m_FileFmt;
        pAPP->Ini.OutFileNaming = m_FileName;
        pAPP->Ini.OutSeqStart   = m_SeqStart;
        pAPP->Ini.OutNblock     = m_Nblock;
        pAPP->Ini.OutDir        = m_WorkDir;
}

void COutputDlg::OnHelp() {

        TRACE0("COutputDlg::OnHelp\n");

        if (!CheckFocus())
                return;

        pAPP->ShowHelp( IDR_HELP_OUTPUT_TITLE, IDR_HELP_OUTPUT_TEXT, this );
}

LONG COutputDlg::OnEditLostFocus( UINT nID, LONG xxx ) {

        TRACE1("COutputDlg::OnEditLostFocus (%u)\n", nID);

        CString tmpStr;
        BOOL    good;
        int     maxValue;


        if (!DoValidate) {
                return( 1 );
                }


        // Don't allow validation when already doing validation

        DoValidate = FALSE;

        // Validate various edit controls

        switch ( nID ) {

                case IDC_TXT_F_NSAMPLE:

                        if (pAPP->MaxSample == 0)
                                maxValue = 9999;
                        else
                                maxValue = pAPP->MaxSample;

                        good = ValidateAsInt( IDC_TXT_F_NSAMPLE,
                                                0,
                                                maxValue,
                                                &m_Nsamples,
                                                "Samples per trace" );

                        if (good)
                                ComputeFileSize();

                        break;



                case IDC_TXT_F_NBLOCK:

                        good = ValidateAsInt( IDC_TXT_F_NBLOCK,
                                                0,
                                                9999,
                                                &m_Nblock,
                                                "Blocks per file" );

                        if (good)
                                ComputeFileSize();

                        break;



                case IDC_TXT_F_WORKDIR:

                        GetDlgItemText( IDC_TXT_F_WORKDIR, tmpStr );

                        if (!SrFileExist((LPCTSTR)tmpStr)) {
                                AfxMessageBox( "Working directory must already exist." );
                                SetDlgItemText( IDC_TXT_F_WORKDIR, m_WorkDir );
                                GotoDlgCtrl( GetDlgItem(IDC_TXT_F_WORKDIR) );
                                good = FALSE;
                                }
                        else {
                                GetDlgItemText( IDC_TXT_F_WORKDIR, m_WorkDir );
                                good = TRUE;
                                }

                        break;



                case IDC_TXT_F_SEQSTART:

                        good = ValidateAsInt( IDC_TXT_F_SEQSTART,
                                                0,
                                                0x7FFFFFFF,
                                                &m_SeqStart,
                                                "Starting sequential file" );
                        break;



                default:
                        good = TRUE;
                        break;


                }


        // Re-allow validation

        DoValidate = TRUE;


        if (good)
                return( 1 );
        else
                return( 0 );
}


void COutputDlg::OnKillfocusTxtNsample() {

        TRACE0("COutputDlg::OnKillfocusTxtNsample\n");

        PostMessage( WM_EDITLOSTFOCUS, IDC_TXT_F_NSAMPLE, 0 );
}

void COutputDlg::OnKillfocusTxtNblock() {

        TRACE0("COutputDlg::OnKillfocusTxtNblock\n");

        PostMessage( WM_EDITLOSTFOCUS, IDC_TXT_F_NBLOCK, 0 );
}

void COutputDlg::ComputeFileSize() {

        TRACE0("COutputDlg::ComputeFileSize\n");

        long fs, gpsbytespersec, nsecperfile;


        // Compute basic file size from number of channels,
        // samples, and output blocks.
        
        fs = pAPP->Ini.Nchannels * m_Nsamples * m_Nblock * pAPP->DataPtSize;

        
        // Extend size to include DAT headers and tags.

        if ( m_FileFmt == FILEFMT_DAT )
                fs += (SRDAT_HDR_TOTALSIZE + SRDAT_TAGSIZE * (m_Nblock+1));

        
        // Extend size to include GPS parameters.  We acquire 1 PPS data
        // and 1 serial data structure per second.  The number of
        // seconds per file is determined from the number of samples,
        // the number of output blocks, and the sampling rate.

        gpsbytespersec = sizeof( PPSTDATA ) + sizeof( SERIALDATA );
        nsecperfile    = (long)((m_Nsamples * m_Nblock / pAPP->Ini.AcqSPS) + 0.5);

        if ( pAPP->Ini.GpsEnable )
                fs += (gpsbytespersec * nsecperfile);


        // Display final size value

        SetDlgItemInt( IDC_LAB_F_SIZEINBYTES, fs );
}

void COutputDlg::OnOptFileFmt() {

        TRACE0("COutputDlg::OnOptFileFmt\n");

        BOOL FmtState, SeqState;
        int  chkbtn;

        chkbtn = GetCheckedRadioButton( IDC_OPT_F_FILEFMT_NONE,
                                        IDC_OPT_F_FILEFMT_DAT );
        m_FileFmt = chkbtn - IDC_OPT_F_FILEFMT_NONE;

        if (chkbtn == IDC_OPT_F_FILEFMT_NONE)
                FmtState = FALSE;
        else
                FmtState = TRUE;


        GetDlgItem(IDC_FRA_F_FILENAME )->EnableWindow( FmtState );
        GetDlgItem(IDC_OPT_F_FILE_ONE )->EnableWindow( FmtState );
        GetDlgItem(IDC_OPT_F_FILE_SEQ )->EnableWindow( FmtState );
        GetDlgItem(IDC_OPT_F_FILE_TIME)->EnableWindow( FmtState );
        GetDlgItem(IDC_LAB_F_WORKDIR  )->EnableWindow( FmtState );
        GetDlgItem(IDC_TXT_F_WORKDIR  )->EnableWindow( FmtState );

        if ( FmtState && (m_FileName == FILE_SEQDEC) )
                SeqState = TRUE;
        else
                SeqState = FALSE;

        GetDlgItem(IDC_LAB_F_SEQSTART )->EnableWindow( SeqState );
        GetDlgItem(IDC_TXT_F_SEQSTART )->EnableWindow( SeqState );

        ComputeFileSize();
}


void COutputDlg::OnOptFileMode() {

        TRACE0("COutputDlg::OnOptFileMode\n");

        BOOL SeqState;
        int  chkbtn;

        chkbtn = GetCheckedRadioButton( IDC_OPT_F_FILE_ONE,
                                        IDC_OPT_F_FILE_TIME );

        if ( chkbtn == IDC_OPT_F_FILE_TIME )
                m_FileName = FILE_TIME;
        else if ( chkbtn == IDC_OPT_F_FILE_SEQ )
                m_FileName = FILE_SEQDEC;
        else // if ( chkbtn == IDC_OPT_F_FILE_ONE )
                m_FileName = FILE_ONE;


        // When sequential mode is selected,
        // the start num label and text are enabled

        if (chkbtn == IDC_OPT_F_FILE_SEQ)
                SeqState = TRUE;
        else
                SeqState = FALSE;

        GetDlgItem(IDC_LAB_F_SEQSTART)->EnableWindow( SeqState );
        GetDlgItem(IDC_TXT_F_SEQSTART)->EnableWindow( SeqState );
}


void COutputDlg::OnKillfocusTxtWorkDir() {

        TRACE0("COutputDlg::OnKillfocusTxtWorkDir\n");

        PostMessage( WM_EDITLOSTFOCUS, IDC_TXT_F_WORKDIR, 0 );
}


void COutputDlg::OnKillfocusTxtSeqStart() {

        TRACE0("COutputDlg::OnKillfocusTxtSeqStart\n");

        PostMessage( WM_EDITLOSTFOCUS, IDC_TXT_F_SEQSTART, 0 );
}





//***********************************************************************
//
// CLASS CDisplayDlg
//
// This class, derived from CScopeDialog, presents and controls the Display
// dialog.  This dialog allows the user to specify various display options
// such as the number of traces to be displayed, the scaling applied
// to those traces, and the overall color scheme.

BEGIN_MESSAGE_MAP(CDisplayDlg, CScopeDialog)
        //{{AFX_MSG_MAP(CDisplayDlg)
        ON_MESSAGE(WM_EDITLOSTFOCUS,           OnEditLostFocus)
        ON_BN_CLICKED(IDC_BTN_D_OK,            OnOK)
        ON_BN_CLICKED(IDC_BTN_D_CANCEL,        OnCancel)
        ON_BN_CLICKED(IDC_BTN_D_HELP,          OnHelp)
        ON_EN_KILLFOCUS(IDC_TXT_D_SHIFTBY,     OnKillfocusTxtShiftby)
        ON_EN_KILLFOCUS(IDC_TXT_D_NDISPLAY,    OnKillfocusTxtNdisplay)
        ON_EN_KILLFOCUS(IDC_TXT_D_FIRSTSAMPLE, OnKillfocusTxtFirstsample)
        ON_EN_KILLFOCUS(IDC_TXT_D_FIRSTTRACE,  OnKillfocusTxtFirsttrace)
        ON_BN_CLICKED(IDC_CHK_D_CLIPPING,      OnChkClipping)
        ON_BN_CLICKED(IDC_CHK_D_AUTODC,        OnChkAutoDC)
        ON_BN_CLICKED(IDC_OPT_D_COLORMONO,     OnOptColorScheme)
        ON_BN_CLICKED(IDC_OPT_D_COLORGREEN,    OnOptColorScheme)
        ON_BN_CLICKED(IDC_OPT_D_COLORBLUE,     OnOptColorScheme)
        ON_BN_CLICKED(IDC_OPT_D_COLORBROWN,    OnOptColorScheme)
        //}}AFX_MSG_MAP
//      ON_CONTROL_RANGE(BN_CLICKED, IDC_OPT_D_COLORMONO, IDC_OPT_D_COLORBROWN, OnOptColorScheme)
END_MESSAGE_MAP()




CDisplayDlg::CDisplayDlg(CWnd* pParent /*=NULL*/)
           : CScopeDialog(CDisplayDlg::IDD, pParent) {

        //{{AFX_DATA_INIT(CDisplayDlg)
        m_ShiftBy     = 0;
        m_Ndisplay    = 0;
        m_FirstSample = 0;
        m_FirstTrace  = 0;
        m_ColorScheme = 0;
        m_TraceWidth  = 0;
        m_Clipping    = FALSE;
        m_AutoDC      = FALSE;
        //}}AFX_DATA_INIT
}


void CDisplayDlg::DoDataExchange(CDataExchange* pDX) {

        CScopeDialog::DoDataExchange(pDX);

        //{{AFX_DATA_MAP(CDisplayDlg)
        DDX_Control(pDX, IDC_TXT_D_SHIFTBY,     m_TXT_ShiftBy);
        DDX_Control(pDX, IDC_TXT_D_NDISPLAY,    m_TXT_Ndisplay);
        DDX_Control(pDX, IDC_TXT_D_FIRSTSAMPLE, m_TXT_FirstSample);
        DDX_Control(pDX, IDC_TXT_D_FIRSTTRACE,  m_TXT_FirstTrace);
        DDX_Control(pDX, IDC_LAB_D_PLOTBITS,    m_LAB_PlotBits);
        DDX_Control(pDX, IDC_LAB_D_NSCAN,       m_LAB_Nscan);
        DDX_Control(pDX, IDC_LAB_D_LASTSAMPLE,  m_LAB_LastSample);
        DDX_Control(pDX, IDC_CHK_D_CLIPPING,    m_CHK_Clipping);
        DDX_Control(pDX, IDC_CHK_D_AUTODC,      m_CHK_AutoDC);
        DDX_Control(pDX, IDC_OPT_D_COLORMONO,   m_OPT_ColorMono);
        DDX_Control(pDX, IDC_OPT_D_COLORGREEN,  m_OPT_ColorGreen);
        DDX_Control(pDX, IDC_OPT_D_COLORBLUE,   m_OPT_ColorBlue);
        DDX_Control(pDX, IDC_OPT_D_COLORBROWN,  m_OPT_ColorBrown);
        DDX_Text(pDX, IDC_TXT_D_SHIFTBY,     m_ShiftBy);
        DDX_Text(pDX, IDC_TXT_D_NDISPLAY,    m_Ndisplay);
        DDX_Text(pDX, IDC_TXT_D_FIRSTSAMPLE, m_FirstSample);
        DDX_Text(pDX, IDC_TXT_D_FIRSTTRACE,  m_FirstTrace);
        //}}AFX_DATA_MAP
}


BOOL CDisplayDlg::OnInitDialog() {

        // Initialize all member variables from application globals

        m_ShiftBy     = pAPP->Ini.PltShiftBy;
        m_Ndisplay    = pAPP->Ini.PltNdisplay;
        m_FirstSample = pAPP->Ini.PltFirstSample;
        m_FirstTrace  = pAPP->Ini.PltFirstTrace;
        m_ColorScheme = pAPP->Ini.PltColorScheme;
        m_Clipping    = pAPP->Ini.PltClipping;
        m_AutoDC      = pAPP->Ini.PltAutoDC;


        CScopeWindow *pWnd = (CScopeWindow *)m_pParentWnd;
        if (pWnd)
                m_TraceWidth = pWnd->TraceWidth;
        else
                m_TraceWidth = 500;




        // Initialize base class and initialize text boxes (via DDX)

        CScopeDialog::OnInitDialog( 0 ); // use 0 not IDC_BTN_D_OK because
                                         // display dialog options are allowed
                                         // even in non-interactive mode


        // Initialize labels and radio buttons

        SetPlotBitsLabel();
        SetNscanLabel();
        SetLastSampleLabel();
        SetColorSchemeRadio();

        if (m_Clipping)
                m_CHK_Clipping.SetCheck(CHECKED);
        else
                m_CHK_Clipping.SetCheck(UNCHECKED);


        if (m_AutoDC)
                m_CHK_AutoDC.SetCheck(CHECKED);
        else
                m_CHK_AutoDC.SetCheck(UNCHECKED);



        return TRUE;  // return TRUE unless you set the focus to a control
                      // EXCEPTION: OCX Property Pages should return FALSE
}

void CDisplayDlg::SetPlotBitsLabel( ) {

        CString str;
        int nscan;

        nscan = pAPP->MaxAmpShift - m_ShiftBy;
        str.Format("Plot %d Bits", nscan );
        CStatic *Label = (CStatic*)GetDlgItem(IDC_LAB_D_PLOTBITS);
        Label->SetWindowText( str );
}

void CDisplayDlg::SetNscanLabel( ) {
        CString str;
        str.Format("Scan %d Traces", pAPP->NchannelsPlot);
        m_LAB_Nscan.SetWindowText( str );
}

void CDisplayDlg::SetLastSampleLabel( ) {

        sprintf(pAPP->szTemp, "Last is %d", (m_FirstSample + m_TraceWidth));
        m_LAB_LastSample.SetWindowText( pAPP->szTemp );
}

void CDisplayDlg::SetColorSchemeRadio( ) {

        TRACE0("CDisplayDlg::SetColorSchemeRadio\n");

        int ibtn;

        ibtn = IDC_OPT_D_COLORMONO + m_ColorScheme;
        CheckRadioButton( IDC_OPT_D_COLORMONO, IDC_OPT_D_COLORBROWN, ibtn );
}

void CDisplayDlg::OnCancel() {

        TRACE0("CDisplayDlg::OnCancel\n");

        CScopeDialog::OnCancel();
}

void CDisplayDlg::OnOK() {

        TRACE0("CDisplayDlg::OnOK\n");

        if (!CheckFocus())
                return;


        CScopeDialog::OnOK();


        // Update globals

        pAPP->Ini.PltShiftBy     = m_ShiftBy;
        pAPP->Ini.PltNdisplay    = m_Ndisplay;
        pAPP->Ini.PltFirstSample = m_FirstSample;
        pAPP->Ini.PltFirstTrace  = m_FirstTrace;
        pAPP->Ini.PltColorScheme = m_ColorScheme;
        pAPP->Ini.PltClipping    = m_Clipping;
        pAPP->Ini.PltAutoDC      = m_AutoDC;
}

void CDisplayDlg::OnHelp() {

        TRACE0("CDisplayDlg::OnHelp\n");

        if (!CheckFocus())
                return;


        pAPP->ShowHelp( IDR_HELP_DISP_TITLE, IDR_HELP_DISP_TEXT, this );
}

LONG CDisplayDlg::OnEditLostFocus( UINT nID, LONG xxx ) {

        TRACE1("CDisplayDlg::OnEditLostFocus (%u)\n", nID);

        int  maxFirst;
        BOOL good;


        if (!DoValidate) {
                return( 1 );
                }


        // Don't allow validation when already doing validation

        DoValidate = FALSE;

        // Validate various edit controls

        switch ( nID ) {

                case IDC_TXT_D_SHIFTBY:

                        good = ValidateAsInt( IDC_TXT_D_SHIFTBY,
                                                pAPP->MinAmpShift,
                                                pAPP->MaxAmpShift-1,
                                                &m_ShiftBy,
                                                "Shift by" );
                        if (good)
                                SetPlotBitsLabel();

                        break;



                case IDC_TXT_D_NDISPLAY:

                        good = ValidateAsInt( IDC_TXT_D_NDISPLAY,
                                                1,
                                                pAPP->NchannelsPlot,
                                                &m_Ndisplay,
                                                "Number to display" );

                        if (good) {
//                              AdjustFirstTrace();
                                maxFirst = pAPP->NchannelsPlot - m_Ndisplay;
                                if (m_FirstTrace > maxFirst) {
                                        m_FirstTrace = maxFirst;
                                        SetDlgItemInt(IDC_TXT_D_FIRSTTRACE,
                                                        m_FirstTrace);
                                        }
                                }

                        break;



                case IDC_TXT_D_FIRSTSAMPLE:

                        maxFirst = pAPP->NsamplesPlot - m_TraceWidth - 1;
                        if (maxFirst < 0)
                                maxFirst = 0;

                        good = ValidateAsInt( IDC_TXT_D_FIRSTSAMPLE,
                                                0,
                                                maxFirst,
                                                &m_FirstSample,
                                                "First sample" );

                        if (good)
                                SetLastSampleLabel();

                        break;




                case IDC_TXT_D_FIRSTTRACE:

                        maxFirst = pAPP->NchannelsPlot - m_Ndisplay;

                        good = ValidateAsInt( IDC_TXT_D_FIRSTTRACE,
                                                0,
                                                maxFirst,
                                                &m_FirstTrace,
                                                "First trace" );

                        break;

                default:
                        good = TRUE;
                        break;

                }


        // Re-allow validation

        DoValidate = TRUE;


        if (good)
                return( 1 );
        else
                return( 0 );
}


void CDisplayDlg::OnKillfocusTxtShiftby() {

        TRACE0("CDisplayDlg::OnKillfocusTxtShiftby\n");

        PostMessage( WM_EDITLOSTFOCUS, IDC_TXT_D_SHIFTBY, 0 );
}

void CDisplayDlg::OnKillfocusTxtNdisplay() {

        TRACE0("CDisplayDlg::OnKillfocusTxtNdisplay\n");

        PostMessage( WM_EDITLOSTFOCUS, IDC_TXT_D_NDISPLAY, 0 );
}


void CDisplayDlg::OnKillfocusTxtFirstsample() {

        PostMessage( WM_EDITLOSTFOCUS, IDC_TXT_D_FIRSTSAMPLE, 0 );
}

void CDisplayDlg::OnKillfocusTxtFirsttrace() {

        PostMessage( WM_EDITLOSTFOCUS, IDC_TXT_D_FIRSTTRACE, 0 );
}

void CDisplayDlg::OnChkClipping(  ) {

        m_Clipping = !m_Clipping;
}

void CDisplayDlg::OnChkAutoDC(  ) {

        m_AutoDC = !m_AutoDC;
}

void CDisplayDlg::OnOptColorScheme(  ) {

        int ibtn;

        ibtn = GetCheckedRadioButton( IDC_OPT_D_COLORMONO, IDC_OPT_D_COLORBROWN );
        m_ColorScheme = ibtn - IDC_OPT_D_COLORMONO;
}







//***********************************************************************
//
// CLASS CGpsDlg
//
// This class, derived from CScopeDialog, presents and controls the Output
// file dialog.  This dialog allows the user to specify the naming
// convention and directory for the output binary data files.  A binary
// header file called SCOPE.HDR is also written to that directory.  The
// user also has some limited control over the size of the resulting files.
//

BEGIN_MESSAGE_MAP(CGpsDlg, CScopeDialog)
        //{{AFX_MSG_MAP(CGpsDlg)
        ON_MESSAGE(WM_EDITLOSTFOCUS,           OnEditLostFocus)
        ON_BN_CLICKED(IDC_BTN_G_OK,            OnOK)
        ON_BN_CLICKED(IDC_BTN_G_CANCEL,        OnCancel)
        ON_BN_CLICKED(IDC_BTN_G_HELP,          OnHelp)
        ON_BN_CLICKED(IDC_CHK_G_GPSENABLE,     OnChkGpsEnable)
        ON_BN_CLICKED(IDC_CHK_G_DISPLAYPPS,    OnChkDisplayPps)
        ON_BN_CLICKED(IDC_CHK_G_DISPLAYSERIAL, OnChkDisplaySerial)
        ON_CBN_SELCHANGE(IDC_CBO_G_DRIVERNAME, OnSelchangeName)
        ON_CBN_KILLFOCUS(IDC_CBO_G_DRIVERNAME, OnKillfocusName)
        ON_CBN_SELCHANGE(IDC_CBO_G_GPSMODEL,   OnSelchangeModel)
        ON_CBN_KILLFOCUS(IDC_CBO_G_GPSMODEL,   OnKillfocusModel)
        ON_CBN_SELCHANGE(IDC_CBO_G_SERIALPORT, OnSelchangePort)
        ON_CBN_KILLFOCUS(IDC_CBO_G_SERIALPORT, OnKillfocusPort)
        //}}AFX_MSG_MAP
END_MESSAGE_MAP()




CGpsDlg::CGpsDlg(CWnd* pParent /*=NULL*/)
              : CScopeDialog(CGpsDlg::IDD, pParent) {

        TRACE0("CGpsDlg::CGpsDlg\n");

        //{{AFX_DATA_INIT(CGpsDlg)

        m_DriverName    = _T("");
        m_Index         = 1;
        m_SerialPort    = 1;
        m_GpsModel      = GPSMODEL_TRIMBLE;
        m_GpsEnable     = FALSE;
        m_DisplayPps    = FALSE;
        m_DisplaySerial = FALSE;

        //}}AFX_DATA_INIT

}


void CGpsDlg::DoDataExchange(CDataExchange* pDX) {

        TRACE0("CGpsDlg::DoDataExchange\n");

        CScopeDialog::DoDataExchange(pDX);

        //{{AFX_DATA_MAP(CGpsDlg)
        DDX_Control(pDX, IDC_CBO_G_DRIVERNAME, m_CBO_DriverName);
        DDX_Control(pDX, IDC_CBO_G_GPSMODEL,   m_CBO_GpsModel);
        DDX_Control(pDX, IDC_CBO_G_SERIALPORT, m_CBO_SerialPortName);
        //}}AFX_DATA_MAP
}

BOOL CGpsDlg::OnInitDialog() {

        CButton *pChk;

        TRACE0("CGpsDlg::OnInitDialog\n");


        // Initialize all member variables from application globals

        m_GpsModel      = pAPP->Ini.GpsModel;
        m_DriverName    = pAPP->Ini.GpsDriverName;
        m_SerialPort    = pAPP->Ini.GpsSerialPort;
        m_GpsEnable     = pAPP->Ini.GpsEnable;
        m_DisplayPps    = pAPP->Ini.GpsDisplayPps;
        m_DisplaySerial = pAPP->Ini.GpsDisplaySerial;


        CScopeDialog::OnInitDialog( IDC_BTN_G_OK );



        // Init combo and check boxes

        SetComboName();
        SetComboModel();
        SetComboPort();

        pChk = (CButton*)GetDlgItem( IDC_CHK_G_GPSENABLE );
        pChk->SetCheck( m_GpsEnable );

        pChk = (CButton*)GetDlgItem( IDC_CHK_G_DISPLAYPPS );
        pChk->SetCheck( m_DisplayPps );

        pChk = (CButton*)GetDlgItem( IDC_CHK_G_DISPLAYSERIAL );
        pChk->SetCheck( m_DisplaySerial );



        // Update activation state of all other controls based on GpsEnable

        SetActivation();



        // Make these items invisible until we decide how best to
        // present the GPS PPS and serial data.
        GetDlgItem(IDC_CHK_G_DISPLAYPPS   )->ShowWindow( SW_HIDE );
        GetDlgItem(IDC_CHK_G_DISPLAYSERIAL)->ShowWindow( SW_HIDE );




        return TRUE;  // return TRUE unless you set the focus to a control
                      // EXCEPTION: OCX Property Pages should return FALSE
}


void CGpsDlg::OnCancel() {

        TRACE0("CGpsDlg::OnCancel\n");

        CScopeDialog::OnCancel();
}

void CGpsDlg::OnOK() {

        CButton* pChk;

        TRACE0("CGpsDlg::OnOK\n");


        if (!CheckFocus())
                return;


        // Don't close if a GPS buffer would overflow

        if ( !pAPP->GpsTimingCheckBufferSize( m_GpsEnable,
                                              pAPP->Ini.Nsamples,
                                              pAPP->Ini.AcqSPS ) ) {
                AfxMessageBox( "Sampling rate is too low, a GPS buffer would overflow." );
                return;
                }


        CScopeDialog::OnOK();


        // Retrieve check state of check boxes

        pChk = (CButton*)GetDlgItem( IDC_CHK_G_GPSENABLE );
        m_GpsEnable = pChk->GetCheck();

        pChk = (CButton*)GetDlgItem( IDC_CHK_G_DISPLAYPPS );
        m_DisplayPps = pChk->GetCheck();

        pChk = (CButton*)GetDlgItem( IDC_CHK_G_DISPLAYSERIAL );
        m_DisplaySerial = pChk->GetCheck();



        // Update globals

        pAPP->Ini.GpsModel         = m_GpsModel;
        pAPP->Ini.GpsDriverName    = m_DriverName;
        pAPP->Ini.GpsSerialPort    = m_SerialPort;
        pAPP->Ini.GpsEnable        = m_GpsEnable;
        pAPP->Ini.GpsDisplayPps    = m_DisplayPps;
        pAPP->Ini.GpsDisplaySerial = m_DisplaySerial;

// WCT: Until there is a dialog button for controlling the counter
//      channel, we just tie it to the GPS mark channel

        if (pAPP->MaxCounterChannel > 0 && pAPP->Ini.GpsEnable)
                pAPP->Ini.CounterEnable = ON;
        else
                pAPP->Ini.CounterEnable = OFF;
}

void CGpsDlg::OnHelp() {

        TRACE0("CGpsDlg::OnHelp\n");

        if (!CheckFocus())
                return;

        pAPP->ShowHelp( IDR_HELP_GPS_TITLE, IDR_HELP_GPS_TEXT, this );
}

LONG CGpsDlg::OnEditLostFocus( UINT nID, LONG xxx ) {

        TRACE1("CGpsDlg::OnEditLostFocus (%u)\n", nID);

        CString tempstr;
        int     temp;
        BOOL    good, good1, good2;


        if (!DoValidate) {
                return( 1 );
                }


        // Don't allow validation when already doing validation

        DoValidate = FALSE;
        good1      = TRUE;



        // Validate various edit controls

        switch ( nID ) {

#ifdef NOT_YET
                             // Just fall through, with no break since
                             // AFX_IDC_FONTNAMES somehow has focus
                case 1001:   // after enter from typing in a DropDown
                             // CBO.  So validate them both to be safe.

                        good1 = TRUE;

                        // break;
#endif // NOT_YET

                case IDC_CBO_G_DRIVERNAME:

                        GetDlgItemText( IDC_CBO_G_DRIVERNAME, pAPP->szTemp, MAXSTR );
                        tempstr = pAPP->szTemp;

                        if ( sscanf((LPCTSTR)tempstr,"SrParGps",&temp) == 1 ) {
                                good2 = TRUE;
                                m_DriverName = tempstr;
                                }
                        else
                                good2 = FALSE;

                        SetComboName();

                        good = (good1 && good2);

                        break;



                case IDC_CBO_G_GPSMODEL:

                        GetDlgItemText( IDC_CBO_G_GPSMODEL, pAPP->szTemp, MAXSTR );
                        tempstr = pAPP->szTemp;

                        if ( sscanf((LPCTSTR)tempstr,"GARMIN",&temp) == 1 ) {
                                good2 = TRUE;
                                m_GpsModel = GPSMODEL_GARMIN;
                                }
                        else if ( sscanf((LPCTSTR)tempstr,"TRIMBLE",&temp) == 1 ) {
                                good2 = TRUE;
                                m_GpsModel = GPSMODEL_TRIMBLE;
                                }
                        else if ( sscanf((LPCTSTR)tempstr,"ONCORE",&temp) == 1 ) {
                                good2 = TRUE;
                                m_GpsModel = GPSMODEL_ONCORE;
                                }
                        else if ( sscanf((LPCTSTR)tempstr,"PCTIME",&temp) == 1 ) {
                                good2 = TRUE;
                                m_GpsModel = GPSMODEL_PCTIME;
                                }
                        else if ( sscanf((LPCTSTR)tempstr,"PARGPS",&temp) == 1 ) {
                                good2 = TRUE;
                                m_GpsModel = GPSMODEL_TRIMBLE;
                                }
                        else
                                good2 = FALSE;

                        SetComboModel();

                        good = (good1 && good2);

                        break;

                case IDC_CBO_G_SERIALPORT:

                        GetDlgItemText( IDC_CBO_G_SERIALPORT, pAPP->szTemp, MAXSTR );
                        tempstr = pAPP->szTemp;

                        if ( sscanf((LPCTSTR)tempstr,"COM%d",&temp) == 1 ) {
                                good2 = TRUE;
                                m_SerialPort = temp;
                                }
                        else
                                good2 = FALSE;

                        SetComboPort();

                        good = (good1 && good2);

                        break;



                default:
                        good = TRUE;
                        break;


                }


        // Re-allow validation

        DoValidate = TRUE;


        if (good)
                return( 1 );
        else
                return( 0 );
}


void CGpsDlg::OnSelchangeName() {

        TRACE0("CGpsDlg::OnSelchangeName\n");


        m_Index = m_CBO_DriverName.GetCurSel();

        if ( (m_Index == CB_ERR) || (m_Index == -1) ) {

                GetDlgItemText( IDC_CBO_G_DRIVERNAME, pAPP->szTemp, MAXSTR );
                m_DriverName = pAPP->szTemp;
                }

        else {
                m_CBO_DriverName.GetLBText( m_Index, pAPP->szTemp );
                m_DriverName = pAPP->szTemp;
                }
}

void CGpsDlg::OnKillfocusName() {

        TRACE0("CGpsDlg::OnKillfocusName\n");

        PostMessage( WM_EDITLOSTFOCUS, IDC_CBO_G_DRIVERNAME, 0 );
}

void CGpsDlg::SetComboName() {

        TRACE1("CGpsDlg::SetComboName to %s\n", (LPCTSTR)m_DriverName);


        // Find string in list

        m_Index = m_CBO_DriverName.FindString( -1, m_DriverName );



        if ( (m_Index == CB_ERR) || (m_Index == -1) )

                SetDlgItemText( IDC_CBO_G_DRIVERNAME, m_DriverName );
        else
                m_CBO_DriverName.SetCurSel( m_Index );
}



void CGpsDlg::OnSelchangeModel() {

        TRACE0("CGpsDlg::OnSelchangeModel\n");

        m_Index = m_CBO_GpsModel.GetCurSel();

        if ( m_Index == GPSMODELINDEX_GARMIN )
                m_GpsModel = GPSMODEL_GARMIN;

        else if ( m_Index == GPSMODELINDEX_TRIMBLE )
                m_GpsModel = GPSMODEL_TRIMBLE;

        else if ( m_Index == GPSMODELINDEX_ONCORE )
                m_GpsModel = GPSMODEL_ONCORE;

        else if ( m_Index == GPSMODELINDEX_PCTIME )
                m_GpsModel = GPSMODEL_PCTIME;

        else { // (m_Index == CB_ERR) || (m_Index == -1) || m_Index unknown
                m_GpsModel = GPSMODEL_TRIMBLE;
                m_CBO_GpsModel.SetCurSel( GPSMODELINDEX_TRIMBLE );
                }
}

void CGpsDlg::SetComboModel() {

        TRACE1("CGpsDlg::SetComboModel to %d\n", m_GpsModel);

        if ( m_GpsModel == GPSMODEL_GARMIN )
                m_Index = GPSMODELINDEX_GARMIN;

        else if ( m_GpsModel == GPSMODEL_TRIMBLE )
                m_Index = GPSMODELINDEX_TRIMBLE;

        else if ( m_GpsModel == GPSMODEL_ONCORE )
                m_Index = GPSMODELINDEX_ONCORE;

        else if ( m_GpsModel == GPSMODEL_PCTIME )
                m_Index = GPSMODELINDEX_PCTIME;

        else
                m_Index = GPSMODELINDEX_TRIMBLE;

        m_CBO_GpsModel.SetCurSel( m_Index );
}

void CGpsDlg::OnKillfocusModel() {

        TRACE0("CGpsDlg::OnKillfocusModel\n");

        PostMessage( WM_EDITLOSTFOCUS, IDC_CBO_G_GPSMODEL, 0 );
}

void CGpsDlg::OnSelchangePort() {

        TRACE0("CGpsDlg::OnSelchangePort\n");


        m_Index = m_CBO_SerialPortName.GetCurSel();

        if ( (m_Index == CB_ERR) || (m_Index == -1) ) {

                GetDlgItemText( IDC_CBO_G_SERIALPORT, pAPP->szTemp, MAXSTR );
                if (sscanf( pAPP->szTemp, "COM%d", &m_SerialPort ) != 1) {
                        // Error
                        }
                }

        else {
                m_CBO_SerialPortName.GetLBText( m_Index, pAPP->szTemp );
                m_SerialPort = m_Index+1;
                }
}

void CGpsDlg::OnKillfocusPort() {

        TRACE0("CGpsDlg::OnKillfocusPort\n");

        PostMessage( WM_EDITLOSTFOCUS, IDC_CBO_G_SERIALPORT, 0 );
}

void CGpsDlg::SetComboPort() {

        TRACE1("CGpsDlg::SetComboPort to COM%d\n", m_SerialPort);


        // Make string and find in list

        sprintf( pAPP->szTemp, "COM%d", m_SerialPort );
        m_Index = m_CBO_SerialPortName.FindString( -1, pAPP->szTemp );



        if ( (m_Index == CB_ERR) || (m_Index == -1) )

                SetDlgItemText( IDC_CBO_G_SERIALPORT, pAPP->szTemp );
        else
                m_CBO_SerialPortName.SetCurSel( m_Index );
}


void CGpsDlg::SetActivation( void  ) {

        // Turn all other controls on or off depending on GpsEnable

        GetDlgItem(IDC_LAB_G_DRIVERNAME   )->EnableWindow( m_GpsEnable );
        GetDlgItem(IDC_CBO_G_DRIVERNAME   )->EnableWindow( m_GpsEnable );
        GetDlgItem(IDC_LAB_G_GPSMODEL     )->EnableWindow( m_GpsEnable );
        GetDlgItem(IDC_CBO_G_GPSMODEL     )->EnableWindow( m_GpsEnable );
        GetDlgItem(IDC_LAB_G_SERIALPORT   )->EnableWindow( m_GpsEnable );
        GetDlgItem(IDC_CBO_G_SERIALPORT   )->EnableWindow( m_GpsEnable );
        GetDlgItem(IDC_CHK_G_DISPLAYPPS   )->EnableWindow( m_GpsEnable );
        GetDlgItem(IDC_CHK_G_DISPLAYSERIAL)->EnableWindow( m_GpsEnable );
}

void CGpsDlg::OnChkGpsEnable(  ) {

        m_GpsEnable = !m_GpsEnable;

        SetActivation();
}

void CGpsDlg::OnChkDisplayPps(  ) {

        m_DisplayPps = !m_DisplayPps;
}

void CGpsDlg::OnChkDisplaySerial(  ) {

        m_DisplaySerial = !m_DisplaySerial;
}





//***********************************************************************
//
// CLASS CContentsDlg
//
// This class, derived from CScopeDialog, presents and controls the Help
// Contents dialog.  This dialog lists the topics for which help text is
// available and allows the user to select and view this text.

BEGIN_MESSAGE_MAP(CContentsDlg, CScopeDialog)
        //{{AFX_MSG_MAP(CContentsDlg)
        ON_BN_CLICKED(IDC_BTN_H_DISPLAY,   OnBtnDisplay)
        ON_BN_CLICKED(IDC_BTN_H_HELP,      OnHelp)
        ON_BN_CLICKED(IDC_BTN_H_CLOSE,     OnCancel)
        ON_LBN_DBLCLK(IDC_LST_H_TOPICS,    OnDblclkTopics)
        //}}AFX_MSG_MAP
END_MESSAGE_MAP()


CContentsDlg::CContentsDlg(CWnd* pParent /*=NULL*/)
            : CScopeDialog(CContentsDlg::IDD, pParent) {

}


void CContentsDlg::DoDataExchange(CDataExchange* pDX) {

        CScopeDialog::DoDataExchange(pDX);

        //{{AFX_DATA_MAP(CContentsDlg)
        DDX_Control(pDX, IDC_LST_H_TOPICS, m_LST_topics);
        //}}AFX_DATA_MAP
}


void CContentsDlg::OnBtnDisplay() {

        int   pos, titleID, textID;
        DWORD item;


        titleID = IDR_HELP_UNKNOWN_TITLE;
        textID  = IDR_HELP_UNKNOWN_TEXT;

        pos = m_LST_topics.GetCurSel();

        if (pos != LB_ERR) {

                item    = m_LST_topics.GetItemData(pos);
                titleID = LOWORD(item);
                textID  = HIWORD(item);

                }

        pAPP->ShowHelp( titleID, textID, &m_LST_topics );
}

void CContentsDlg::OnHelp() {

        pAPP->ShowHelp( IDR_HELP_HELP_TITLE, IDR_HELP_HELP_TEXT, this );
}

void CContentsDlg::OnCancel() {

        CScopeDialog::OnCancel();
}

BOOL CContentsDlg::OnInitDialog() {

        CScopeDialog::OnInitDialog( 0 );

        m_LST_topics.ResetContent();

        pAPP->LoadHelpTitles( &m_LST_topics );

        m_LST_topics.SetCurSel(0);


        m_LST_topics.SetFocus();

        return FALSE; // return TRUE unless you set the focus to a control
                      // EXCEPTION: OCX Property Pages should return FALSE
}

void CContentsDlg::OnDblclkTopics() {

        OnBtnDisplay();
}






//***********************************************************************
//
// CLASS CTopicDlg
//
// This class, derived from CScopeDialog, presents and controls the Help
// Topic dialog.  This dialog displays help text for the selected topic.
// This may have been selected directly by the user from the Help
// Contents dialog or it may have been selected automatically by the
// program based on context.

BEGIN_MESSAGE_MAP(CTopicDlg, CScopeDialog)
        //{{AFX_MSG_MAP(CTopicDlg)
        ON_WM_SIZE()
        ON_EN_SETFOCUS(IDC_TXT_H_TOPICTEXT, OnSetfocusTopic)
        //}}AFX_MSG_MAP
END_MESSAGE_MAP()


CTopicDlg::CTopicDlg(CWnd* pParent /*=NULL*/)
          : CScopeDialog(CTopicDlg::IDD, pParent) {

        TRACE0("CTopicDlg::CTopicDlg\n");

        //{{AFX_DATA_INIT(CTopicDlg)
        //}}AFX_DATA_INIT

        m_TitleID = IDR_HELP_UNKNOWN_TITLE;
        m_TextID  = IDR_HELP_UNKNOWN_TEXT;
}

CTopicDlg::CTopicDlg(int TitleID, int TextID, CWnd* pParent /*=NULL*/)
          : CScopeDialog(CTopicDlg::IDD, pParent) {

        TRACE0("CTopicDlg::CTopicDlg\n");

        m_TitleID = TitleID;
        m_TextID  = TextID;
}

void CTopicDlg::DoDataExchange(CDataExchange* pDX) {

        CScopeDialog::DoDataExchange(pDX);

        //{{AFX_DATA_MAP(CTopicDlg)
        DDX_Control(pDX, IDC_TXT_H_TOPICTEXT, m_TXT_Topic);
        //}}AFX_DATA_MAP
}


BOOL CTopicDlg::OnInitDialog() {

        int iret;

        TRACE0("CTopicDlg::OnInitDialog\n");

        // Initialize all member variables from application globals


        CScopeDialog::OnInitDialog( 0 );


        // Set left and right margin borders in pixels
        
        m_TXT_Topic.SetMargins( 15, 20 );



        // Set title

        htitle = (LPCSTR)::LockResource(
                         ::LoadResource( NULL,
                         ::FindResource( NULL,
                            (LPCTSTR)m_TitleID, RT_RCDATA)) );
        SetWindowText( htitle );




        // Set text

        htext = (LPCSTR)::LockResource(
                        ::LoadResource( NULL,
                        ::FindResource( NULL,
                            (LPCTSTR)m_TextID, RT_RCDATA)) );
        m_TXT_Topic.SetWindowText( htext );


        CWnd *pWnd = GetDlgItem( IDC_TXT_H_TOPICTEXT );
        if (pWnd) {
                pWnd->SetFocus( );
                iret = FALSE;
                }
        else
                iret = TRUE;


        TRACE0("CTopicDlg::OnInitDialog END\n");

        return iret;  // return TRUE unless you set the focus to a control
                      // EXCEPTION: OCX Property Pages should return FALSE
}



void CTopicDlg::OnSize(UINT nType, int cx, int cy) {

        TRACE0("CTopicDlg::OnSize\n");

        CScopeDialog::OnSize( nType, cx, cy );

        m_TXT_Topic.MoveWindow( 0, 0, cx, cy );
}

void CTopicDlg::OnSetfocusTopic() {

        TRACE0("CTopicDlg::OnSetfocusTopic\n");

        m_TXT_Topic.SetSel( 0, 0, FALSE );
}




//***********************************************************************
//
// CLASS CSrParser
//
// This class, derived from CObject, provides the basic parser helper
// functions.  The CScopeApp object must provide its own top level parsing
// function that contains the information about what keywords are
// understood and what actions to take when they are found.
//



CSrParser::CSrParser( )
          :CObject( ) {

        TRACE0("CSrParser::CSrParser\n");

        IniFp   = NULL;
        fline   = -1;
        c       = ' ';
        token   = NULL;
        ttype   = TOKEN_UNKNOWN;

        strcpy( IniName, "" );
}


int CSrParser::OpenIni( const char *filename ) {

        TRACE0("CSrParser::OpenIni\n");

        // Check for already open

        if (IniFp) {
                pAPP->ErrString.Format( "ERROR[%s] : File already open.", filename );
                pAPP->LogMessage( MSGLEVEL_ERR, MSGDLG_SHOW, ERROR_NOT_FATAL, NULL );
                return( 0 );
                }


        // Open and check for failure

        IniFp = fopen( filename, "r" );

        if (!IniFp) {
                pAPP->ErrString.Format( "ERROR[%s] : Could not open file.", filename );
                pAPP->LogMessage( MSGLEVEL_ERR, MSGDLG_SHOW, ERROR_NOT_FATAL, NULL );
                return( 0 );
                }


        // Prepare for parsing

        fline = 1;
        c     = ' ';
        token = NULL;
        ttype = TOKEN_UNKNOWN;
        SrStringCopy( IniName, filename, MAXSTR );

        return( 1 );

}

int CSrParser::CloseIni( void ) {

        TRACE0("CSrParser::CloseIni\n");

        // Close file

        fclose( IniFp );


        // Reset parsing variables

        IniFp = NULL;
        fline = -1;
        c     = ' ';
        token = NULL;
        ttype = TOKEN_UNKNOWN;
        strcpy( IniName, "" );

        return( 1 );
}

int CSrParser::ErrorIni( const char *msg ) {

        TRACE0("CSrParser::ErrorIni\n");

        // Display an .ini error to the log file

        pAPP->ErrString.Format( "ERROR[%s] : line %d : %s.", IniName, fline, msg );
        pAPP->LogMessage( MSGLEVEL_ERR, MSGDLG_NO_SHOW, ERROR_NOT_FATAL, NULL );


        // Give the user a choice about what to do

        pAPP->ErrString += "\n\nPress ok to continue processing ini, ";
        pAPP->ErrString += "cancel to restore saved values.\n";

        if (AfxMessageBox(pAPP->ErrString, MB_OKCANCEL) == IDCANCEL)
                pAPP->ErrCode = 1;
        else
                pAPP->ErrCode = 0;


        // Send back if ok or cancel pressed
        
        return( !pAPP->ErrCode );
}

void CSrParser::AddToToken( int lc ) {

//        TRACE0("CSrParser::AddToToken\n");

        // Add c to TOKEN, checking for token buffer overflow.

        if ( (token - TOKEN) < MAXTOKEN )

                *token++ = (char)lc;

        else
                ErrorIni( "internal token length exceeded" );
}


int CSrParser::SkipSpace( int lc ) {

//        TRACE0("CSrParser::SkipSpace\n");

        // Skip over white space, including comments.
skip:

        if ( isspace( lc ) ) {

                if ( lc == '\n' )

                        fline++;


                lc = fgetc( IniFp );

                goto skip;

                }


        else if ( lc == ';' ) { // ; comment ...

                while ( lc != '\n' && lc != EOF )

                        lc = fgetc( IniFp );

                goto skip;

                }


        return( lc );
}


int CSrParser::GetToken( void ) {

//        TRACE0("CSrParser::GetToken\n");


        // Skip over any white space.

        c = SkipSpace( c );



        // Gather the next token.

        token = TOKEN;


        if ( isalpha( c ) || c == '_' ) {

                ttype = TOKEN_SYMBOL;

                while ( isalnum(  c ) || c == '_' || c == '/' )

                        { AddToToken( c ); c = fgetc( IniFp ); }
                }


        else if ( c == '=' ) {

                ttype = TOKEN_EQUALS;

                AddToToken( c ); c = fgetc( IniFp );

                }


        else if ( isdigit( c ) ) {

                // ttype could be HEX, INTEGER, or FLOAT ...

                AddToToken( c );

                c = fgetc( IniFp );



                if ( TOKEN[0] == '0' && c == 'x' ) {

                        ttype = TOKEN_HEX;
                        token = TOKEN;

                        c = fgetc( IniFp );

                        if ( isxdigit( c ) ) {

                                while ( isxdigit( c ) )

                                        { AddToToken( c ); c = fgetc( IniFp ); }
                                }

                        else

                                ttype = TOKEN_UNKNOWN;
                        }

                else {

                        ttype = TOKEN_INTEGER;

                        while ( isdigit( c ) )

                                { AddToToken( c ); c = fgetc( IniFp ); }



                        if ( c == '.' ) {

                                ttype = TOKEN_FLOAT;

                                AddToToken( c );

                                c = fgetc( IniFp );

                                while ( isdigit( c ) )

                                        { AddToToken( c ); c = fgetc( IniFp ); }
                                }
                        }
                }



        else if ( c == '"' ) {

                ttype = TOKEN_STRING;

                c = fgetc( IniFp );

                  while ( c != '"' && c != '\n' && c != EOF )

                        { AddToToken( c ); c = fgetc( IniFp ); }



                if ( c == '"' )

                        c = fgetc( IniFp );

                else
                        ErrorIni( "missing quote on string" );

                }



        else if ( c == EOF ) {

                ttype = TOKEN_EOF;
                c = ' ';                // re-initialize for next file
                }


        else {

                ttype = TOKEN_UNKNOWN;

                AddToToken( c );

                c = fgetc( IniFp );

                }



        // Null terminate the TOKEN and return its type.

        AddToToken( '\0' );

        return( ttype );
}


int CSrParser::Match( const char *s ) {

//        TRACE0("CSrParser::Match\n");

        char *t;


        // Check for a match between s and TOKEN.

        t = TOKEN;

        while ( *s && *t && toupper( *s ) == toupper( *t ) )

                { s++; t++; }


        return ( *s == '\0' && *t == '\0' ? 1 : 0 );
}


int CSrParser::KeywordEquals( const char *keyword ) {

//        TRACE0("CSrParser::KeywordEquals\n");

        char msg[80];


        if ( Match( keyword ) ) {

                if ( GetToken() == TOKEN_EQUALS )

                     return( 1 );

                else {
                     sprintf( msg, "expected = sign after %s keyword", keyword );
                     ErrorIni( msg );
                     }
                }

        return( 0 );
}


int CSrParser::GetOnOff( int *parameter, const char *error ) {

//        TRACE0("CSrParser::GetOnOff\n");

        // Get an {ON,OFF} parameter value.

        GetToken();

        if ( Match( "ON" ) )
                *parameter = ON;

        else if ( Match( "OFF" ) )
                *parameter = OFF;

        else {
                ErrorIni( error );
                return( 0 );
                }

        return( 1 );
}


int CSrParser::GetInt( int *parameter, const char *error ) {

//        TRACE0("CSrParser::GetInt\n");

        // Get a decimal int parameter value.

        GetToken();

        if ( ttype != TOKEN_INTEGER ) {

                ErrorIni( error );
                return( 0 );
                }

        else {
                sscanf( TOKEN, "%d", parameter );
                return( 1 );
                }
}


int CSrParser::GetUint( unsigned int *parameter, const char *error ) {

//        TRACE0("CSrParser::GetUint\n");

        // Get a decimal int parameter value.

        GetToken();

        if ( ttype != TOKEN_INTEGER ) {

                ErrorIni( error );
                return( 0 );
                }

        else {
                sscanf( TOKEN, "%u", parameter );
                return( 1 );
                }
}


int CSrParser::GetUlong( unsigned long *parameter, const char *error ) {

//        TRACE0("CSrParser::GetUlong\n");

        // Get an unsigned long parameter value.

        GetToken();

        if ( ttype != TOKEN_INTEGER ) {

                ErrorIni( error );
                return( 0 );
                }

        else {
                sscanf( TOKEN, "%lu", parameter );
                return( 1 );
                }
}


int CSrParser::GetHex( int *parameter, const char *error ) {

//        TRACE0("CSrParser::GetHex\n");

        // Get a hex parameter value.

        GetToken();

        if ( ttype != TOKEN_HEX ) {

                ErrorIni( error );
                return( 0 );
                }

        else {
                sscanf( TOKEN, "%x", parameter );
                return( 1 );
                }
}


int CSrParser::GetFloat( double *parameter, const char *error ) {

//        TRACE0("CSrParser::GetFloat\n");

        // Get a floating point parameter value.

        GetToken();

        if ( (ttype != TOKEN_FLOAT) && (ttype != TOKEN_INTEGER) ) {

                ErrorIni( error );
                return( 0 );
                }

        else {
                sscanf( TOKEN, "%lf", parameter );
                return( 1 );
                }
}


int CSrParser::IsEOF( void ) {

        TRACE0("CSrParser::IsEOF\n");

        if ( ttype == TOKEN_EOF )
                return( 1 );
        else
                return( 0 );
}
